cockpit.locale({
 "": {
  "plural-forms": (n) => 0,
  "language": "undefined",
  "language-direction": "ltr"
 },
 "Task Log": [
  null,
  "タスクログ"
 ],
 "Close": [
  null,
  "閉じる"
 ],
 "Cancel": [
  null,
  "キャンセル"
 ],
 "Replication Agreement Details (": [
  null,
  "レプリカ合意の詳細 ("
 ],
 "Replica": [
  null,
  "レプリカ"
 ],
 "Agreement Enabled": [
  null,
  "有効なレプリカ合意"
 ],
 "Last Init Started": [
  null,
  "前回の初期化開始日時"
 ],
 "Last Init Ended": [
  null,
  "前回の初期化終了日時"
 ],
 "Last Initialization Status": [
  null,
  "前回の初期化ステータス"
 ],
 "Replication In Progress": [
  null,
  "現在のレプリケーションの進行状況"
 ],
 "Changes Sent": [
  null,
  "送信された変更の数"
 ],
 "Last Update Started": [
  null,
  "最後に更新が開始された日時"
 ],
 "Last Update Ended": [
  null,
  "最後の更新が終了した日時"
 ],
 "Last Update Status": [
  null,
  "最後の更新の状態"
 ],
 "Resolve Replication Conflict": [
  null,
  "レプリケーションの競合を解決"
 ],
 "Resolve Conflict": [
  null,
  "競合を解決"
 ],
 "Valid Entry": [
  null,
  "有効なエントリ"
 ],
 "Child Entries: ": [
  null,
  "子エントリ: "
 ],
 "Conflict Entry": [
  null,
  "競合エントリ"
 ],
 "Delete Conflict Entry": [
  null,
  "競合エントリを削除"
 ],
 "This will delete the conflict entry, and the \"valid\" entry will remain intact.": [
  null,
  "これにより、競合エントリが削除され、'有効'なエントリがそのまま残ります。"
 ],
 "Swap Conflict Entry With Valid Entry": [
  null,
  "競合エントリを有効なエントリと入れ替える"
 ],
 "This will replace the \"valid\" entry with the conflict entry, but keeping the valid entry DN intact.": [
  null,
  "「有効な」エントリを競合エントリで置き換えますが、有効なエントリのDNはそのまま維持されます。"
 ],
 "Convert Conflict Entry Into New Entry": [
  null,
  "競合エントリを新しいエントリに変換"
 ],
 "The conflict entry uses a multi-valued RDN to specify the original DN and it's nsUniqueID.  To convert the conflict entry to a new entry you must provide a new RDN attribute/value for the new entry.  \"RDN_ATTRIBUTE=VALUE\".  For example: cn=my_new_entry": [
  null,
  "競合エントリは、元のDNとそのnsUniqueIDを指定するために複数の値を持つRDNを使用します。競合エントリを新しいエントリに変換するには、新しいエントリのRDN属性/値を指定する必要があります。「RDN属性=値」の形式で指定します。例えば、cn=my_new_entry とします。"
 ],
 "Enter new RDN here": [
  null,
  "ここに新しいRDNを入力してください"
 ],
 "Add": [
  null,
  "追加"
 ],
 "Edit": [
  null,
  "編集"
 ],
 "$0 Report Credentials": [
  null,
  "$0 レポートの認証情報"
 ],
 "Save": [
  null,
  "保存"
 ],
 "Hostname": [
  null,
  "ホスト名"
 ],
 "Port": [
  null,
  "ポート"
 ],
 "Bind DN for the specified instances": [
  null,
  "指定されたインスタンスのバインドDNを設定します。"
 ],
 "Bind DN": [
  null,
  "バインドDN"
 ],
 "Bind password for the specified instances": [
  null,
  "指定されたインスタンスのバインドパスワード"
 ],
 "Password": [
  null,
  "パスワード"
 ],
 "Input the password interactively": [
  null,
  "パスワードを対話形式で入力します"
 ],
 "Interactive Input": [
  null,
  "パスワードを対話形成で入力する"
 ],
 "Add Replica Connection": [
  null,
  "レプリカ接続を追加"
 ],
 "Connection Name": [
  null,
  "接続名"
 ],
 "Bind password for the specified instance.  You can also specify a password file but the filename needs to be inside of brackets [/PATH/FILE]": [
  null,
  "指定されたインスタンスのバインドパスワードを入力してください。または、パスワードファイルを指定する場合は、ファイル名を角かっこで囲んでください([/PATH/FILE])。"
 ],
 "Input the password interactively, stores '*' as the password value in .dsrc": [
  null,
  "パスワードを対話形式で入力し、'.dsrc' ファイル内のパスワード値を'*' で保存します。"
 ],
 "$0 Report Alias": [
  null,
  "レポートのエイリアスを$0"
 ],
 "Alias name for the instance": [
  null,
  "インスタンスのエイリアス名を設定します"
 ],
 "Alias": [
  null,
  "エイリアス"
 ],
 "An instance hostname": [
  null,
  "インスタンスのホスト名を設定します"
 ],
 "An instance port": [
  null,
  "インスタンスのポートを設定します"
 ],
 "Replication Login Credentials for $0": [
  null,
  "$0 のレプリケーションログイン資格情報"
 ],
 "Confirm Credentials Input": [
  null,
  "資格情報の入力を確認"
 ],
 "In order to get the replication agreement lag times and state, the authentication credentials to the remote replicas must be provided.": [
  null,
  "リモートレプリカへのレプリカ合意の遅延時間と状態を取得するために、リモートレプリカへの認証資格情報を提供する必要があります。"
 ],
 "Bind DN was acquired from <b>Replica Credentials</b> table. If you want to bind as another user, change or remove the Bind DN there.": [
  null,
  "バインドDNはレプリカの認証情報テーブルから取得されました。別のユーザとしてバインドする場合は、そこでバインドDNを変更または削除してください。"
 ],
 "Bind DN for the instance": [
  null,
  "インスタンスのバインド名"
 ],
 "Password for the Bind DN": [
  null,
  "バインドDNのパスワード"
 ],
 "Refreshing": [
  null,
  "更新中..."
 ],
 "Loading": [
  null,
  "読み込み中"
 ],
 "Do the refresh every few seconds": [
  null,
  "数秒ごとに更新中..."
 ],
 "$0 the report...": [
  null,
  "$0をレポートしています..."
 ],
 "There is no report, you must first generate the report in the <b>Prepare Report</b> tab.": [
  null,
  "レポートはありません。まずは「レポートの準備」タブでレポートを生成する必要があります。"
 ],
 "Display all agreements including the disabled ones and the ones we failed to connect to": [
  null,
  "全ての合意を表示します。これには無効な合意や接続に失敗した合意も含まれます。"
 ],
 "Show Disabled Agreements": [
  null,
  "無効なレプリカ合意を表示する"
 ],
 "Refresh Report": [
  null,
  "レポートを更新"
 ],
 "Show all data in one table (it makes it easier to check lag times)": [
  null,
  "すべてのデータを1つのテーブルに表示します(遅延時間を確認するのに便利です)"
 ],
 "Table View": [
  null,
  "テーブルの一覧を表示する"
 ],
 "Can not get replication information from Replica:": [
  null,
  "レプリカからのレプリケーション情報を取得できません:"
 ],
 "Replica Root": [
  null,
  "レプリカのルートエントリ"
 ],
 "Replica ID": [
  null,
  "レプリカＩＤ"
 ],
 "Max CSN": [
  null,
  "最大CSN"
 ],
 "Replica Status": [
  null,
  "レプリカの状態"
 ],
 "No Agreements Were Found": [
  null,
  "合意は見つかりませんでした"
 ],
 "Supplier host:port (and alias if applicable)": [
  null,
  "サプライヤーのホスト:ポート(および適用される場合はエイリアス)"
 ],
 "Supplier:": [
  null,
  "サプライヤー"
 ],
 "Refresh chaining monitor": [
  null,
  "チェイン監視を更新"
 ],
 "Add Operations": [
  null,
  "操作を追加"
 ],
 "Delete Operations": [
  null,
  "操作を削除"
 ],
 "Modify Operations": [
  null,
  "操作を更新"
 ],
 "ModRDN Operations": [
  null,
  "ModRDN操作"
 ],
 "Base Searches": [
  null,
  "ベース検索"
 ],
 "One-Level Searches": [
  null,
  "ワンレベル検索"
 ],
 "Subtree Searches": [
  null,
  "サブツリー検索"
 ],
 "Abandon Operations": [
  null,
  "キャンセル操作"
 ],
 "Bind Operations": [
  null,
  "バインド操作"
 ],
 "Unbind Operations": [
  null,
  "バインド解除操作"
 ],
 "Compare Operations": [
  null,
  "比較操作"
 ],
 "Outgoing Connections": [
  null,
  "外部接続"
 ],
 "Outgoing Bind Connections": [
  null,
  "外部バインド接続"
 ],
 "Cache Hit Ratio": [
  null,
  "キャッシュヒット率"
 ],
 "Cache Utilization": [
  null,
  "キャッシュ利用率"
 ],
 "Loading database monitor information ...": [
  null,
  "データベースの監視情報を読み込んでいます..."
 ],
 "Database Cache": [
  null,
  "データベースキャッシュ"
 ],
 "The database cache hit ratio (dbcachehitratio).": [
  null,
  "データベースのキャッシュヒット率(dbcachehitratio)"
 ],
 "Live Database Cache Statistics": [
  null,
  "ライブデータベースのキャッシュ統計"
 ],
 "Database Cache Hit Ratio:": [
  null,
  "データベースのキャッシュヒット率:"
 ],
 "Database Cache Tries:": [
  null,
  "データベースのキャッシュトライ数:"
 ],
 "Database Cache Hits:": [
  null,
  "データベースのキャッシュヒット数:"
 ],
 "Cache Pages Read:": [
  null,
  "読み込まれたキャッシュページの数"
 ],
 "Cache Pages Written:": [
  null,
  "書き込まれたキャッシュページの数"
 ],
 "Read-Only Page Evictions:": [
  null,
  "読み取り専用のページが削除された回数"
 ],
 "Read-Write Page Evictions:": [
  null,
  "読み書き可能なページが削除された回数"
 ],
 "Normalized DN Cache": [
  null,
  "正規化されたDNキャッシュ"
 ],
 "The normalized DN cache hit ratio (normalizeddncachehitratio).": [
  null,
  "正規化されたDNキャッシュヒット率(normalizeddncachehitratio)"
 ],
 "Live Normalized DN Cache Statistics": [
  null,
  "正規化されたDNキャッシュのリアルタイム統計情報"
 ],
 "The amount of the cache that is being used: max size (maxnormalizeddncachesize) vs current size (currentnormalizeddncachesize)": [
  null,
  "使用されているキャッシュの量: 最大サイズ(maxnormalizeddncachesize)と現在のサイズ(currentnormalizeddncachesize)の比較"
 ],
 "Cached DN's": [
  null,
  "キャッシュされたDN"
 ],
 "Live Normalized DN Cache Utilization Statistics": [
  null,
  "正規化されたDNキャッシュの利用状況の統計情報"
 ],
 "NDN Cache Hit Ratio:": [
  null,
  "正規化されたDNキャッシュのヒット率"
 ],
 "NDN Cache Max Size:": [
  null,
  "正規化されたDNキャッシュの最大サイズ"
 ],
 "NDN Cache Tries:": [
  null,
  "正規化されたDNキャッシュへのアクセス試行回数"
 ],
 "NDN Current Cache Size:": [
  null,
  "現在の正規化されたDNキャッシュのサイズ"
 ],
 "NDN Cache Hits:": [
  null,
  "正規化されたDNキャッシュへのヒット数"
 ],
 "NDN Cache DN Count:": [
  null,
  "正規化されたDNキャッシュ内のDNの数"
 ],
 "NDN Cache Evictions:": [
  null,
  "正規化されたDNキャッシュから削除されたエントリの数"
 ],
 "NDN Cache Thread Size:": [
  null,
  "正規化されたDNキャッシュのスレッドサイズ"
 ],
 "NDN Cache Thread Slots:": [
  null,
  "正規化されたDNキャッシュのスレッドスロット数"
 ],
 "Database Performance Statistics": [
  null,
  "データベースのパフォーマンス統計"
 ],
 "Replication agreement has been poked": [
  null,
  "レプリカ合意が変更されました"
 ],
 "Failed to poke replication agreement $0 - $1": [
  null,
  "レプリカ合意 $0 の変更に失敗しました - $1"
 ],
 "Monitor Replication Agreements": [
  null,
  "レプリカ合意の監視"
 ],
 "Refresh replication monitor": [
  null,
  "レプリケーション監視を更新する"
 ],
 "Replication conflict entry was converted into a valid entry": [
  null,
  "競合エントリが有効なエントリに変換されました"
 ],
 "Failed to convert conflict entry entry: $0 - $1": [
  null,
  "競合エントリの変換に失敗しました: $0 - $1"
 ],
 "Replication Conflict Entry is now the Valid Entry": [
  null,
  "レプリケーションの競合エントリは、有効なエントリに変換されました。"
 ],
 "Failed to swap in conflict entry: $0 - $1": [
  null,
  "競合エントリの置換に失敗しました: $0 - $1"
 ],
 "Replication conflict entry was deleted": [
  null,
  "競合エントリが削除されました"
 ],
 "Failed to delete conflict entry: $0 - $1": [
  null,
  "競合エントリの削除に失敗しました: $0 - $1"
 ],
 "Failed to get conflict entries: $0 - $1": [
  null,
  "競合エントリを取得できませんでした: $0 - $1"
 ],
 "Replication glue entry was converted": [
  null,
  "レプリケーションのグルーエントリが変換されました"
 ],
 "Failed to convert glue entry: $0 - $1": [
  null,
  "グルーエントリの変換に失敗しました: $0 - $1"
 ],
 "Replication glue entry was deleted": [
  null,
  "グルーエントリが削除されました"
 ],
 "Failed to delete glue entry: $0 - $1": [
  null,
  "グルーエントリの削除に失敗しました。エラー: $0 - $1"
 ],
 "You must provide a RDN if you want to convert the Conflict Entry": [
  null,
  "競合エントリを変換するには、RDNを指定する必要があります"
 ],
 "Monitor Conflict and Glue Entries": [
  null,
  "競合エントリとグルーエントリの監視"
 ],
 "Conflict Entries ": [
  null,
  "競合エントリ"
 ],
 "Replication conflict entries occur when two entries are created with the same DN (or name) on different servers at about the same time.  The automatic conflict resolution procedure renames the entry created last.  Its RDN is changed into a multi-valued RDN that includes the entry's original RDN and it's unique identifier (nsUniqueId).  There are several ways to resolve a conflict, but choosing which option to use is up to you.": [
  null,
  "レプリケーションの競合エントリは、異なるサーバでほぼ同時に同じDN(または名前)で作成された場合に発生します。自動的な競合解決手順では、最後に作成されたエントリのRDNが変更されます。元のRDNと一意の識別子(nsUniqueId)を含む複数の値を持つRDNに変更されます。競合を解決するためのいくつかの方法がありますが、使用するオプションはあなた次第です。"
 ],
 "What Is A Replication Conflict Entry?": [
  null,
  "レプリケーション競合エントリとは何ですか？"
 ],
 "Glue Entries ": [
  null,
  "グルーエントリ"
 ],
 "When a Delete operation is replicated and the consumer server finds that the entry to be deleted has child entries, the conflict resolution procedure creates a \"glue entry\" to avoid having orphaned entries in the database.  In the same way, when an Add operation is replicated and the consumer server cannot find the parent entry, the conflict resolution procedure creates a \"glue entry\", representing the \"parent entry\", so that the new entry is not an orphaned entry.  You can choose to convert the glue entry, or remove the glue entry and all its child entries.": [
  null,
  "削除操作がレプリケートされ、コンシューマーサーバが削除対象のエントリに子エントリが存在することを検出した場合、競合解決手順ではデータベース内に孤立したエントリが残らないように、「グルーエントリ」を作成します。同様に、追加操作がレプリケートされ、コンシューマーサーバが親エントリを見つけることができない場合、競合解決手順では新しいエントリが孤立したエントリにならないように、「グルーエントリ」が作成されます。グルーエントリを変換するか、グルーエントリとその子エントリをすべて削除するかを選択することができます。"
 ],
 "What Is A Replication Glue Entry?": [
  null,
  "レプリケーショングルーエントリとは何ですか？"
 ],
 "Delete Glue Entry": [
  null,
  "グルーエントリを削除"
 ],
 "Are you really sure you want to delete this glue entry and its child entries?": [
  null,
  "本当にこのグルーエントリとその子エントリを削除しますか？"
 ],
 "Deleting Glue Entry ...": [
  null,
  "グルーエントリを削除しています ..."
 ],
 "Delete Glue": [
  null,
  "グルーエントリを削除"
 ],
 "Convert Glue Entry": [
  null,
  "グルーエントリを変換"
 ],
 "Are you really sure you want to convert this glue entry to a regular entry?": [
  null,
  "本当にこのグルーエントリを通常のエントリに変換しますか？"
 ],
 "Converting Glue Entry ...": [
  null,
  "グルーエントリを変換しています ..."
 ],
 "Convert Glue": [
  null,
  "グルーを変換"
 ],
 "Are you really sure you want to convert this conflict entry?": [
  null,
  "本当にこの競合エントリを変換しますか？"
 ],
 "Converting Conflict Entry ...": [
  null,
  "競合エントリを変換しています ..."
 ],
 "Convert Conflict": [
  null,
  "競合を変換"
 ],
 "Swap Conflict Entry": [
  null,
  "競合エントリを入れ替え"
 ],
 "Are you really sure you want to swap this conflict entry with the valid entry?": [
  null,
  "本当にこの競合エントリを有効なエントリと入れ替えますか？"
 ],
 "Swapping Conflict Entry ...": [
  null,
  "競合エントリを入れ替えています ..."
 ],
 "Swap Conflict": [
  null,
  "競合を入れ替え"
 ],
 "Delete Replication Conflict Entry": [
  null,
  "レプリケーション競合エントリを削除"
 ],
 "Are you really sure you want to delete this conflict entry?": [
  null,
  "本当にこの競合エントリを削除しますか？"
 ],
 "Deleting Conflict Entry ...": [
  null,
  "競合エントリを削除しています ..."
 ],
 "Delete Conflict": [
  null,
  "競合を削除"
 ],
 "Monitor Replication Tasks": [
  null,
  "レプリケーションタスクの監視"
 ],
 "CleanAllRUV Tasks ": [
  null,
  "CleanAllRUVタスク"
 ],
 "Abort CleanAllRUV Tasks ": [
  null,
  "CleanAllRUVタスクの中止"
 ],
 "Replication Winysnc agreement has been poked": [
  null,
  "レプリカWinysnc合意を変更しました"
 ],
 "Failed to poke replication winsync agreement $0 - $1": [
  null,
  "レプリカWinysnc合意 $0 の変更に失敗しました - $1"
 ],
 "Monitor Winsync Agreements": [
  null,
  "Winsync合意の監視"
 ],
 "Reloading security log...": [
  null,
  "セキュリティログを更新しています..."
 ],
 "Security Log": [
  null,
  "セキュリティログ"
 ],
 "Refresh security log": [
  null,
  "セキュリティログを更新する"
 ],
 "Continuously Refresh": [
  null,
  "継続的に更新する"
 ],
 "CPU": [
  null,
  "CPU"
 ],
 "Virtual Memory": [
  null,
  "仮想メモリ"
 ],
 "Resident Memory": [
  null,
  "常駐メモリ"
 ],
 "Connections": [
  null,
  "接続"
 ],
 "Server Statistics": [
  null,
  "サーバの統計情報"
 ],
 "Refresh suffix monitor": [
  null,
  "サフィックス監視を更新する"
 ],
 "Resource Charts": [
  null,
  "リソースチャート"
 ],
 "Established client connections to the server": [
  null,
  "サーバへの確立済みクライアント接続数"
 ],
 "Live Connection Statistics": [
  null,
  "リアルタイムの接続の統計"
 ],
 "Memory Usage": [
  null,
  "メモリ使用量"
 ],
 "Virtual Size": [
  null,
  "仮想サイズ"
 ],
 "Resident Size": [
  null,
  "常駐サイズ"
 ],
 "Live Server Memory Statistics": [
  null,
  "リアルタイムのサーバのメモリ統計"
 ],
 "CPU Usage": [
  null,
  "CPU使用率"
 ],
 "Server CPU Usage": [
  null,
  "サーバCPU使用率"
 ],
 "Server Stats": [
  null,
  "サーバの統計"
 ],
 "Server Instance": [
  null,
  "サーバインスタンス"
 ],
 "Version": [
  null,
  "バージョン"
 ],
 "Server Started": [
  null,
  "サーバ起動"
 ],
 "Server Uptime": [
  null,
  "サーバ稼働時間"
 ],
 "Worker Threads": [
  null,
  "ワーカースレッド"
 ],
 "Threads Waiting To Read": [
  null,
  "読み取りを待っているスレッド"
 ],
 "Conns At Max Threads": [
  null,
  "最大スレッド数の接続数"
 ],
 "Conns Exceeded Max Threads": [
  null,
  "最大スレッド数を超えた接続数"
 ],
 "Total Connections": [
  null,
  "総接続数"
 ],
 "Current Connections": [
  null,
  "現在の接続数"
 ],
 "Operations Started": [
  null,
  "開始された操作数"
 ],
 "Operations Completed": [
  null,
  "完了した操作数"
 ],
 "Entries Returned To Clients": [
  null,
  "クライアントに返されたエントリ数"
 ],
 "Bytes Sent to Clients": [
  null,
  "クライアントに送信されたバイト数"
 ],
 "Connection Table": [
  null,
  "接続テーブル"
 ],
 "Disk Space": [
  null,
  "ディスクスペース"
 ],
 "Refresh": [
  null,
  "更新"
 ],
 "SNMP Counters": [
  null,
  "SNMPカウンタ"
 ],
 "Anonymous Binds": [
  null,
  "匿名バインド"
 ],
 "Referrals": [
  null,
  "リファラル"
 ],
 "Unauthenticated Binds": [
  null,
  "未認証のバインド"
 ],
 "Returned Referrals": [
  null,
  "返ってきたリファラル"
 ],
 "Simple Auth Binds": [
  null,
  "単純な認証のバインド"
 ],
 "Bind Security Errors": [
  null,
  "バインドセキュリティエラー"
 ],
 "Strong Auth Binds": [
  null,
  "強力な認証のバインド"
 ],
 "Security Errors": [
  null,
  "セキュリティエラー"
 ],
 "Initiated Operations": [
  null,
  "開始した操作"
 ],
 "Errors": [
  null,
  "エラー"
 ],
 "Conns in Max Threads": [
  null,
  "最大スレッド数の接続数"
 ],
 "Bytes Received": [
  null,
  "受信したバイト数"
 ],
 "Search Operations": [
  null,
  "検索操作"
 ],
 "Bytes Sent": [
  null,
  "送信されたバイト数"
 ],
 "One Level Searches": [
  null,
  "ワンレベル検索"
 ],
 "Entries Returned": [
  null,
  "返ってきたエントリ"
 ],
 "Whole Tree Searches": [
  null,
  "ツリー全体を検索"
 ],
 "Loading Suffix Monitor Information ...": [
  null,
  "サフィックスの監視情報を読み込んでいます ..."
 ],
 "Entry Cache": [
  null,
  "エントリキャッシュ"
 ],
 "The entry cache hit ratio (entrycachehitratio)": [
  null,
  "エントリキャッシュのヒット率(entrycachehitratio)"
 ],
 "Live Entry Cache Statistics": [
  null,
  "リアルタイムのエントリキャッシュの統計情報"
 ],
 "The amount of the cache that is being used: max size (maxentrycachesize) vs current size (currententrycachesize)": [
  null,
  "使用中のキャッシュ量: 最大サイズ(maxentrycachesize) vs 現在のサイズ(currententrycachesize)"
 ],
 "Cached Entries": [
  null,
  "キャッシュされたエントリ"
 ],
 "Live Entry Cache Utilization Statistics": [
  null,
  "エントリキャッシュの利用状況の統計情報"
 ],
 "Entry Cache Hit Ratio:": [
  null,
  "エントリキャッシュのヒット率"
 ],
 "Entry Cache Max Size:": [
  null,
  "エントリキャッシュの最大サイズ"
 ],
 "Entry Cache Hits:": [
  null,
  "エントリキャッシュのヒット数"
 ],
 "Entry Cache Current Size:": [
  null,
  "エントリキャッシュの現在のサイズ"
 ],
 "Entry Cache Tries:": [
  null,
  "エントリキャッシュへのアクセス試行回数"
 ],
 "Entry Cache Max Entries:": [
  null,
  "エントリキャッシュの最大エントリ数"
 ],
 "Entry Cache Count:": [
  null,
  "エントリキャッシュの現在のエントリ数"
 ],
 "DN Cache": [
  null,
  "DNキャッシュ"
 ],
 "The DN cache hit ratio (dncachehitratio)": [
  null,
  "DNキャッシュのヒット率(dncachehitratio)"
 ],
 "Live DN Cache Statistics": [
  null,
  "リアルタイムのDNキャッシュの統計情報"
 ],
 "The amount of the cache that is being used: max size (maxdncachesize) vs current size (currentdncachesize)": [
  null,
  "キャッシュの使用量に関する情報です。最大サイズ(maxdncachesize)と現在のサイズ(currentdncachesize)が含まれます。"
 ],
 "Live DN Cache Utilization Statistics": [
  null,
  "キャッシュの利用状況に関する情報"
 ],
 "DN Cache Hit Ratio:": [
  null,
  "DNキャッシュのヒット率:"
 ],
 "DN Cache Max Size:": [
  null,
  "DNキャッシュの最大サイズ:"
 ],
 "DN Cache Hits:": [
  null,
  "DNキャッシュへのヒット数"
 ],
 "DN Cache Current Size:": [
  null,
  "DNキャッシュの現在のサイズ"
 ],
 "DN Cache Tries:": [
  null,
  "DNキャッシュへのアクセス試行回数"
 ],
 "DN Cache Max Count:": [
  null,
  "DNキャッシュの最大エントリ数"
 ],
 "DN Cache Current Count:": [
  null,
  "DNキャッシュの現在のエントリ数"
 ],
 "Task": [
  null,
  "タスク"
 ],
 "Created": [
  null,
  "作成した"
 ],
 "Status": [
  null,
  "ステータス"
 ],
 "No Tasks": [
  null,
  "タスクはありません"
 ],
 "Abort CleanAllRUV Tasks": [
  null,
  "CleanAllRUVタスクの中止"
 ],
 "CleanAllRUV Tasks": [
  null,
  "CleanAllRUVのタスク"
 ],
 "Agreement": [
  null,
  "合意"
 ],
 "Enabled": [
  null,
  "有効"
 ],
 "Session In Progress:": [
  null,
  "進行中のセッション:"
 ],
 "Changes Sent:": [
  null,
  "送信された変更:"
 ],
 "Last Init Started:": [
  null,
  "前回の初期化開始日時:"
 ],
 "Last Init Ended:": [
  null,
  "前回の初期化終了日時:"
 ],
 "Last Init Status:": [
  null,
  "前回の初期化ステータス:"
 ],
 "Last Updated Started:": [
  null,
  "前回の更新開始日時:"
 ],
 "Last Update Ended:": [
  null,
  "前回の更新完了日時:"
 ],
 "Last Update Status:": [
  null,
  "前回の更新ステータス:"
 ],
 "Awaken the winsync replication agreement": [
  null,
  "winsyncレプリカ合意を検証し再有効化します"
 ],
 "Poke": [
  null,
  "検証"
 ],
 "No Agreements": [
  null,
  "合意がありません"
 ],
 "Winsync Agreements": [
  null,
  "Winsync合意"
 ],
 "Changes Skipped:": [
  null,
  "スキップされた変更:"
 ],
 "Reap Active:": [
  null,
  "アクティブなレプリカ:"
 ],
 "Awaken the replication agreement": [
  null,
  "レプリカ合意を検証し再有効化します"
 ],
 "Replication Agreements": [
  null,
  "レプリカ合意"
 ],
 "Connection Opened": [
  null,
  "接続開始日時"
 ],
 "IP Address": [
  null,
  "IPアドレス"
 ],
 "Conn ID": [
  null,
  "接続ID"
 ],
 "Max Threads": [
  null,
  "最大スレッド"
 ],
 "If connection is currently at \"Max Threads\" then it will block new operations": [
  null,
  "'最大スレッド'に達している場合、新しい操作をブロックします。"
 ],
 "IP Address:": [
  null,
  "IPアドレス:"
 ],
 "File Descriptor:": [
  null,
  "ファイルディスクリプタ:"
 ],
 "Connection Opened:": [
  null,
  "接続の開始日時:"
 ],
 "Operations Started:": [
  null,
  "操作の開始日時:"
 ],
 "Connection ID:": [
  null,
  "接続ID:"
 ],
 "Operations Finished:": [
  null,
  "操作の完了日時:"
 ],
 "Bind DN:": [
  null,
  "バインドDN:"
 ],
 "Read/write Blocked:": [
  null,
  "ブロックされた読み込み/書き込み:"
 ],
 "Connection Currently At Max Threads:": [
  null,
  "現在の接続は最大スレッドになっているか:"
 ],
 "Yes": [
  null,
  "はい"
 ],
 "No": [
  null,
  "いいえ"
 ],
 "Number Of Times Connection Hit Max Threads:": [
  null,
  "接続が最大スレッドにヒットした回数:"
 ],
 "Number Of Operations Blocked By Max Threads:": [
  null,
  "最大スレッド数によってブロックされた操作の数:"
 ],
 "Client Connections:": [
  null,
  "クライアント接続:"
 ],
 "Search connections": [
  null,
  "接続を検索"
 ],
 "Glue Entry": [
  null,
  "グルーエントリ"
 ],
 "Description": [
  null,
  "説明"
 ],
 "No Glue Entries": [
  null,
  "グルーエントリがありません"
 ],
 "Replication Conflict Glue Entries": [
  null,
  "レプリケーションの競合するグルーエントリ"
 ],
 "Conflict DN": [
  null,
  "競合DN"
 ],
 "Resolve": [
  null,
  "解決"
 ],
 "No Conflict Entries": [
  null,
  "競合エントリはありません"
 ],
 "Replication Conflict Entries": [
  null,
  "レプリケーションの競合エントリ"
 ],
 "Disk Partition": [
  null,
  "ディスクパーティション"
 ],
 "Disk Size": [
  null,
  "ディスクサイズ"
 ],
 "Used Space": [
  null,
  "使用済みスペース"
 ],
 "Available Space": [
  null,
  "利用可能なスペース"
 ],
 "Connection Data": [
  null,
  "接続データ"
 ],
 "Edit Alias": [
  null,
  "エイリアスの編集"
 ],
 "Delete Alias": [
  null,
  "エイリアスの削除"
 ],
 "No Aliases": [
  null,
  "エイリアスはありません"
 ],
 "Instance Aliases": [
  null,
  "インスタンスエイリアス"
 ],
 "Replica Naming Aliases": [
  null,
  "レプリカの命名エイリアス"
 ],
 "Edit Connection": [
  null,
  "接続の編集"
 ],
 "Delete Connection": [
  null,
  "接続の削除"
 ],
 "No Credentials": [
  null,
  "資格情報がありません"
 ],
 "Credentials Table": [
  null,
  "資格情報テーブル"
 ],
 "Interactive Input is set": [
  null,
  "対話型入力が設定されています"
 ],
 "Both Password or Interactive Input flag are not set": [
  null,
  "パスワードまたは対話型入力フラグのいずれも設定されていません"
 ],
 "Replication Report Credentials": [
  null,
  "レプリケーションレポートの資格情報"
 ],
 "Supplier": [
  null,
  "サプライヤー"
 ],
 "Lag": [
  null,
  "遅延"
 ],
 "No agreement information": [
  null,
  "合意情報がありません"
 ],
 "Suffix & Replica ID:": [
  null,
  "サフィックス & レプリカID:"
 ],
 "Replica Server Status:": [
  null,
  "レプリカサーバステータス:"
 ],
 "Replication Enabled:": [
  null,
  "レプリケーションを有効:"
 ],
 "Consumer:": [
  null,
  "コンシューマー:"
 ],
 "Agreement is disabled": [
  null,
  "合意が無効になっています"
 ],
 "Agreement Name": [
  null,
  "合意名"
 ],
 "Name": [
  null,
  "名前"
 ],
 "There is no saved replication monitor connections": [
  null,
  "保存されたレプリケーション監視の接続はありません"
 ],
 "Replication Connections": [
  null,
  "レプリケーション接続"
 ],
 "There are no saved replication monitor aliases": [
  null,
  "保存されたレプリケーション監視のエイリアスはありません"
 ],
 "Replication Monitoring Aliases": [
  null,
  "レプリケーション監視のエイリアス"
 ],
 "Failed to get config nsslapd-port, nsslapd-localhost and nasslapd-rootdn: $0": [
  null,
  "nsslapd-port、nsslapd-localhost、nsslapd-rootdnの取得に失敗しました: $0"
 ],
 "Delete replica connection": [
  null,
  "レプリカ接続の削除"
 ],
 "Delete replica alias": [
  null,
  "レプリカエイリアスの削除"
 ],
 "The agreement doesn't exist!": [
  null,
  "合意が存在しません！"
 ],
 "Host, Port, and Bind DN are required.": [
  null,
  "Host、Port、Bind DNは必須です。"
 ],
 "Password field can't be empty, if Interactive Input is not selected": [
  null,
  "対話型入力が選択されていない場合、パスワードフィールドを空にすることはできません"
 ],
 "Credentials \"$0:$1\" already exists": [
  null,
  "認証情報 \"$0:$1\" はすでに存在します"
 ],
 "Host, Port, and Alias are required.": [
  null,
  "Host、Port、Aliasは必須です。"
 ],
 "Alias \"$0\" already exists": [
  null,
  "エイリアス$0は既に存在します"
 ],
 "Successfully saved monitor configuration to the .dsrc file": [
  null,
  "監視設定の.dsrcファイルへの保存に成功しました"
 ],
 "Failed to delete from .dsrc file: $0": [
  null,
  ".dsrc ファイルからの削除に失敗しました: $0"
 ],
 "Failed to add to .dsrc content: $0": [
  null,
  ".dsrc コンテンツへの追加に失敗しました: $0"
 ],
 "Failed to get .dsrc content: $0": [
  null,
  ".dsrc コンテンツの取得に失敗しました: $0"
 ],
 "Sync report has failed - $0": [
  null,
  "同期レポートに失敗しました - $0"
 ],
 "Bind DN and password are required.": [
  null,
  "バインドDNとパスワードが必要です。"
 ],
 "Failed to update .dsrc information: $0": [
  null,
  ".dsrc 情報の更新に失敗しました: $0"
 ],
 "Successfully added connection to .dsrc config file": [
  null,
  ".dsrc 設定ファイルへの接続の追加に成功しました"
 ],
 "Successfully added alias to .dsrc config file": [
  null,
  ".dsrc設定ファイルへのエイリアスの追加に成功しました"
 ],
 "Generate Report": [
  null,
  "レポートを作成"
 ],
 "Generating ...": [
  null,
  "作成しています ..."
 ],
 "Generating": [
  null,
  "作成する"
 ],
 "Saved Report Configuration": [
  null,
  "保存済みのレポート設定"
 ],
 "Replica Credentials": [
  null,
  "レプリカ認証情報"
 ],
 "Add a replica credential to the .dsrc file": [
  null,
  ".dsrc ファイルにレプリカの認証情報を追加する"
 ],
 "Add Connection": [
  null,
  "接続を追加"
 ],
 "Add a replica alias to the .dsrc file": [
  null,
  ".dsrc ファイルにレプリカのエイリアスを追加"
 ],
 "Add Alias": [
  null,
  "エイリアスを追加"
 ],
 "Use the specified credentials and display full topology report": [
  null,
  "指定された認証情報を使用し、完全なトポロジーレポートを表示します。"
 ],
 "Prepare New Report": [
  null,
  "新しいレポートを作成"
 ],
 "Hide Help": [
  null,
  "ヘルプを非表示"
 ],
 "Show Help": [
  null,
  "ヘルプを表示"
 ],
 "How To Use Replication Sync Report": [
  null,
  "レプリケーション同期レポートの使用方法"
 ],
 "Update The <b>Replication Credentials</b>": [
  null,
  "レプリケーション認証情報を更新する"
 ],
 "• Initially, the table is populated with the local instance's replication agreements, which includes the local instance.": [
  null,
  "初期状態では、このテーブルにはローカルインスタンスを含むレプリカ合意が設定されています。"
 ],
 "• Add the remaining replica server credentials from your replication topology.": [
  null,
  "レプリケーショントポロジーから、残りのレプリカサーバ認証情報を追加します。"
 ],
 "• It is advised to use an <b>Interactive Input</b> option for the password because it's more secure.": [
  null,
  "パスワードは対話型入力オプションを使用するのがより安全なのでお勧めです。"
 ],
 "Add <b>Replica Aliases</b> (if desired)": [
  null,
  "レプリカエイリアスを追加する(必要な場合)"
 ],
 "• Adding aliases will make the report more readable.": [
  null,
  "エイリアスを追加することで、レポートがより読みやすくなります。"
 ],
 "• Each Replica can have one alias. For example, you can give names like this: <b> Alias</b>=Main Supplier, <b>Hostname</b>=192.168.122.01, <b>Port</b>=38901": [
  null,
  "各レプリカは1つのエイリアスを持つことができます。例えば次のようなエイリアスを追加できます: this:Alias=Main Supplier, Hostname=192.168.122.01, Port=38901"
 ],
 "• In the report result, the report will have an entry like this: <b> Supplier: Main Supplier (192.168.122.01:38901)</b>.": [
  null,
  "レポート結果では、このような項目が表示されます: Supplier: Main Supplier (192.168.122.01:38901)"
 ],
 "Press <b>Generate Report</b> Button": [
  null,
  "レポート生成ボタンを押してください"
 ],
 "• It will initiate the report creation.": [
  null,
  "レポート作成が開始されます。"
 ],
 "• You may be asked for the credentials while the process is running through the agreements.": [
  null,
  "合意処理中に認証情報の入力を求められることがあります。"
 ],
 "Save the report configuration in the .dsrc file for future use.": [
  null,
  "後で使用できるよう、レポート設定を .dsrc ファイルに保存します。"
 ],
 "Save Report Configuration": [
  null,
  "レポート設定の保存"
 ],
 "Add Credentials": [
  null,
  "認証情報の追加"
 ],
 "Report Result": [
  null,
  "結果を確認"
 ],
 "Loading Replication DSRC Information ...": [
  null,
  "レプリカDSRC情報を読み込んでいます ..."
 ],
 "Only one monitor configuration can be saved in the server's '~/.dsrc' file.  There is already an existing monitor configuration, and if you proceed it will be completely overwritten with the new configuration.": [
  null,
  "サーバの '~/.dsrc' ファイルに保存できる監視設定は1つだけです。すでに既存の監視設定がありますが、このまま進むと新しい設定で完全に上書きされます。"
 ],
 "This will save the current credentials and aliases to the server's '~/.dsrc' file so it can be reused in the future.": [
  null,
  "これにより、現在の認証情報とエイリアスがサーバの '~/.dsrc' ファイルに保存され、将来的に再利用できるようになります。"
 ],
 "Synchronization Report": [
  null,
  "同期レポート"
 ],
 "Delete Replica Connection": [
  null,
  "レプリカ接続の削除"
 ],
 "Are you really sure you want to delete this connection from the '~/.dsrc' file?": [
  null,
  "本当に'~/.dsrc'ファイルからこの接続を削除してもいいのでしょうか？"
 ],
 "Deleting Connection ...": [
  null,
  "接続を削除しています ..."
 ],
 "Delete Replica Alias": [
  null,
  "レプリカエイリアスを削除"
 ],
 "Are you really sure you want to delete this alias from the '~/.dsrc' file?": [
  null,
  "本当にこのエイリアスを'~/.dsrc'ファイルから削除しますか？"
 ],
 "Deleting Alias ...": [
  null,
  "エイリアスを削除しています ..."
 ],
 "Are you sure you want to proceed?": [
  null,
  "本当に続けますか？"
 ],
 "Overwrite Monitor Configuration": [
  null,
  "監視設定を上書き"
 ],
 "Writing DSRC ...": [
  null,
  "DSRCを書き込み中 ..."
 ],
 "Write DSRC": [
  null,
  "DSRCを書き込み"
 ],
 "Reloading access log...": [
  null,
  "アクセスログを更新しています..."
 ],
 "Access Log": [
  null,
  "アクセスログ"
 ],
 "Refresh access log": [
  null,
  "アクセスログを更新する"
 ],
 "Reloading audit failure log...": [
  null,
  "監査失敗ログを更新しています..."
 ],
 "Audit Failure Log": [
  null,
  "監査失敗ログ"
 ],
 "Refresh audit failure log": [
  null,
  "監査失敗ログを更新する"
 ],
 "Reloading audit log...": [
  null,
  "監査ログを更新しています..."
 ],
 "Audit Log": [
  null,
  "監査ログ"
 ],
 "Refresh audit log": [
  null,
  "監査ログを更新する"
 ],
 "Reloading errors log...": [
  null,
  "エラーログを更新しています..."
 ],
 "Errors Log": [
  null,
  "エラーログ"
 ],
 "Refresh error log": [
  null,
  "エラーログを更新する"
 ],
 "Filter": [
  null,
  "フィルタ"
 ],
 "Everything": [
  null,
  "すべて"
 ],
 "Error Messages": [
  null,
  "エラーメッセージ"
 ],
 "Info Messages": [
  null,
  "情報メッセージ"
 ],
 ", I am sure.": [
  null,
  "、問題ありません。"
 ],
 "Warning! Account Policy config entry $0 doesn't exist!": [
  null,
  "警告! Account Policyの設定エントリ $0 が存在しません!"
 ],
 "Config entry $0 was successfully $1": [
  null,
  "構成エントリ $0 の $1 に成功しました"
 ],
 "Error during the config entry $0 operation - $1": [
  null,
  "構成エントリの$0に失敗しました - $1"
 ],
 "Config entry $0 was successfully deleted": [
  null,
  "構成エントリ $0 の削除に成功しました"
 ],
 "Error during the config entry removal operation - $0": [
  null,
  "構成エントリの削除中にエラーが発生しました - $0"
 ],
 "Successfully updated Account Policy Plugin": [
  null,
  "Account Policy Pluginを更新しました"
 ],
 "Error during update - $0": [
  null,
  "更新中にエラーが発生しました - $0"
 ],
 "Save Config": [
  null,
  "設定を保存"
 ],
 "Saving ...": [
  null,
  "保存しています ..."
 ],
 "Delete Config": [
  null,
  "設定の削除"
 ],
 "Saving": [
  null,
  "保存"
 ],
 "Adding ...": [
  null,
  "追加しています..."
 ],
 "Add Config": [
  null,
  "構成を追加"
 ],
 "Manage Account Policy Plugin Shared Config Entry": [
  null,
  "Account Policyプラグインの共有構成エントリを管理"
 ],
 "DN of the config entry": [
  null,
  "設定エントリのDN"
 ],
 "Config DN": [
  null,
  "設定DN"
 ],
 "Value must be a valid DN": [
  null,
  "値は有効なDNである必要があります"
 ],
 "Specifies the attribute to store the time of the last successful login in this attribute in the users directory entry (alwaysRecordLoginAttr)": [
  null,
  "ユーザディレクトリエントリのこの属性に、ログインに成功した最後の時刻を保存する属性を指定します (alwaysRecordLoginAttr)"
 ],
 "Always Record Login Attribute": [
  null,
  "常にログインを記録する属性"
 ],
 "Type an attribute name ...": [
  null,
  "属性名を入力してください ..."
 ],
 "There are no matching entries": [
  null,
  "該当するエントリはありません"
 ],
 "Sets that every entry records its last login time (alwaysRecordLogin)": [
  null,
  "すべてのエントリが最後のログイン時間を記録するように設定します。(alwaysRecordLogin)"
 ],
 "Always Record Login": [
  null,
  "常にログインを記録する"
 ],
 "Specifies the attribute to identify which entries are account policy configuration entries (specAttrName)": [
  null,
  "アカウントポリシー構成エントリを識別するために使用される属性を指定します。(specAttrName)"
 ],
 "Specific Attribute": [
  null,
  "特定の属性"
 ],
 "Specifies the primary time attribute used to evaluate an account policy (stateAttrName)": [
  null,
  "アカウントポリシーの評価に使用される主な時間属性を指定します。(stateAttrName)"
 ],
 "State Attribute": [
  null,
  "ステータス属性"
 ],
 "Provides a backup attribute for the server to reference to evaluate the expiration time (altStateAttrName)": [
  null,
  "有効期限を評価するためにサーバが参照するのバックアップ属性を提供します。(altStateAttrName)"
 ],
 "Alternative State Attribute": [
  null,
  "代替ステータス属性"
 ],
 "Specifies the attribute within the policy to use for the account inactivation limit (limitAttrName)": [
  null,
  "アカウントの無効化制限に使用するポリシー内の属性を指定します。(limitAttrName)"
 ],
 "Limit Attribute": [
  null,
  "制限属性"
 ],
 "DN of the shared config entry (nsslapd-pluginConfigArea)": [
  null,
  "共有構成エントリのDNを設定します。(nsslapd-pluginConfigArea)"
 ],
 "Shared Config Entry": [
  null,
  "共有構成エントリ"
 ],
 "Manage Config": [
  null,
  "構成を管理"
 ],
 "Create Config": [
  null,
  "構成を作成"
 ],
 "Delete Account Policy Config Entry": [
  null,
  "アカウントポリシーの構成エントリを削除"
 ],
 "Are you sure you want to delete this config entry?": [
  null,
  "このエントリを削除しますか？"
 ],
 "Deleting ...": [
  null,
  "削除しています ..."
 ],
 "Delete": [
  null,
  "削除"
 ],
 "The $0 operation was successfully done on \"$1\" entry": [
  null,
  "\"$1\" エントリに対する $0 の操作が成功しました"
 ],
 "The set operation was successfully done on \"$0:$1\" entry": [
  null,
  "\"$0:$1\" エントリへの設定が完了しました"
 ],
 "Error during the config entry set operation - $0": [
  null,
  "構成エントリの設定に失敗しました - $0"
 ],
 "Shared config entry $0 was successfully deleted": [
  null,
  "共有構成エントリ $0 を削除しました"
 ],
 "Error during the shared config entry removal operation - $0": [
  null,
  "共有構成エントリの削除に失敗しました - $0"
 ],
 "You can not use 'SSL' if the connection protocol is not 'TLS'": [
  null,
  "接続プロトコルが'TLS'でない場合、'SSL'は使用できません"
 ],
 "You can not use a 'SASL' method if the connection protocol is 'TLS'": [
  null,
  "接続プロトコルが'TLS'の場合、'SASL'メソッドは使用できません"
 ],
 "You must either set, or unset, both preferences": [
  null,
  "両方の設定を行うか、両方の設定を解除する必要があります"
 ],
 "You can not use the 'TLS' protocol if the BindMethod is a 'SASL' method": [
  null,
  "BindMethodが'SASL'メソッドの場合、'TLS'プロトコルは使用できません"
 ],
 "You can not use the 'LDAP' protocol if the BindMethod is a 'TLS'": [
  null,
  "BindMethodが'TLS'の場合、'LDAP'プロトコルは使用できません"
 ],
 "Prefix": [
  null,
  "プレフィックス"
 ],
 "Defines a prefix that can be prepended to the generated number values for the attribute (dnaPrefix)": [
  null,
  "属性の生成された数値の前に付加できるプレフィックスを定義します。(dnaPrefix)"
 ],
 "Remote Bind DN": [
  null,
  "リモートバインドDN"
 ],
 "Specifies the Replication Manager DN (dnaRemoteBindDN)": [
  null,
  "レプリケーションマネージャーのDNを指定します。(dnaRemoteBindDN)"
 ],
 "Remote Bind Credentials": [
  null,
  "リモートバインド資格情報"
 ],
 "Specifies the Replication Manager's password (dnaRemoteBindCred)": [
  null,
  "レプリケーションマネージャーのパスワードを指定します。(dnaRemoteBindCred)"
 ],
 "Next Range": [
  null,
  "次の範囲"
 ],
 "Defines the next range to use when the current range is exhausted.  Format is '####-####', or '1000-5000' (dnaNextRange).": [
  null,
  "現在の範囲が使い尽くされた場合に使用する次の範囲を定義します。形式は '####-####' または '1000-5000'です。(dnaNextRange)"
 ],
 "Save Shared Config": [
  null,
  "共有設定を保存"
 ],
 "Adding Config ...": [
  null,
  "設定を追加しています..."
 ],
 "Saving Config ...": [
  null,
  "設定を保存しています..."
 ],
 "Create DNA Config Entry": [
  null,
  "DNA設定エントリを作成する"
 ],
 "Edit DNA Config Entry": [
  null,
  "DNA設定エントリを編集する"
 ],
 "DNA Configuration": [
  null,
  "DNA設定"
 ],
 "Config Name": [
  null,
  "設定名"
 ],
 "Sets which attributes that will have unique numbers generated for them (dnaType).": [
  null,
  "一意の数字が生成される属性を設定します。(dnaType)"
 ],
 "DNA Managed Attributes": [
  null,
  "DNA管理属性"
 ],
 "Type an attribute...": [
  null,
  "属性名を入力してください..."
 ],
 "Sets an LDAP filter to use to search for and identify the entries to which to apply the distributed numeric assignment range (dnaFilter)": [
  null,
  "分散数値割り当て範囲を適用するエントリを検索および識別するために使用するLDAPフィルタを設定します。(dnaFilter)"
 ],
 "Sets the base DN to search for entries to which to apply the distributed numeric assignment (dnaScope)": [
  null,
  "分散数値割り当てを適用するエントリを検索するためのベースDNを設定します。(dnaScope)"
 ],
 "Subtree Scope": [
  null,
  "サブツリースコープ"
 ],
 "Gives the next available number which can be assigned (dnaNextValue)": [
  null,
  "割り当て可能な次の利用可能な番号を与えます。(dnaNextValue)"
 ],
 "Next Value": [
  null,
  "次の値へ"
 ],
 "Sets the maximum value that can be assigned for the range, default is -1 (dnaMaxValue)": [
  null,
  "範囲に割り当てることができる最大値を設定します。デフォルト値は-1です。(dnaMaxValue)"
 ],
 "Max Value": [
  null,
  "最大値"
 ],
 "Sets a user-defined value that instructs the plug-in to assign a new value for the entry (dnaMagicRegen)": [
  null,
  "プラグインに、エントリの新しい値を割り当てるよう指示するユーザ定義値を設定します。(dnaMagicRegen)"
 ],
 "Magic Regeneration Value": [
  null,
  "Magic Regeneration値"
 ],
 "Sets a threshold of remaining available numbers in the range. When the server hits the threshold, it sends a request for a new range (dnaThreshold)": [
  null,
  "範囲内に残っている利用可能な数値の閾値を設定します。サーバがこの閾値に達すると、新しい範囲の要求を送信します。(dnaThreshold)"
 ],
 "Threshold": [
  null,
  "閾値"
 ],
 "Sets a timeout period, in seconds, for range requests so that the server does not stall waiting on a new range from one server and can request a range from a new server (dnaRangeRequestTimeout)": [
  null,
  "範囲のリクエストのタイムアウト期間を秒単位で設定します。これにより、サーバが新しい範囲を1つのサーバから待ち続け、新しいサーバから範囲をリクエストできるようになります (dnaRangeRequestTimeout)"
 ],
 "Range Request Timeout": [
  null,
  "リクエストタイムアウト範囲"
 ],
 "Sets an interval, or increment, to add to the next DNA assigned number. If the next DNA number is '10', and the 'interval' is '1000', then the assigned value in the entry will be '1010' (dnaInterval)": [
  null,
  "次に割り当てられるDNA番号の間隔、つまり増分を設定します。次のDNA番号が '10' で、'数値間隔' が '1000' の場合、エントリに割り当てられる値は '1010' になります。(dnaInterval)"
 ],
 "Number Interval": [
  null,
  "数値間隔"
 ],
 "Shared Config Settings": [
  null,
  "共有設定"
 ],
 "The DNA Shared Config Entry is an entry in the database that DNA configurations about remote replicas are stored under.  Think of the Shared Config Entry as a container of DNA remote configurations.  These remote configurations are only used if the database is being replicated, otherwise these settings can be ignored. This entry must already be created prior to setting the <b> Shared Config Entry</b> field.": [
  null,
  "DNA共有構成エントリは、リモートレプリカに関するDNA構成が保存されるデータベース内のエントリです。共有構成エントリは、DNAリモート構成のコンテナと考えてください。これらのリモート構成は、データベースが複製されている場合にのみ使用されます。それ以外の場合、これらの設定は無視できます。このエントリは、共有構成エントリを設定する前に作成しておく必要があります。"
 ],
 "What is a Shared Config Entry?": [
  null,
  "共有構成エントリとは？"
 ],
 "Defines a container entry DN for DNA remote server configuration that the servers can use to transfer ranges between one another (dnaSharedCfgDN)": [
  null,
  "サーバがお互いに範囲を転送できるDNAリモートサーバ設定のコンテナエントリDNを定義します。(dnaSharedCfgDN)"
 ],
 "Shared Config Entry DN": [
  null,
  "共有構成エントリDN"
 ],
 "Manage DNA Plugin Shared Config Entry": [
  null,
  "DNAプラグイン共有構成エントリの管理"
 ],
 "Used with Replication to share ranges between replicas": [
  null,
  "レプリカ間で範囲を共有するためにレプリケーションと共に使用されます"
 ],
 "Remote Authentication Preferences": [
  null,
  "リモート認証設定"
 ],
 "Remote Bind Method": [
  null,
  "リモートバインド方法"
 ],
 "Loading DNA configuration": [
  null,
  "DNA設定を読み込んでいます"
 ],
 "Distributed Numeric Assignment": [
  null,
  "分散数値割り当て"
 ],
 "Delete DNA Configuration": [
  null,
  "DNA設定の削除"
 ],
 "Are you really sure you want to delete this DNA configuration?": [
  null,
  "このDNA設定を本当に削除しますか？"
 ],
 "Deleting config ...": [
  null,
  "設定を削除しています..."
 ],
 "Delete config": [
  null,
  "設定を削除"
 ],
 "Delete Shared Config Entry": [
  null,
  "共有設定エントリの削除"
 ],
 "Are you really sure you want to delete this Shared Config Entry?": [
  null,
  "この共有設定エントリを本当に削除しますか？"
 ],
 "Deleting Shared Config ...": [
  null,
  "共有設定を削除しています..."
 ],
 "Delete Shared Config": [
  null,
  "共有設定を削除"
 ],
 "$0 plugin was successfully $1.": [
  null,
  "$0プラグインを$1にしました。"
 ],
 "Please, restart the instance.": [
  null,
  "インスタンスを再起動してください。"
 ],
 "$0 plugin was successfully $1. Please, restart the instance.": [
  null,
  "$0プラグインを$1にしました。インスタンスを再起動してください。"
 ],
 "Error during $0 plugin modification - $1": [
  null,
  "$0プラグインの変更に失敗しました - $1"
 ],
 "Plugin is enabled": [
  null,
  "プラグインは有効です"
 ],
 "Plugin is disabled": [
  null,
  "プラグインは無効です"
 ],
 "Plugin": [
  null,
  "プラグイン"
 ],
 "Hide Plugin Details": [
  null,
  "プラグインの詳細を非表示"
 ],
 "Show Plugin Details": [
  null,
  "プラグインの詳細を表示"
 ],
 "Plugin Type": [
  null,
  "プラグインタイプ"
 ],
 "Plugin Path": [
  null,
  "プラグインパス"
 ],
 "Plugin Initfunc": [
  null,
  "プラグインInitfunc"
 ],
 "Plugin Depends On Type": [
  null,
  "プラグイン依存タイプ"
 ],
 "Plugin Depends On Named": [
  null,
  "プラグイン依存名"
 ],
 "Plugin Vendor": [
  null,
  "プラグインベンダー"
 ],
 "Plugin Version": [
  null,
  "プラグインバージョン"
 ],
 "Plugin Description": [
  null,
  "プラグイン説明"
 ],
 "Plugin ID": [
  null,
  "プラグインID"
 ],
 "Plugin Precedence": [
  null,
  "プラグイン優先順位"
 ],
 "Plugin Name": [
  null,
  "プラグイン名"
 ],
 "Plugin Description:": [
  null,
  "プラグインの説明:"
 ],
 "Plugin Path:": [
  null,
  "プラグインのパス:"
 ],
 "Plugin Init Function:": [
  null,
  "プラグインの初期化関数:"
 ],
 "Plugin ID:": [
  null,
  "プラグインID:"
 ],
 "Plugin Vendor:": [
  null,
  "プラグインベンダー:"
 ],
 "Plugin Version:": [
  null,
  "プラグインバージョン:"
 ],
 "Plugin Depends On Named:": [
  null,
  "プラグインが依存する名前付き要素:"
 ],
 "Plugin Depends On Type:": [
  null,
  "プラグインが依存する型:"
 ],
 "Plugin Precedence:": [
  null,
  "プラグインの優先度:"
 ],
 "Search Plugins": [
  null,
  "プラグインを検索"
 ],
 "Attribute": [
  null,
  "属性"
 ],
 "Edit Config": [
  null,
  "設定の編集"
 ],
 "Attribute Uniqueness Configurations": [
  null,
  "属性の一意性設定"
 ],
 "No Configurations": [
  null,
  "設定がありません"
 ],
 "Search Configurations": [
  null,
  "設定を検索"
 ],
 "Link Type": [
  null,
  "リンクタイプ"
 ],
 "Managed Type": [
  null,
  "管理タイプ"
 ],
 "Link Scope": [
  null,
  "リンクスコープ"
 ],
 "Linked Attributes Configurations": [
  null,
  "リンクした属性の設定"
 ],
 "Scope": [
  null,
  "スコープ"
 ],
 "DNA Configurations": [
  null,
  "DNA設定"
 ],
 "Remaining Values": [
  null,
  "残りの値"
 ],
 "DNA Shared Configurations": [
  null,
  "DNA共有設定"
 ],
 "Search Shared Configurations": [
  null,
  "共有設定の検索"
 ],
 "Definition Name": [
  null,
  "定義名"
 ],
 "Default Group": [
  null,
  "デフォルトグループ"
 ],
 "Automembership Definitions": [
  null,
  "自動メンバーシップ定義"
 ],
 "No Definitions": [
  null,
  "定義がありません"
 ],
 "Search Definitions": [
  null,
  "定義を検索"
 ],
 "Exclusive Regex": [
  null,
  "除外正規表現"
 ],
 "Inclusive Regex": [
  null,
  "包含正規表現"
 ],
 "Target Group": [
  null,
  "対象グループ"
 ],
 "Automembership Regular Expressions": [
  null,
  "自動メンバーシップ正規表現"
 ],
 "No regular expressions": [
  null,
  "正規表現がありません"
 ],
 "Search Regular Expressions": [
  null,
  "正規表現を検索"
 ],
 "Managed Base": [
  null,
  "管理ベース"
 ],
 "Managed Entry Definitions": [
  null,
  "Managed Entryの定義"
 ],
 "No definitions": [
  null,
  "定義がありません"
 ],
 "Template DN": [
  null,
  "テンプレートDN"
 ],
 "Managed Entry Templates": [
  null,
  "Managed Entryテンプレート"
 ],
 "No templates": [
  null,
  "テンプレートがありません"
 ],
 "Search Templates": [
  null,
  "テンプレートを検索"
 ],
 "URL": [
  null,
  "URL"
 ],
 "Edit URL": [
  null,
  "URLを編集"
 ],
 "Delete URL": [
  null,
  "URLを削除"
 ],
 "Pass-Through Authentication URLs": [
  null,
  "パススルー認証URL"
 ],
 "No URLs": [
  null,
  "URLがありません"
 ],
 "Search": [
  null,
  "検索"
 ],
 "Map Method": [
  null,
  "マップ方法"
 ],
 "PAM Configurations": [
  null,
  "PAM設定"
 ],
 "No PAM configurations": [
  null,
  "PAM設定がありません"
 ],
 "Successfully updated Referential Integrity Plugin": [
  null,
  "参照整合プラグインの更新に成功しました"
 ],
 "Manage Referential Integrity Plugin Shared Config Entry": [
  null,
  "参照整合プラグインの共有構成エントリを管理"
 ],
 "The config entry full DN": [
  null,
  "構成エントリの完全なDN"
 ],
 "Specifies attributes to check for and update (referint-membership-attr)": [
  null,
  "チェックおよび更新する属性を指定します。 (referint-membership-attr)"
 ],
 "Membership Attribute": [
  null,
  "メンバーシップ属性"
 ],
 "At least one attribute must be specified": [
  null,
  "少なくとも1つの属性を指定する必要があります"
 ],
 "Defines the subtree in which the plug-in looks for the delete or rename operations of a user entry (nsslapd-pluginEntryScope)": [
  null,
  "ユーザエントリの削除または名前変更操作をプラグインが検索するサブツリーを定義します。 (nsslapd-pluginEntryScope)"
 ],
 "Entry Scope": [
  null,
  "エントリスコープ"
 ],
 "Defines the subtree in which the plug-in ignores any operations for deleting or renaming a user (nsslapd-pluginExcludeEntryScope)": [
  null,
  "ユーザの削除または名前変更操作をプラグインが無視するサブツリーを定義します。 (nsslapd-pluginExcludeEntryScope)"
 ],
 "Exclude Entry Scope": [
  null,
  "エントリスコープを除外する"
 ],
 "Specifies which branch the plug-in searches for the groups to which the user belongs. It only updates groups that are under the specified container branch, and leaves all other groups not updated (nsslapd-pluginContainerScope).": [
  null,
  "ユーザが所属するグループをプラグインが検索するブランチを指定します。指定されたコンテナブランチに属するグループのみが更新され、その他のグループは更新されません。 (nsslapd-pluginContainerScope)"
 ],
 "Container Scope": [
  null,
  "コンテナの範囲"
 ],
 "Logfile": [
  null,
  "ログファイル"
 ],
 "Invalid log file name": [
  null,
  "無効なログファイル名です"
 ],
 "Sets the update interval in seconds. Special values: 0 - The check is performed immediately, -1 - No check is performed. (referint-update-delay)": [
  null,
  "更新間隔を秒単位で設定します。ただし次の値は特別な値です。 0 : 即座にチェックが実行されます。 -1 : チェックが実行されません。 (referint-update-delay)"
 ],
 "Update Delay": [
  null,
  "更新の遅延"
 ],
 "Specifies attributes to check for and update (referint-membership-attr).": [
  null,
  "確認および更新する属性を指定します。(referint-membership-attr)"
 ],
 "At least one attribute needs to be specified": [
  null,
  "少なくとも1つの属性を指定する必要があります"
 ],
 "The value must be a valid DN": [
  null,
  "値は有効なDNである必要があります"
 ],
 "Specifies a path to the Referential integrity logfile. For example: /var/log/dirsrv/slapd-": [
  null,
  "参照整合性ログファイルへのパスを指定します。例:/var/log/dirsrv/slapd-"
 ],
 "Invalid log nameN": [
  null,
  "無効なログ名"
 ],
 "Sets the update interval. Special values: 0 - The check is performed immediately, -1 - No check is performed (referint-update-delay)": [
  null,
  "更新間隔を設定します。ただし次の値は特別な値です。 0 : 即座にチェックが実行されます。 -1 : チェックが実行されません。 (referint-update-delay)"
 ],
 "The value to set as nsslapd-pluginConfigArea": [
  null,
  "nsslapd-pluginConfigAreaの値を設定します"
 ],
 "Manage": [
  null,
  "管理"
 ],
 "Delete RI Plugin Config Entry": [
  null,
  "RIプラグインの設定エントリを削除"
 ],
 "Successfully updated the Retro Changelog": [
  null,
  "Retro Changelogを更新しました"
 ],
 "Failed to update Retro Changelog Plugin - $0": [
  null,
  "Retro Changelog Pluginの更新に失敗しました - $0"
 ],
 "This attribute specifies the suffix which will be excluded from the scope of the plugin (nsslapd-exclude-suffix)": [
  null,
  "プラグインの範囲から除外されるサフィックスを指定します。 (nsslapd-exclude-suffix)"
 ],
 "Exclude Suffix": [
  null,
  "除外するサフィックス"
 ],
 "Type a suffix...": [
  null,
  "サフィックスを入力してください..."
 ],
 "Values must be valid DN !": [
  null,
  "値は有効なDNである必要があります！"
 ],
 "Sets a flag to indicate on a change in the changelog whether the change is newly made on that server or whether it was replicated over from another server (isReplicated)": [
  null,
  "変更履歴の変更について、そのサーバで新しく変更されたものか、他のサーバからレプリケートされたものであるかを示すフラグを設定します (isReplicated)"
 ],
 "Is Replicated": [
  null,
  "複製する"
 ],
 "This specifies attribute which will be excluded from the scope of the plugin (nsslapd-exclude-attrs)": [
  null,
  "プラグインの範囲から除外される属性を指定します。 (nsslapd-exclude-attrs)"
 ],
 "Exclude attribute": [
  null,
  "除外する属性"
 ],
 "Specifies the maximum age of any entry in the changelog before it is trimmed from the database (nsslapd-changelogmaxage)": [
  null,
  "変更履歴のエントリがデータベースから削除されるまでの最大保存期間を指定します。 (nsslapd-changelogmaxage)"
 ],
 "Max Age": [
  null,
  "最大保持時間"
 ],
 "Weeks": [
  null,
  "週"
 ],
 "Days": [
  null,
  "日"
 ],
 "Hours": [
  null,
  "時間"
 ],
 "Minutes": [
  null,
  "分"
 ],
 "Sets in seconds how often the plugin checks if changes can be trimmed from the database based on the entry's max age (nsslapd-changelog-trim-interval)": [
  null,
  "エントリの最大保存期間に基づいて、データベースから変更を削除できるかどうかをプラグインが確認する頻度を秒単位で設定します。(nsslapd-changelog-trim-interval)"
 ],
 "Trimming Interval": [
  null,
  "トリミング間隔"
 ],
 "Successfully updated the RootDN Access Control": [
  null,
  "RootDNアクセス制御の更新に成功しました"
 ],
 "Failed to update RootDN Access Control Plugin - $0": [
  null,
  "RootDNアクセス制御プラグインの更新に失敗しました - $0"
 ],
 "Sets what hosts, by fully-qualified domain name, the root user is allowed to use to access the Directory Server. Wildcards are accepted. Any hosts not listed are implicitly denied (rootdn-allow-host)": [
  null,
  "ルートユーザがディレクトリサーバへのアクセスに使用できるホストを、完全修飾ドメイン名で設定します。ワイルドカードが使用できます。設定されていないホストは暗黙的に拒否されます (rootdn-allow-host)"
 ],
 "Allow Host": [
  null,
  "許可ホスト"
 ],
 "Type a hostname ...": [
  null,
  "ホスト名を入力してください..."
 ],
 "Sets what hosts, by fully-qualified domain name, the root user is not allowed to use to access the Directory Server.  Wildcards are accepted.  Any hosts not listed are implicitly allowed (rootdn-deny-host). If a host address is listed in both the rootdn-allow-host and rootdn-deny-host attributes, it is denied access.": [
  null,
  "ルートユーザがディレクトリサーバへのアクセスに使用できないホストを、完全修飾ドメイン名で設定します。ワイルドカードが使用できます。設定されていないホストは暗黙的に許可されます (rootdn-deny-host)。ホストアドレスが、rootdn-allow-host属性とrootdn-deny-host属性の両方に設定されている場合、アクセスが拒否されます。"
 ],
 "Deny Host": [
  null,
  "拒否ホスト"
 ],
 "Sets what IP addresses, either IPv4 or IPv6, for machines the root user is allowed to use to access the Directory Server. Wildcards are accepted.  Any IP addresses not listed are implicitly denied (rootdn-allow-ip)": [
  null,
  "ルートユーザがディレクトリサーバへのアクセスに使用できるIPアドレス(IPv4 または IPv6)を設定します。ワイルドカードが使用できます。設定されていないIPアドレスは暗黙的に拒否されます (rootdn-allow-ip)"
 ],
 "Allow IP address": [
  null,
  "許可IPアドレス"
 ],
 "Type an IP address ...": [
  null,
  "IPアドレスを入力してください..."
 ],
 "Sets what IP addresses, either IPv4 or IPv6, for machines the root user is not allowed to use to access the Directory Server. Wildcards are accepted. Any IP addresses not listed are implicitly allowed (rootdn-deny-ip) If an IP address is listed in both the rootdn-allow-ip and rootdn-deny-ip attributes, it is denied access.": [
  null,
  "ルートユーザがディレクトリサーバへのアクセスに使用できないIPアドレス(IPv4 または IPv6)を設定します。ワイルドカードが使用できます。設定されていないIPアドレスは暗黙的に許可されます。(rootdn-deny-ip)IPアドレスが、rootdn-allow-ip属性とrootdn-deny-ip属性の両方に設定されている場合、アクセスは拒否されます。"
 ],
 "Deny IP address": [
  null,
  "拒否IPアドレス"
 ],
 "Sets part of a time period or range when the root user is allowed to access the Directory Server. This sets when the time-based access begins (rootdn-open-time)": [
  null,
  "ルートユーザがディレクトリサーバにアクセスできる期間または範囲の一部を設定します。これにより、時間ベースのアクセスが開始されるタイミングを設定します。(rootdn-open-time)。"
 ],
 "Open Time": [
  null,
  "開始時間"
 ],
 "Sets part of a time period or range when the root user is allowed to access the Directory Server. This sets when the time-based access ends (rootdn-close-time)": [
  null,
  "ルートユーザがディレクトリサーバにアクセスできる期間または範囲の一部を設定します。これにより、時間ベースのアクセスが終了するタイミングを設定します。(rootdn-close-time)"
 ],
 "Close Time": [
  null,
  "終了時間"
 ],
 "Gives a comma-separated list of what days the root user is allowed to use to access the Directory Server. Any days listed are implicitly denied (rootdn-days-allowed)": [
  null,
  "ルートユーザがディレクトリサーバにアクセスできる日を、指定します。設定された日は暗黙的に許可されます (rootdn-days-allowed)"
 ],
 "Days To Allow Access": [
  null,
  "アクセス許可日"
 ],
 "Monday": [
  null,
  "月曜日"
 ],
 "Friday": [
  null,
  "金曜日"
 ],
 "Tuesday": [
  null,
  "火曜日"
 ],
 "Saturday": [
  null,
  "土曜日"
 ],
 "Wednesday": [
  null,
  "水曜日"
 ],
 "Sunday": [
  null,
  "日曜日"
 ],
 "Thursday": [
  null,
  "木曜日"
 ],
 "You must set at least one day": [
  null,
  "少なくとも1日以上設定する必要があります"
 ],
 "Global USN mode was successfully set to $0.": [
  null,
  "グローバルUSNモードが$0に設定されました。"
 ],
 "Error during global USN mode modification - $0": [
  null,
  "グローバルUSNモードの変更に失敗しました - $0"
 ],
 "Suffix is required.": [
  null,
  "サフィックスが必要です。"
 ],
 "Cleanup USN Tombstones task was successful": [
  null,
  "USN Tombstonesのクリーンアップタスクが正常に終了しました"
 ],
 "Cleanup USN Tombstones task has failed $0": [
  null,
  "USN Tombstonesのクリーンアップタスクが失敗しました $0"
 ],
 "USN Tombstone Cleanup Task": [
  null,
  "USN Tombstonesのクリーンアップタスク"
 ],
 "Run": [
  null,
  "実行"
 ],
 "Gives the suffix in the Directory Server to run the cleanup operation against": [
  null,
  "クリーンアップ操作を実行するディレクトリサーバ内のサフィックスを指定します。"
 ],
 "Cleanup Suffix": [
  null,
  "クリーンアップサフィックス"
 ],
 "Gives the highest USN value to delete when removing tombstone entries. All tombstone entries up to and including that number are deleted. Tombstone entries with higher USN values (that means more recent entries) are not deleted": [
  null,
  "tombstoneエントリを削除する時の最大USN値を指定します。指定した番号までのすべてのtombstoneエントリが削除されます。より新しいUSN値を持つtombstoneエントリは削除されません。"
 ],
 "Cleanup Max USN": [
  null,
  "クリーンアップ最大USN値"
 ],
 "The USN Plug-in enables LDAP clients and servers to identify if entries have been changed. When the USN Plug-in is enabled, update sequence numbers (USNs) are sequential numbers that are assigned to an entry whenever a write operation is performed against the entry. (Write operations include add, modify, modrdn, and delete operations. Internal database operations, like export operations, are not counted in the update sequence.) A USN counter keeps track of the most recently assigned USN.": [
  null,
  "USNプラグインを有効にすると、LDAPクライアントとサーバはエントリが変更されたかどうかを識別できます。USNプラグインが有効の場合、エントリに対して書き込み操作が行われるたびに、更新シーケンス番号(USN)が割り当てられます。(書き込み操作には、add、modify、modrdn、delete操作が含まれます。エクスポート操作などの内部データベース操作は、更新シーケンスにカウントされません)USNカウンターは、最後に割り当てられたUSNを追跡します。"
 ],
 "The USN plug-in also moves entries to tombstone entries when the entry is deleted. If replication is enabled, then separate tombstone entries are kept by both the USN and Replication plug-in. Note that both tombstone entries are deleted by the replication process.": [
  null,
  "エントリが削除されると、USNプラグインはエントリをtombstoneエントリに移動します。レプリケーションが有効になっている場合、USNプラグインとレプリケーションプラグインの両方によって別々のtombstoneエントリが保持されます。レプリケーションプロセスによって、両方のtombstoneエントリが削除されることに注意してください。"
 ],
 "What is the USN Plugin?": [
  null,
  "USNプラグインとは？"
 ],
 "Enable entryUSN assignment across all backends, instead of per backend.": [
  null,
  "バックエンド毎ではなく、すべてのバックエンドでentryUSNの割り当てを有効にします。"
 ],
 "USN Global": [
  null,
  "USNグローバル"
 ],
 "This task deletes the tombstone entries maintained by the USN plugin.  This task should not run against a suffix that is being replicated": [
  null,
  "このタスクは、USNプラグインによって保持されているtombstoneエントリを削除します。このタスクは、レプリケーションされているサフィックスに対して実行しないでください。"
 ],
 "Tombstone Cleanup Task": [
  null,
  "Tombstoneクリーンアップタスク"
 ],
 "Run Task": [
  null,
  "タスクを実行"
 ],
 "Fixup task for $0 was successfully started": [
  null,
  "$0 の修正タスクを開始しました"
 ],
 "Fixup task for $0 has failed $1": [
  null,
  "$0 の修正タスクが失敗しました $1"
 ],
 "Successfully updated Posix winsync plugin": [
  null,
  "Posix winsyncプラグインを更新しました"
 ],
 "Failed to update Posix winsync plugin - $0": [
  null,
  "Posix winsyncプラグインの更新に失敗しました - $0"
 ],
 "Task running ...": [
  null,
  "タスクを実行しています..."
 ],
 "MemberOf Task": [
  null,
  "MemberOfタスク"
 ],
 "Base DN that contains entries to fix up": [
  null,
  "修復対象のエントリを含むベースDN"
 ],
 "Base DN": [
  null,
  "ベースDN"
 ],
 "Filter for entries to fix up. If omitted, all entries with objectclass inetuser/inetadmin/nsmemberof under the specified base will have their memberOf attribute regenerated.": [
  null,
  "修復対象のエントリをフィルタリングします。省略した場合、指定されたベースの下に、オブジェクトクラス(inetuser/inetadmin/nsmemberof)を持つエントリのmemberOf属性が再生成されます。"
 ],
 "Filter DN": [
  null,
  "フィルタDN"
 ],
 "Enter an LDAP search filter": [
  null,
  "フィルタでエントリ検索"
 ],
 "Sets whether to run the memberOf fix-up task immediately after a sync run in order to update group memberships for synced users": [
  null,
  "同期実行後すぐにmemberOf修正タスクを実行して、同期されたユーザのグループメンバーシップを更新するかどうかを設定します"
 ],
 "Create MemberOf Task": [
  null,
  "MemberOfタスクを作成する"
 ],
 "Sets whether to store (and, if necessary, convert) the UID value in the memberUID attribute in lower case": [
  null,
  "UID値を小文字でmemberUID属性に格納 (必要に応じて変換)するかどうかを設定します"
 ],
 "Lower Case UID": [
  null,
  "UIDを小文字にする"
 ],
 "Sets whether to map the memberUID attribute in an Active Directory group to the uniqueMember attribute in a Directory Server group": [
  null,
  "Active DirectoryグループのmemberUID属性を、ディレクトリサーバグループのuniqueMember属性にマップするかどうかを設定します"
 ],
 "Map Member UID": [
  null,
  "メンバーUIDをマップ"
 ],
 "Manages if nested groups are updated when memberUID attributes in an Active Directory POSIX group change": [
  null,
  "Active Directory POSIXグループのmemberUID属性が変更されたときに、ネストされたグループが更新されるかどうかを管理します"
 ],
 "Map Nested Grouping": [
  null,
  "ネストされたグループをマップする"
 ],
 "Sets whether to the older Microsoft System Services  for Unix 3.0 (msSFU30) schema when syncing Posix attributes from Active Directory": [
  null,
  "Active DirectoryからPOSIX属性を同期する際に、古いMicrosoft System Services for Unix 3.0 (msSFU30)スキーマを使用するかどうかを設定します"
 ],
 "Microsoft System Services for Unix 3.0 (msSFU30) schema": [
  null,
  "Microsoft System Services for Unix 3.0 (msSFU30)スキーマを使用する"
 ],
 "Corrects mismatched member and uniquemember values": [
  null,
  "一致しないメンバーとuniqueMemberの値を修正"
 ],
 "Run MemberOf Task": [
  null,
  "MemberOfタスクを実行"
 ],
 "Fixup DN is required.": [
  null,
  "修正DNは必須です。"
 ],
 "Fixup task for $0 was successful": [
  null,
  "$0 の修正タスクが成功しました"
 ],
 "Config Attribute and Group Attribute are required.": [
  null,
  "構成属性とグループ属性は必須です。"
 ],
 "Successfully updated MemberOf Plugin": [
  null,
  "MemberOfプラグインを更新しました"
 ],
 "MemberOf Plugin FixupTask": [
  null,
  "MemberOfプラグインの修正タスク"
 ],
 "This task only needs to be run after enabling the plugin for the first time, or if the plugin configuration has changed in a way that will impact the group memberships.": [
  null,
  "このタスクは、プラグインを初めてに有効にした後、またはプラグインの設定がグループメンバーシップに影響を与えるような変更がされた場合にのみ実行する必要があります。"
 ],
 "Base DN that contains entries to fix up.": [
  null,
  "修正対象のエントリを含むベースDN。"
 ],
 "Subtree DN": [
  null,
  "サブツリーDN"
 ],
 "Optional. Filter for finding entries to fix up. For example:  (uid=*).  If omitted, all entries with objectclass 'inetuser', 'inetadmin', or 'nsmemberof' under the specified subtree DN will have their memberOf attribute regenerated.": [
  null,
  "これはオプションです。修正対象のエントリを検索するためのフィルタを指定します。 例:(uid=*) 省略した場合、指定されたサブツリーDN配下にあるオブジェクトクラス 'inetuser'、'inetadmin'、または 'nsmemberof' を持つすべてのエントリのmemberOf属性が再生成されます。"
 ],
 "Search Filter": [
  null,
  "検索フィルタ"
 ],
 "Manage MemberOf Plugin Shared Config Entry": [
  null,
  "MemberOfプラグイン共有構成エントリの管理"
 ],
 "Specifies the attribute in the user entry for the Directory Server to manage to reflect group membership (memberOfAttr)": [
  null,
  "グループメンバーシップを反映するためにディレクトリサーバが管理するユーザエントリの属性を指定します。(memberOfAttr)"
 ],
 "Type a member attribute...": [
  null,
  "メンバー属性を入力してください..."
 ],
 "Specifies the attribute in the group entry to use to identify the DNs of group members (memberOfGroupAttr)": [
  null,
  "グループメンバーのDNを識別するために使用するグループエントリの属性を指定します。(memberOfGroupAttr)"
 ],
 "Group Attribute": [
  null,
  "グループ属性"
 ],
 "Type a member group attribute...": [
  null,
  "メンバーグループ属性を入力してください..."
 ],
 "Specifies backends or multiple-nested suffixes for the MemberOf plug-in to work on (memberOfEntryScope)": [
  null,
  "MemberOfプラグインが動作するバックエンドまたはネストされたサフィックスを指定します。(memberOfEntryScope)"
 ],
 "Type a subtree DN...": [
  null,
  "サブツリーDNを入力してください..."
 ],
 "Values must be valid DN's": [
  null,
  "値は有効なDNである必要があります"
 ],
 "Specifies whether to search the local suffix for user entries on all available suffixes (memberOfAllBackends)": [
  null,
  "すべての利用可能なサフィックスでユーザエントリを検索するかどうかを指定します。(memberOfAllBackends)"
 ],
 "All Backends": [
  null,
  "すべてのバックエンド"
 ],
 "Specifies backends or multiple-nested suffixes for the MemberOf plug-in to exclude (memberOfEntryScopeExcludeSubtree)": [
  null,
  "MemberOfプラグインが除外するバックエンドまたは複数のネストされたサフィックスを指定します (memberOfEntryScopeExcludeSubtree)"
 ],
 "Exclude Subtree": [
  null,
  "サブツリーを除外"
 ],
 "Specifies wherher to skip nested groups or not (memberOfSkipNested)": [
  null,
  "ネストされたグループをスキップするかどうかを指定します (memberOfSkipNested)"
 ],
 "Skip Nested Groups": [
  null,
  "ネストをスキップ"
 ],
 "If an entry does not have an object class that allows the memberOf attribute then the memberOf plugin will automatically add the object class listed in the memberOfAutoAddOC parameter": [
  null,
  "エントリにmemberOf属性を許可するオブジェクトクラスがない場合、memberOfプラグインはmemberOfAutoAddOCパラメータに設定されているオブジェクトクラスを自動的に追加します"
 ],
 "Auto Add OC": [
  null,
  "自動的にOCを追加"
 ],
 "A subtree is required, and values must be valid DN's": [
  null,
  "サブツリーが必要であり、値は有効なDNである必要があります"
 ],
 "The fixup task will add the memberOf attribute to entries that are missing it.  This is typically only run once after enabling or changing the plugin.": [
  null,
  "修正タスクはmemberOf属性が欠落しているエントリに追加します。これは通常、プラグインを有効にしたり変更した後に1回だけ実行されます。"
 ],
 "MemberOf Fixup Task": [
  null,
  "MemberOf修正タスク"
 ],
 "Delete MemberOf Config Entry": [
  null,
  "MemberOf設定エントリを削除"
 ],
 "There must be at least one Subtree or Subtree Entries OC set": [
  null,
  "少なくとも1つのSubtreeまたはSubtree Entries OCを設定する必要があります"
 ],
 "$0 Attribute Uniqueness Plugin Config Entry": [
  null,
  "Attribute Uniquenessプラグインの設定エントリを$0"
 ],
 " Config": [
  null,
  "設定 "
 ],
 "Sets the name of the plug-in configuration record. (cn) You can use any string, but \"attribute_name Attribute Uniqueness\" is recommended.": [
  null,
  "プラグイン設定レコードの名前を設定します。(cn) 任意の文字列を使用できますが、\"attribute_name Attribute Uniqueness\" を使用することを推奨します。"
 ],
 "Sets the name of the attribute whose values must be unique. This attribute is multi-valued. (uniqueness-attribute-name)": [
  null,
  "値が一意でなければならない属性の名前を設定します。この属性は複数の値を持ちます。 (uniqueness-attribute-name)"
 ],
 "Attribute Names": [
  null,
  "属性名"
 ],
 "Type an attribute name...": [
  null,
  "属性名を入力してください..."
 ],
 "There are no matching attributes": [
  null,
  "一致する属性がありません"
 ],
 "Sets the DN under which the plug-in checks for uniqueness of the attributes value. This attribute is multi-valued (uniqueness-subtrees)": [
  null,
  "プラグインが属性値の一意性を確認する際のDNを設定します。この属性は複数の値を持ちます。 (uniqueness-subtrees)"
 ],
 "Subtrees": [
  null,
  "サブツリー"
 ],
 "Verifies that the value of the attribute set in uniqueness-attribute-name is unique in this subtree (uniqueness-top-entry-oc)": [
  null,
  "属性名(uniqueness-attribute-name)で設定された値が、このサブツリー内で一意であることを検証します (uniqueness-top-entry-oc)"
 ],
 "Top Entry OC": [
  null,
  "トップエントリのOC"
 ],
 "If enabled (on), the plug-in checks that the attribute is unique across all subtrees set. If you set the attribute to off, uniqueness is only enforced within the subtree of the updated entry (uniqueness-across-all-subtrees)": [
  null,
  "有効になっている場合、プラグインはすべての設定されたサブツリー全体で属性が一意であることを確認します。属性をオフに設定すると、一意性は更新されたエントリのサブツリー内でのみ強制されます(uniqueness-across-all-subtrees)"
 ],
 "Across All Subtrees": [
  null,
  "すべてのサブツリー"
 ],
 "Verifies if an attribute is unique, if the entry contains the objectclass set in this parameter (uniqueness-subtree-entries-oc)": [
  null,
  "属性が一意かどうかを検証します。エントリには、このパラメータで設定されたオブジェクトクラスが含まれている必要があります (uniqueness-subtree-entries-oc)"
 ],
 "Subtree Entry's OC": [
  null,
  "サブツリーエントリのOC"
 ],
 "Identifies whether or not the config is enabled.": [
  null,
  "設定が有効かどうかを識別します。"
 ],
 "Enable config": [
  null,
  "設定を有効にする"
 ],
 "Configuration is enabled": [
  null,
  "設定が有効になっています"
 ],
 "Configuration is disabled": [
  null,
  "設定が無効になっています"
 ],
 "Delete Attribute Uniqueness Configuration": [
  null,
  "属性一意性の設定を削除"
 ],
 "Are you sure you want to delete this configuration?": [
  null,
  "この設定を削除しますか？"
 ],
 "Deleting attribute uniqueness configuration...": [
  null,
  "属性一意性の設定を削除しています..."
 ],
 "Delete Configuration": [
  null,
  "設定を削除"
 ],
 "Name, Scope, Filter and Grouping Attribute are required.": [
  null,
  "名前、スコープ、フィルタ、およびグループ属性は必須です。"
 ],
 "Error during the definition entry $0 operation - $1": [
  null,
  "定義エントリの $0 に失敗しました - $1"
 ],
 "Error during the regex \"$0\" entry delete operation - $1": [
  null,
  "正規表現 \"$0\" エントリの削除に失敗しました - $1"
 ],
 "Error during the regex \"$0\" entry $1 operation - $2": [
  null,
  "正規表現 \"$0\" エントリの $1 に失敗しました - $2"
 ],
 "Definition entry $0 was successfully deleted": [
  null,
  "定義エントリ $0 を削除しました"
 ],
 "Error during the definition entry removal operation - $0": [
  null,
  "定義エントリの削除中にエラーが発生しました - $0"
 ],
 "Name and Target Group are required.": [
  null,
  "名前と対象グループは必須です。"
 ],
 "Regex \"$0\" already exists": [
  null,
  "正規表現 \"$0\" はすでに存在します"
 ],
 "Regex \"$0\" does not exist - \"$1\" is impossible": [
  null,
  "正規表現 \"$0\" は存在しません -  \"$1\" はできません"
 ],
 "Regex \"$0\" does not exist - impossible to delete": [
  null,
  "正規表現 \"$0\" は存在しません - 削除できません"
 ],
 "$0  Auto Membership Plugin Definition Entry": [
  null,
  "Auto Membershipプラグイン定義エントリを$0"
 ],
 "Add Definition": [
  null,
  "定義を追加"
 ],
 "Save Definition": [
  null,
  "定義を保存"
 ],
 "Sets the subtree DN to search for entries (autoMemberScope).": [
  null,
  "エントリを検索するサブツリーDNを設定します。(autoMemberScope)"
 ],
 "Sets a standard LDAP search filter to use to search for matching entries (autoMemberFilter)": [
  null,
  "一致するエントリの検索に使用する標準のLDAP検索フィルタを設定します。(autoMemberFilter)"
 ],
 "Entry Filter": [
  null,
  "エントリフィルタ"
 ],
 "Specifies the name of the member attribute in the group entry and the attribute in the object entry that supplies the member attribute value, in the format group_member_attr:entry_attr (autoMemberGroupingAttr)": [
  null,
  "グループエントリのメンバー属性名と、メンバー属性値を提供するオブジェクトエントリの属性名を、group_member_attr:entry_attrの形式で指定します。(autoMemberGroupingAttr)"
 ],
 "Grouping Attribute": [
  null,
  "グループ属性"
 ],
 "Sets an optional default or fallback group to add the entry to as a member (autoMemberDefaultGroup)": [
  null,
  "エントリをメンバーとして追加するオプションのデフォルトまたはフォールバックグループを設定します。(autoMemberDefaultGroup)"
 ],
 "Membership Regular Expressions": [
  null,
  "メンバーシップ正規表現"
 ],
 "Add Regex": [
  null,
  "正規表現を追加"
 ],
 "Manage Auto Membership Plugin Regex Entry": [
  null,
  "自動メンバーシッププラグイン正規表現エントリを管理"
 ],
 "Regex Name": [
  null,
  "正規表現名"
 ],
 "Sets a single regular expression to use to identify entries to exclude (autoMemberExclusiveRegex)": [
  null,
  "エントリを除外するために使用する単一の正規表現を設定します。(autoMemberExclusiveRegex)"
 ],
 "Type a regex...": [
  null,
  "正規表現を入力してください..."
 ],
 "Sets which group to add the entry to as a member, if it meets the regular expression conditions (autoMemberTargetGroup)": [
  null,
  "正規表現の条件を満たす場合に、エントリをメンバーとして追加するグループを設定します。(autoMemberTargetGroup)"
 ],
 "Delete Autemebership Configuration": [
  null,
  "自動メンバーシップ設定を削除"
 ],
 "Deleting Configuration...": [
  null,
  "設定を削除しています..."
 ],
 "Linked Attributes Plugin Config Entry": [
  null,
  "Linked Attributesプラグインの設定エントリ"
 ],
 "The Linked Attributes configuration name": [
  null,
  "リンク属性の設定名を設定します"
 ],
 "Sets the attribute that is managed manually by administrators (linkType)": [
  null,
  "管理者によって手動で管理される属性を設定します。(linkType)"
 ],
 "Sets the attribute that is created dynamically by the plugin (managedType)": [
  null,
  "プラグインによって動的に作成される属性を設定します。(managedType)"
 ],
 "Sets the base DN that restricts the plugin to a specific part of the directory tree (linkScope)": [
  null,
  "プラグインをディレクトリツリーの特定の部分に制限するベースDNを設定します。(linkScope)"
 ],
 "Delete Linked Attribute Configuration": [
  null,
  "リンクされた属性の構成を削除"
 ],
 "Error during the template entry $0 operation - $1": [
  null,
  "テンプレートエントリの$0に失敗しました - $1"
 ],
 "Template entry $0 was successfully deleted": [
  null,
  "テンプレートエントリ $0 を削除しました"
 ],
 "Error during the template entry removal operation - $0": [
  null,
  "テンプレートエントリの削除操作中にエラーが発生しました - $0"
 ],
 "Failed to get attributes - $0": [
  null,
  "属性の取得に失敗しました - $0"
 ],
 "$0 Managed Entries Template Entry": [
  null,
  "Managed Entriesテンプレートエントリを$0"
 ],
 "$0 Managed Entries Definition Entry": [
  null,
  "Managed Entries定義エントリを$0"
 ],
 "A valid DN must be provided": [
  null,
  "有効なDNを入力してください"
 ],
 "Sets the search base DN to use to find candidate entries (originScope)": [
  null,
  "エントリの候補を見つけるために使用する検索ベースDNを設定します。(originScope)"
 ],
 "Sets the search filter to use to search for and identify the entries within the subtree which require a managed entry (originFilter)": [
  null,
  "サブツリー内で、管理対象エントリが必要なエントリを検索し、特定するために使用する検索フィルタを設定します。 (originFilter)"
 ],
 "Sets the subtree where the managed entries are created (managedBase)": [
  null,
  "管理対象エントリが作成されるサブツリーを設定します(managedBase)"
 ],
 "Template": [
  null,
  "テンプレート"
 ],
 "Choose which template to use for this definition": [
  null,
  "この定義に使用するテンプレートを選択してください"
 ],
 "Save Template": [
  null,
  "テンプレートを保存"
 ],
 "The template DN must be set to a valid DN": [
  null,
  "テンプレートDNは有効なDNに設定する必要があります"
 ],
 "DN of the template entry": [
  null,
  "テンプレートエントリのDNを設定します"
 ],
 "RDN Attribute": [
  null,
  "RDN属性"
 ],
 "Mapped Attributes": [
  null,
  "マッピングされた属性"
 ],
 "Dynamic attribute mappings": [
  null,
  "動的属性マッピングを設定します"
 ],
 "You must have at least one Mapped Attribute that matches the RDN Attribute": [
  null,
  "少なくとも1つのマッピングされた属性がRDN属性に一致する必要があります"
 ],
 "Add Attribute": [
  null,
  "属性を追加"
 ],
 "Delete Attribute": [
  null,
  "属性を削除"
 ],
 "Static Attributes": [
  null,
  "静的属性"
 ],
 "Static attribute mappings": [
  null,
  "静的属性マッピングを設定します"
 ],
 "Add Static Attribute": [
  null,
  "静的属性を追加"
 ],
 "Static Attribute": [
  null,
  "静的属性"
 ],
 "The attribute that is set in the managed entry": [
  null,
  "管理されたエントリに設定される属性を追加します"
 ],
 "Attribute Value": [
  null,
  "属性値"
 ],
 "The value set for the static attribute in the entry": [
  null,
  "エントリの静的属性に設定される値を設定します"
 ],
 "Add Mapped Attribute": [
  null,
  "マップされた属性を追加"
 ],
 "Mapped Attribute": [
  null,
  "マップされた属性"
 ],
 "Mapped Value": [
  null,
  "マップされた値"
 ],
 "The value set for the mapped attribute": [
  null,
  "マップされた属性に設定される値を設定します"
 ],
 "Loading Managed Entries Configuration ...": [
  null,
  "Managed Entriesの構成を読み込んでいます..."
 ],
 "Templates": [
  null,
  "テンプレート"
 ],
 "Templates are used by Definitions to construct the \"Managed Entry\".  You must have at least one Template before you can create a Definition entry.": [
  null,
  "テンプレートは定義によって \"Managed Entry\" を構築するために使用されます。定義エントリを作成する前に、少なくとも1つのテンプレートが必要です。"
 ],
 "What is a Template?": [
  null,
  "テンプレートとは？"
 ],
 "Create Template": [
  null,
  "テンプレート作成"
 ],
 "Definitions": [
  null,
  "定義"
 ],
 "Definitions describe the criteria that entries must meet in order for its \"managed entry\" to be created.  The Managed entry will be created using the Template specified in its configuration.  You can not create a Definition until there is at least one Template.": [
  null,
  "定義は、エントリが \"Managed Entry\" として作成されるために必要な基準を説明します。Managed Entryは、その構成で指定されたテンプレートを使用して作成されます。テンプレートが少なくとも1つないと、定義を作成できません。"
 ],
 "What is a Definition?": [
  null,
  "定義とは？"
 ],
 "Shared Config Area": [
  null,
  "共有構成領域"
 ],
 "The DN for the shared conifig area is invalid": [
  null,
  "共有構成領域のDNが無効です"
 ],
 "Delete Template Entry": [
  null,
  "テンプレートエントリを削除"
 ],
 "Are you really sure you want to delete this template entry?": [
  null,
  "本当にこのテンプレートを削除しますか？"
 ],
 "Deleting template ...": [
  null,
  "テンプレートを削除しています..."
 ],
 "Delete Template": [
  null,
  "テンプレートを削除"
 ],
 "Delete Definition Entry": [
  null,
  "定義エントリを削除"
 ],
 "Are you really sure you want to delete this definition entry?": [
  null,
  "本当にこの定義エントリを削除しますか？"
 ],
 "Deleting definition ...": [
  null,
  "定義を削除しています..."
 ],
 "Delete Definition": [
  null,
  "定義を削除"
 ],
 "Delete Mapped Attribute From Template": [
  null,
  "テンプレートからマップされた属性を削除"
 ],
 "Are you really sure you want to delete this mapped attribute?": [
  null,
  "本当にこのマップされた属性を削除しますか？"
 ],
 "Deleting mapped attribute ...": [
  null,
  "マップされた属性を削除しています..."
 ],
 "Delete Static Attribute From Template": [
  null,
  "テンプレートから静的属性を削除"
 ],
 "Are you really sure you want to delete this static attribute?": [
  null,
  "本当にこの静的属性を削除しますか？"
 ],
 "Deleting static attribute ...": [
  null,
  "静的属性を削除しています..."
 ],
 "PAMConfig entry $0 was successfully deleted": [
  null,
  "PAMConfigエントリ $0 を削除しました"
 ],
 "Error during the pamConfig entry removal operation - $0": [
  null,
  "pamConfigエントリの削除に失敗しました - $0"
 ],
 "Error during the pamConfig entry $0 operation - $1": [
  null,
  "pamConfigエントリの$0に失敗しました - $1"
 ],
 "$0 PAM Passthrough Auth Config Entry": [
  null,
  "PAMパススルー認証の設定を$0"
 ],
 "Specifies a suffix to exclude from PAM authentication (pamExcludeSuffix)": [
  null,
  "PAM認証から除外するサフィックスを指定します。(pamExcludeSuffix)"
 ],
 "Type a suffix DN ...": [
  null,
  "サフィックスDNを入力してください..."
 ],
 "Sets a suffix to include for PAM authentication (pamIncludeSuffix)": [
  null,
  "PAM認証に含めるサフィックスを設定します。(pamIncludeSuffix)"
 ],
 "Include Suffix": [
  null,
  "含めるサフィックス"
 ],
 "Contains the attribute name which is used to hold the PAM user ID (pamIDAttr)": [
  null,
  "PAMユーザIDを保持するために使用される属性名を含みます。(pamIDAttr)"
 ],
 "ID Attribute": [
  null,
  "ID属性"
 ],
 "Choose an attribute...": [
  null,
  "属性を選択してください..."
 ],
 "Identifies how to handle missing include or exclude suffixes (pamMissingSuffix)": [
  null,
  "含まれるまたは除外されるサフィックスがない場合の処理方法を示します。(pamMissingSuffix)"
 ],
 "Missing Suffix": [
  null,
  "サフィックス無しの場合"
 ],
 "Sets an LDAP filter to use to identify specific entries within the included suffixes for which to use PAM pass-through authentication (pamFilter)": [
  null,
  "PAMパススルー認証を使用するために含まれる、サフィックス内の特定のエントリを識別するためにLDAPフィルタを設定します。(pamFilter)"
 ],
 "Gives the method to use to map the LDAP bind DN to a PAM identity (pamIDMapMethod)": [
  null,
  "LDAPバインドDNをPAMアイデンティティにマップするために使用する方法を指定します (pamIDMapMethod)"
 ],
 "Contains the service name to pass to PAM (pamService)": [
  null,
  "PAMに渡すサービス名を含みます(pamService)"
 ],
 "Service": [
  null,
  "サービス"
 ],
 "Fallback Auth Enabled": [
  null,
  "フォールバック認証が有効"
 ],
 "Sets whether to fallback to regular LDAP authentication if PAM authentication fails (pamFallback)": [
  null,
  "PAM認証が失敗した場合に通常のLDAP認証にフォールバックするかどうかを設定します (pamFallback)"
 ],
 "Require Secure Connection": [
  null,
  "セキュア接続を必須"
 ],
 "Requires secure TLS connection for PAM authentication (pamSecure)": [
  null,
  "PAM認証にはセキュアなTLS接続が必要です(pamSecure)"
 ],
 "Delete PAM Passthrough Configuration": [
  null,
  "PAMパススルー構成を削除"
 ],
 "URL $0 was successfully deleted": [
  null,
  "URL $0 を削除しました"
 ],
 "Error during the URL removal operation - $0": [
  null,
  "URLの削除に失敗しました - $0"
 ],
 "Error during the URL $0 operation - $1": [
  null,
  "URLの$0に失敗しました - $1"
 ],
 "Authentication Hostname": [
  null,
  "認証ホスト名"
 ],
 "The authenticating directory host name. The port number of the Directory Server can be given by adding a colon and then the port number. For example, dirserver.example.com:389. If the port number is not specified, the PTA server attempts to connect using either of the standard ports: Port 389 if ldap:// is specified in the URL. Port 636 if ldaps:// is specified in the URL.": [
  null,
  "認証用ディレクトリのホスト名です。ディレクトリサーバのポート番号は、コロンとポート番号を追加することで指定できます。(例: dirserver.example.com:389 ) ポート番号が指定されていない場合、PTAサーバは標準ポートのいずれかを使用して接続を試行します。URLにldap://が指定されている場合は、ポート389を使用します。URLにldaps://が指定されている場合は、ポート636を使用して接続を試みます。"
 ],
 "Subtree": [
  null,
  "Subtree"
 ],
 "The pass-through subtree. The PTA Directory Server passes through bind requests to the authenticating Directory Server from all clients whose DN is in this subtree.": [
  null,
  "パススルーサブツリーです。PTAディレクトリサーバは、DNがこのサブツリー内にあるすべてのクライアントからのバインドリクエストを認証用ディレクトリサーバに渡します。"
 ],
 "Maximum Number of Connections": [
  null,
  "最大接続数"
 ],
 "The maximum number of connections the PTA directory can simultaneously open to the authenticating directory.": [
  null,
  "PTAディレクトリが認証用ディレクトリに対して同時に開くことのできる最大接続数です。"
 ],
 "Maximum Number of Simultaneous Operations": [
  null,
  "同時操作の最大数"
 ],
 "The maximum number of simultaneous operations (usually bind requests) the PTA directory can send to the authenticating directory within a single connection.": [
  null,
  "PTAディレクトリが単一の接続内で認証用ディレクトリに送信できる同時操作(通常はバインドリクエスト)の最大数です。"
 ],
 "Timeout": [
  null,
  "タイムアウト"
 ],
 "The time limit, in seconds, that the PTA directory waits for a response from the authenticating Directory Server. If this timeout is exceeded, the server returns an error to the client. The default is 300 seconds (five minutes). Specify zero (0) to indicate no time limit should be enforced.": [
  null,
  "PTAディレクトリが認証ディレクトリサーバからの応答を待つタイムリミットを秒単位で設定します。このタイムアウトを超えると、サーバはクライアントにエラーを返します。デフォルトは300秒(5分)です。タイムリミットを強制しない場合はゼロ(0)を指定します。"
 ],
 "Connection Life Time": [
  null,
  "接続ライフタイム"
 ],
 "The time limit, in seconds, within which a connection may be used.": [
  null,
  "接続が使用可能な時間制限を秒単位で設定します。"
 ],
 "$0 Pass-Though Authentication URL": [
  null,
  "パススルー認証URLを$0"
 ],
 "Defines whether TLS is used for communication between the two Directory Servers.": [
  null,
  "2つのディレクトリサーバ間の通信にTLSを使用するかどうかを定義します。"
 ],
 "Connection Type": [
  null,
  "接続タイプ"
 ],
 "The version of the LDAP protocol used to connect to the authenticating directory. Directory Server supports LDAP version 2 and 3. The default is version 3, and Red Hat strongly recommends against using LDAPv2, which is old and will be deprecated.": [
  null,
  "認証ディレクトリに接続するために使用されるLDAPプロトコルのバージョン。ディレクトリサーバは、LDAPバージョン2と3をサポートしています。デフォルトはバージョン3で、古いため非推奨となるLDAPv2は使用しないよう、Red Hatが強く推奨しています。"
 ],
 "A flag of whether to use Start TLS for the connection to the authenticating directory. Start TLS establishes a secure connection over the standard port, so it is useful for connecting using LDAP instead of LDAPS. The TLS server and CA certificates need to be available on both of the servers. To use Start TLS, the LDAP URL must use ldap:, not ldaps:.": [
  null,
  "認証ディレクトリへの接続にStart TLSを使用するかどうかのフラグです。Start TLSは、標準ポートを介して安全な接続を確立するため、LDAPSの代わりにLDAPを使用して接続するのに便利です。TLSサーバおよびCA証明書は、両方のサーバで利用可能である必要があります。Start TLSを使用するには、LDAP URLはldaps:ではなくldap:を使用する必要があります。"
 ],
 "Enable StartTLS": [
  null,
  "StartTLSを有効にする"
 ],
 "The URL that will be added or modified after you click 'Save'": [
  null,
  "'保存'をクリックした後に追加または変更されるURL"
 ],
 "Result URL": [
  null,
  "結果のURL"
 ],
 "Add URL": [
  null,
  "URLを追加"
 ],
 "Delete Passthru Authentication URL": [
  null,
  "パススルー認証URLを削除"
 ],
 "Are you sure you want to delete this URL?": [
  null,
  "このURLを削除しますか？"
 ],
 "Deleting URL...": [
  null,
  "URLを削除しています..."
 ],
 "Create Database Link": [
  null,
  "データベースリンクの作成"
 ],
 "Creating Link ...": [
  null,
  "リンクを作成しています ..."
 ],
 "Creating": [
  null,
  "作成"
 ],
 "Creating Link": [
  null,
  "リンクの作成"
 ],
 "The RDN of the link suffix.": [
  null,
  "リンクサフィックスのRDN"
 ],
 "Link Sub-Suffix": [
  null,
  "リンクのサブサフィックス"
 ],
 "Required field": [
  null,
  "必須項目です"
 ],
 "A name for the backend chaining database link.": [
  null,
  "バックエンドチェイニングデータベースのリンクの名前です。"
 ],
 "Link Database Name": [
  null,
  "リンクデータベース名"
 ],
 "The LDAP URL for the remote server.  Add additional failover LDAP URLs separated by a space. (nsfarmserverurl).": [
  null,
  "リモートサーバのLDAP URLです。スペースで区切られた追加のフェイルオーバーLDAP URLを追加します。(nsfarmserverurl)"
 ],
 "Remote Server URL(s)": [
  null,
  "リモートサーバのURL"
 ],
 "The DN of the entry to authenticate with on the remote server.": [
  null,
  "リモートサーバで認証するエントリのDN"
 ],
 "Remote Server Bind DN": [
  null,
  "リモートサーバのバインドDN"
 ],
 "The credentials for the bind DN (nsmultiplexorcredentials).": [
  null,
  "バインドDNの認証情報を設定します(nsmultiplexorcredentials)"
 ],
 "Bind DN Credentials": [
  null,
  "バインドDNの認証情報"
 ],
 "Password does not match": [
  null,
  "パスワードが一致しません"
 ],
 "Confirm credentials for the bind DN (nsmultiplexorcredentials).": [
  null,
  "バインドDN(nsmultiplexorcredentials)の認証情報を確認します。"
 ],
 "Confirm Password": [
  null,
  "パスワードの確認"
 ],
 "The bind method for contacting the remote server  (nsbindmechanism).": [
  null,
  "リモートサーバに接続するためのバインドメソッドを設定します(nsbindmechanism)"
 ],
 "Bind Method": [
  null,
  "バインドメソッド"
 ],
 "Use StartTLS for the remote server LDAP URL.": [
  null,
  "リモートサーバのLDAP URLにStartTLSを使用します。"
 ],
 "Use StartTLS": [
  null,
  "StartTLSを使用する"
 ],
 "Create Sub-Suffix": [
  null,
  "サブサフィックスの作成"
 ],
 "Creating suffix ...": [
  null,
  "サフィックスを作成しています ..."
 ],
 "Create Sub Suffix": [
  null,
  "サブサフィックスの作成"
 ],
 "Creating Suffix": [
  null,
  "サフィックスの作成"
 ],
 "Database suffix, like 'dc=example,dc=com'.  The suffix must be a valid LDAP Distinguished Name (DN)": [
  null,
  "データベースのサフィックスは 'dc=example,dc=com' のようなものです。サフィックスは有効なLDAPの識別名(DN)でなければなりません。"
 ],
 "Sub-Suffix DN": [
  null,
  "サブサフィックスDN"
 ],
 "The name for the backend database, like 'userroot'.  The name can be a combination of alphanumeric characters, dashes (-), and underscores (_). No other characters are allowed, and the name must be unique across all backends.": [
  null,
  "「userroot」のようなバックエンドデータベースの名前です。 英数字、ダッシュ(-)、アンダースコア(_)を組み合わせた名前になります。それ以外の文字は使用できません。また、この名前はすべてのバックエンドで一意である必要があります。"
 ],
 "Database Name": [
  null,
  "データベース名"
 ],
 "Database initialization options": [
  null,
  "データベース初期化オプション"
 ],
 "Initialization Option": [
  null,
  "初期化オプション"
 ],
 "Do Not Initialize Database": [
  null,
  "データベースを初期化しない"
 ],
 "Create The Top Sub-Suffix Entry": [
  null,
  "トップサフィックスエントリを作成"
 ],
 "Add Sample Entries": [
  null,
  "サンプルエントリを追加"
 ],
 "Export Database": [
  null,
  "データベースのエクスポート"
 ],
 "Exporting ...": [
  null,
  "エクスポートしています ..."
 ],
 "Export Database To LDIF File": [
  null,
  "データベースをLDIFファイルに書き出す"
 ],
 "Name of exported LDIF file.": [
  null,
  "エクスポートされるLDIFファイルの名前。"
 ],
 "LDIF File Name": [
  null,
  "LDIFファイル名"
 ],
 "Include the replication metadata needed to restore or initialize another replica.": [
  null,
  "別のレプリカを復元または初期化するために必要なレプリケーションメタデータを含む。"
 ],
 "Include Replication Data": [
  null,
  "レプリケーションデータを含める"
 ],
 "Initialize Database via LDIF File": [
  null,
  "LDIFファイルによるデータベースの初期化"
 ],
 "The full path to the LDIF file.  The server must have permissions to read it.": [
  null,
  "LDIFファイルへのフルパスです。サーバに読み込む権限が必要です。"
 ],
 "Or, enter LDIF location": [
  null,
  "または、LDIFの場所を入力してください。"
 ],
 "Import": [
  null,
  "インポート"
 ],
 "Referral": [
  null,
  "リファラル"
 ],
 "Delete Referral": [
  null,
  "リファラルを削除"
 ],
 "Delete this referral": [
  null,
  "このリファラルを削除する"
 ],
 "No Referrals": [
  null,
  "リファラルがありません"
 ],
 "Indexing Types": [
  null,
  "インデックスタイプ"
 ],
 "Matching Rules": [
  null,
  "マッチングルール"
 ],
 "Edit Index": [
  null,
  "インデックスを編集"
 ],
 "Reindex": [
  null,
  "再インデックス"
 ],
 "Delete Index": [
  null,
  "インデックスを削除"
 ],
 "Indexes": [
  null,
  "インデックス"
 ],
 "No Indexes": [
  null,
  "インデックスがありません"
 ],
 "Search indexes": [
  null,
  "インデックスを検索"
 ],
 "Encrypted Attribute": [
  null,
  "暗号化属性"
 ],
 "Delete this attribute": [
  null,
  "この属性を削除する"
 ],
 "No Attributes": [
  null,
  "属性がありません"
 ],
 "LDIF File": [
  null,
  "LDIFファイル"
 ],
 "Creation Date": [
  null,
  "作成日"
 ],
 "File Size": [
  null,
  "ファイルサイズ"
 ],
 "Initialize the database with this LDIF file": [
  null,
  "このLDIFファイルを使ってデータベースを初期化します。"
 ],
 "No LDIF files": [
  null,
  "LDIFファイルがありません"
 ],
 "Import LDIF File": [
  null,
  "LDIFファイルのインポート"
 ],
 "Suffix": [
  null,
  "サフィックス"
 ],
 "Import LDIF": [
  null,
  "LDIFをインポート"
 ],
 "Delete LDIF": [
  null,
  "LDIFを削除"
 ],
 "Backup": [
  null,
  "バックアップ"
 ],
 "Size": [
  null,
  "サイズ"
 ],
 "No Backups": [
  null,
  "バックアップがありません"
 ],
 "Backups": [
  null,
  "バックアップ"
 ],
 "Restore Backup": [
  null,
  "バックアップを復元"
 ],
 "Delete Backup": [
  null,
  "バックアップを削除"
 ],
 "Target DN": [
  null,
  "ターゲットDN"
 ],
 "Policy Type": [
  null,
  "ポリシータイプ"
 ],
 "Database Suffix": [
  null,
  "データベースサフィックス"
 ],
 "No Local Policies": [
  null,
  "ローカルポリシーがありません"
 ],
 "Local Password Policies": [
  null,
  "ローカルパスワードポリシー"
 ],
 "Edit Policy": [
  null,
  "ポリシーを編集"
 ],
 "Delete policy": [
  null,
  "ポリシーを削除"
 ],
 "Search Base": [
  null,
  "検索ベース"
 ],
 "Disabled": [
  null,
  "無効"
 ],
 "Uses: ": [
  null,
  "使用します: "
 ],
 "Search Base:": [
  null,
  "検索ベース:"
 ],
 "Search Filter:": [
  null,
  "検索フィルタ:"
 ],
 "Scope:": [
  null,
  "スコープ:"
 ],
 "Sort Indexes:": [
  null,
  "インデックスを並べ替える:"
 ],
 "Create Sort Index": [
  null,
  "ソートインデックスの作成"
 ],
 "No VLV Indexes": [
  null,
  "VLVインデックスがありません"
 ],
 "VLV Indexes": [
  null,
  "VLVインデックス"
 ],
 "Reindex VLV": [
  null,
  "VLVを再インデックスする"
 ],
 "Delete VLV": [
  null,
  "VLVを削除する"
 ],
 "LDIF was successfully imported": [
  null,
  "LDIFは正常にインポートされました"
 ],
 "Failure importing LDIF - $0": [
  null,
  "LDIFのインポートに失敗しました - $0"
 ],
 "LDIF file was successfully deleted": [
  null,
  "LDIFファイルの削除に成功しました"
 ],
 "Failure deleting LDIF file - $0": [
  null,
  "LDIFファイルの削除に失敗しました - $0"
 ],
 "Backup name should not be a path.  All backups are stored in the server's backup directory": [
  null,
  "バックアップの名前はパスではないようにしてください。すべてのバックアップはサーバのバックアップディレクトリに保存されます。"
 ],
 "Server has been backed up. You can find the backup in $0 directory on the server machine.": [
  null,
  "サーバをバックアップしました。バックアップはサーバの$0ディレクトリにあります。"
 ],
 "Server has been backed up.": [
  null,
  "サーバがバックアップされました。"
 ],
 "Error while trying to get the server's backup directory- $0": [
  null,
  "サーバのバックアップ中にエラーが発生しました- $0"
 ],
 "Failure backing up server - $0": [
  null,
  "サーバのバックアップに失敗しました。 - $0"
 ],
 "There are no databases defined to restore": [
  null,
  "復元するデータベースが定義されていません"
 ],
 "Server has been restored": [
  null,
  "サーバが復元されました"
 ],
 "Failure restoring up server - $0": [
  null,
  "サーバの復元に失敗しました。 - $0"
 ],
 "Backup was successfully deleted": [
  null,
  "バックアップの削除に成功しました"
 ],
 "Failure deleting backup - $0": [
  null,
  "バックアップの削除に失敗しました - $0"
 ],
 "LDIF name is empty": [
  null,
  "LDIF名が空です"
 ],
 "LDIF name should not be a path.  All export files are stored in the server's LDIF directory": [
  null,
  "LDIF名にパスを指定しないでください。すべてのエクスポートファイルはサーバのLDIFディレクトリに保存されます。"
 ],
 "Database export complete. You can find the LDIF file in $0 directory on the server machine.": [
  null,
  "データベースのエクスポートが完了しました。LDIFファイルはサーバの$0ディレクトリにあります。"
 ],
 "Database export complete.": [
  null,
  "データベースのエクスポートが完了しました。"
 ],
 "Error while trying to get the server's LDIF directory- $0": [
  null,
  "LDIFディレクトリの取得中にエラーが発生しました- $0"
 ],
 "Error exporting database - $0": [
  null,
  "データベースのエクスポート中にエラーが発生しました- $0"
 ],
 "Refreshing ...": [
  null,
  "更新しています ..."
 ],
 "Database Backups & LDIFs": [
  null,
  "データベースのバックアップとLDIF"
 ],
 "Backups ": [
  null,
  "バックアップ"
 ],
 "Create Backup": [
  null,
  "バックアップの作成"
 ],
 "LDIFs ": [
  null,
  "LDIFs"
 ],
 "Create LDIF": [
  null,
  "LDIFの作成"
 ],
 "Delete LDIF File": [
  null,
  "LDIFファイルの削除"
 ],
 "Are you sure you want to delete this LDIF?": [
  null,
  "本当にこのLDIFを削除しますか？"
 ],
 "Deleting LDIF ...": [
  null,
  "LDIFを削除しています ..."
 ],
 "Are you sure you want to import this LDIF?": [
  null,
  "本当にこのLDIFをインポートしますか？"
 ],
 "Importing LDIF ...": [
  null,
  "LDIF をインポートしています ..."
 ],
 "Restore Database": [
  null,
  "データベースを復元"
 ],
 "Are you sure you want to restore from this backup?": [
  null,
  "本当にこのバックアップから復元しますか？"
 ],
 "Restoring ...": [
  null,
  "復元しています ..."
 ],
 "Restore": [
  null,
  "復元"
 ],
 "Are you sure you want to delete this backup?": [
  null,
  "本当にこのバックアップを削除しますか？"
 ],
 "Deleting backup ...": [
  null,
  "バックアップを削除しています ..."
 ],
 "Replace Existing Backup": [
  null,
  "既存のバックアップを置き換えます"
 ],
 "A backup already eixsts with the same name, do you want to replace it?": [
  null,
  "同じ名前のバックアップが既にありますが、置き換えますか？"
 ],
 "Replacing backup ...": [
  null,
  "バックアップを置き換えています ..."
 ],
 "Replace Backup": [
  null,
  "バックアップの置き換え"
 ],
 "Replace Existing LDIF File": [
  null,
  "既存のLDIFファイルの置き換え"
 ],
 "A LDIF file already exists with the same name, do you want to replace it?": [
  null,
  "同じ名前のLDIFファイルが既に存在します。置き換えますか？"
 ],
 "Replacing LDIF ...": [
  null,
  "LDIfを置き換えています ..."
 ],
 "Replace LDIF": [
  null,
  "LDIFを置き換える"
 ],
 "Creating ...": [
  null,
  "作成しています ..."
 ],
 "(You can safely close this window)": [
  null,
  "(このウィンドウは安全に閉じることができます)"
 ],
 "Select Suffix": [
  null,
  "サフィックスを選択"
 ],
 "Name of exported LDIF file, if left blank the data and time will be used as the file name": [
  null,
  "エクスポートされるLDIFファイルの名前を設定します。空白の場合、データと時刻がファイル名として使用されます。"
 ],
 "Backing up ...": [
  null,
  "バックアップを取得しています ..."
 ],
 "Backing up": [
  null,
  "バックアップ"
 ],
 "Backup The Server": [
  null,
  "サーバのバックアップを作成"
 ],
 "Backup name, if left blank the date and time will be used as the name": [
  null,
  "バックアップ名を設定します。空白の場合は日付と時刻が名前として使用されます。"
 ],
 "Backup Name": [
  null,
  "バックアップ名"
 ],
 "Importing LDIF ": [
  null,
  "LDIFをインポートする"
 ],
 "Successfully added encrypted attribute": [
  null,
  "暗号化属性の追加に成功しました"
 ],
 "Failed to add encrypted attribute - $0": [
  null,
  "暗号化属性の追加に失敗しました - $0"
 ],
 "Encrypted attribute successfully deleted": [
  null,
  "暗号化属性を削除しました"
 ],
 "Failure deleting encrypted attribute - $0": [
  null,
  "暗号化属性の削除に失敗しました - $0"
 ],
 "Adding Attribute ...": [
  null,
  "属性を追加しています ..."
 ],
 "Type attribute name to be encrypted": [
  null,
  "暗号化属性名を入力してください"
 ],
 "Remove Attribute Encryption": [
  null,
  "暗号化属性の削除"
 ],
 "Are you sure you want to remove this encrypted attribute?": [
  null,
  "本当にこの暗号化属性を削除しますか？"
 ],
 "Invalid DN": [
  null,
  "無効なDN"
 ],
 "Create A New Local Password Policy": [
  null,
  "ローカルパスワードポリシーの新規作成"
 ],
 "Password Policy Type": [
  null,
  "パスワードポリシーのタイプ"
 ],
 "FormSelect Input": [
  null,
  "選択フォームの入力"
 ],
 "Subtree Policy": [
  null,
  "サブツリーポリシー"
 ],
 "User Policy": [
  null,
  "ユーザポリシー"
 ],
 "The DN of the entry to apply this password policy to.": [
  null,
  "このパスワードポリシーを適用するエントリのDNを設定します。"
 ],
 "Hide General Settings": [
  null,
  "一般設定を非表示"
 ],
 "Show General Settings": [
  null,
  "一般設定を表示"
 ],
 "Set the password storage scheme (passwordstoragescheme).": [
  null,
  "パスワードの保存方式を設定します。(passwordstoragescheme)"
 ],
 "Password Storage Scheme": [
  null,
  "パスワード保存方式"
 ],
 "Indicates the number of seconds that must pass before a user can change their password again. (passwordMinAge).": [
  null,
  "ユーザがパスワードを再度変更できるようになるまでの経過秒数を設定します。(passwordMinAge)"
 ],
 "Password Minimum Age": [
  null,
  "パスワードの有効期限の最小値"
 ],
 "The DN for a password administrator or administrator group (passwordAdminDN).": [
  null,
  "パスワード管理者または管理者グループのDNを設定します。(passwordAdminDN)"
 ],
 "Password Administrator": [
  null,
  "パスワード管理者"
 ],
 "Disable updating password state attributes like passwordExpirationtime, passwordHistory, etc, when setting a user's password as a Password Administrator (passwordAdminSkipInfoUpdate).": [
  null,
  "パスワード管理者としてユーザのパスワードを設定する際、passwordExpirationtime、passwordHistoryなどのパスワードの状態属性の更新を無効にします。(passwordAdminSkipInfoUpdate)"
 ],
 "Do not update target entry's password state attributes": [
  null,
  "対象のエントリのパスワード状態属性を更新しないようにする"
 ],
 "Record a separate timestamp specifically for the last time that the password for an entry was changed. If this is enabled, then it adds the pwdUpdateTime operational attribute to the user account entry (passwordTrackUpdateTime).": [
  null,
  "エントリのパスワードが最後に変更されたときのタイムスタンプを記録します。これが有効な場合、ユーザアカウントエントリにpwdUpdateTime運用属性が追加されます。(passwordTrackUpdateTime)"
 ],
 "Track Password Update Time": [
  null,
  "パスワードの更新時間を記録する"
 ],
 "Allow user's to change their passwords (passwordChange).": [
  null,
  "ユーザがパスワードを変更できるようにします。(passwordChange)"
 ],
 "Allow Users To Change Their Passwords": [
  null,
  "ユーザがパスワードを変更できるようにする"
 ],
 "User must change its password after its been reset by an administrator (passwordMustChange).": [
  null,
  "ユーザは、管理者によってパスワードがリセットされた後、そのパスワードの変更が必要になります。 (passwordMustChange)"
 ],
 "User Must Change Password After Reset": [
  null,
  "パスワードのリセット後、ユーザのパスワード変更を必須にする"
 ],
 "Maintain a password history for each user (passwordHistory).": [
  null,
  "各ユーザのパスワード履歴を保持します。(passwordHistory)"
 ],
 "Keep Password History": [
  null,
  "パスワードの履歴を残す"
 ],
 "Hide Expiration Settings": [
  null,
  "有効期限設定の非表示"
 ],
 "Show Expiration Settings": [
  null,
  "有効期限設定の表示"
 ],
 "Enable a password expiration policy (passwordExp).": [
  null,
  "パスワードの有効期限ポリシーを有効にします。(passwordExp)"
 ],
 "Enforce Password Expiration": [
  null,
  "パスワードの有効期限を設定する"
 ],
 "The maximum age of a password in seconds before it expires (passwordMaxAge).": [
  null,
  "パスワードの有効期限の最大値を秒単位で指定します。(passwordMaxAge)"
 ],
 "Password Expiration Time": [
  null,
  "パスワードの有効期限"
 ],
 "The number of logins that are allowed after the password has expired (passwordGraceLimit).": [
  null,
  "パスワードの有効期限が切れた後に許可されるログインの回数を設定します。(passwordGraceLimit)"
 ],
 "Allowed Logins After Password Expires": [
  null,
  "期限切れパスワードでのログインを許可"
 ],
 "Set the time (in seconds), before a password is about to expire, to send a warning. (passwordWarning).": [
  null,
  "パスワードの有効期限が切れる前に、警告を送信する時間を秒単位で設定します。(passwordWarning)"
 ],
 "Send Password Expiring Warning": [
  null,
  "パスワードの有効期限切れ警告を送信"
 ],
 "Always return a password expiring control when requested (passwordSendExpiringTime).": [
  null,
  "要求があれば、常にパスワードの期限切れ制御を返します。(passwordSendExpiringTime)"
 ],
 "Always send Password Expiring Control": [
  null,
  "パスワードの有効期限切れを警告する"
 ],
 "Hide Lockout Settings": [
  null,
  "ロックアウト設定を非表示"
 ],
 "Show Lockout Settings": [
  null,
  "ロックアウト設定を表示"
 ],
 "Enable account lockout (passwordLockout).": [
  null,
  "アカウントロックアウトを有効にします。(passwordLockout)"
 ],
 "Enable Account Lockout": [
  null,
  "アカウントロックアウトの有効にする"
 ],
 "The maximum number of failed logins before account gets locked (passwordMaxFailure).": [
  null,
  "アカウントがロックされるまでのログイン失敗回数を設定します。(passwordMaxFailure)"
 ],
 "Number of Failed Logins That Locks out Account": [
  null,
  "アカウントをロックアウトするログイン失敗回数"
 ],
 "The number of seconds until an accounts failure count is reset (passwordResetFailureCount).": [
  null,
  "アカウント障害カウントがリセットされるまでの秒数を設定します。(passwordResetFailureCount)"
 ],
 "Time Until Failure Count Resets": [
  null,
  "失敗回数がリセットされるまでの時間"
 ],
 "The number of seconds, duration, before the account gets unlocked (passwordLockoutDuration).": [
  null,
  "アカウントのロックが解除されるまでの秒数、期間を設定します。(passwordLockoutDuration)"
 ],
 "Time Until Account Unlocked": [
  null,
  "アカウントロックが解除されるまでの時間"
 ],
 "Do not lockout the user account forever, instead the account will unlock based on the lockout duration (passwordUnlock).": [
  null,
  "ユーザアカウントを永遠にロックアウトするのではなく、ロックアウト期間に応じてアカウントのロックを解除します。(passwordUnlock)"
 ],
 "Do Not Lockout Account Forever": [
  null,
  "アカウントを永久にロックアウトしない"
 ],
 "Hide Syntax Settings": [
  null,
  "構文設定を非表示"
 ],
 "Show Syntax Settings": [
  null,
  "構文設定を表示"
 ],
 "Enable password syntax checking (passwordCheckSyntax).": [
  null,
  "パスワードの構文チェックを有効にします。(passwordCheckSyntax)"
 ],
 "Enable Password Syntax Checking": [
  null,
  "パスワードの構文チェックを有効にします"
 ],
 "Minimum Length": [
  null,
  "必要な長さ"
 ],
 "The minimum number of characters in the password (passwordMinLength).": [
  null,
  "パスワードの最小文字数を設定します。(passwordMinLength)"
 ],
 "Reject passwords with fewer than this many alpha characters (passwordMinAlphas).": [
  null,
  "アルファベットの数がこれより少ないパスワードは拒否されます。(passwordMinAlphas)"
 ],
 "Minimum Alpha's": [
  null,
  "必要な文字数"
 ],
 "Minimum Digits": [
  null,
  "必要な桁数"
 ],
 "Reject passwords with fewer than this many digit characters (0-9) (passwordMinDigits).": [
  null,
  "数字(0～9)がこの数より少ないパスワードは拒否されます。(passwordMinDigits)"
 ],
 "Reject passwords with fewer than this many special non-alphanumeric characters (passwordMinSpecials).": [
  null,
  "英数字以外の特殊文字がこの数より少ないパスワードは拒否されます。 (passwordMinSpecials)"
 ],
 "Minimum Special": [
  null,
  "必要な特殊文字数"
 ],
 "Minimum Uppercase": [
  null,
  "必要な大文字数"
 ],
 "Reject passwords with fewer than this many uppercase characters (passwordMinUppers).": [
  null,
  "大文字がこの数より少ないパスワードは拒否されます。(passwordMinUppers)"
 ],
 "Reject passwords with fewer than this many lowercase characters (passwordMinLowers).": [
  null,
  "小文字がこの数より少ないパスワードは拒否されます。(passwordMinLowers)"
 ],
 "Minimum Lowercase": [
  null,
  "必要な小文字数"
 ],
 "Minimum 8-bit": [
  null,
  "必要な8ビット文字数"
 ],
 "Reject passwords with fewer than this many 8-bit or multi-byte characters (passwordMin8Bit).": [
  null,
  "8ビットまたはマルチバイトの文字数がこの数より少ないパスワードは拒否されます。 (passwordMin8Bit)"
 ],
 "The minimum number of character categories that a password must contain (categories are upper, lower, digit, special, and 8-bit) (passwordMinCategories).": [
  null,
  "パスワードに含めなければいけない最小のカテゴリ数(カテゴリは、大文字、小文字、桁、特殊文字、8ビット)を設定します。(passwordMinCategories)"
 ],
 "Minimum Categories": [
  null,
  "必要なカテゴリ数"
 ],
 "Minimum Token Length": [
  null,
  "最小トークン長"
 ],
 "The smallest attribute value used when checking if the password contains any of the user's account information (passwordMinTokenLength).": [
  null,
  "ユーザのアカウント情報のいずれかがパスワードに含まれているかをチェックする際に使用される最小の属性値。(passwordMinTokenLength)"
 ],
 "The maximum number of times the same character can sequentially appear in a password (passwordMaxRepeats).": [
  null,
  "パスワードに同じ文字が連続して出現できる最大回数。(passwordMaxRepeats)"
 ],
 "Max Repeated Chars": [
  null,
  "最大繰り返し文字数"
 ],
 "Max Sequences": [
  null,
  "最大シーケンス数"
 ],
 "The maximum number of allowed monotonic characters sequences (passwordMaxSequence).": [
  null,
  "許容される単調な文字列の最大数を設定します。(passwordMaxSequence)"
 ],
 "Max Sequence Sets": [
  null,
  "最大の配列数"
 ],
 "Max Seq Per Class": [
  null,
  "1クラスあたりの最大配列数"
 ],
 "The maximum number of consecutive characters from the same character class/category (passwordMaxClassChars)..": [
  null,
  "同じ文字クラス/カテゴリから連続する文字の最大数。(passwordMaxClassChars)"
 ],
 "A space-separated list of words that are not allowed to be contained in the new password (passwordBadWords).": [
  null,
  "新しいパスワードに含むことができない文字のスペース区切りのリストを設定します。(passwordBadWords)"
 ],
 "Prohibited Words": [
  null,
  "禁止文字"
 ],
 "A list of entry attributes to compare to the new password (passwordUserAttributes).": [
  null,
  "新しいパスワードと比較するためのエントリ属性のリストを設定します。(passwordUserAttributes)"
 ],
 "Check User Attributes": [
  null,
  "ユーザ属性の確認"
 ],
 "Type attributes to check...": [
  null,
  "入力した属性を確認中..."
 ],
 "Check the password against the system's CrackLib dictionary (passwordDictCheck).": [
  null,
  "パスワードをシステムのCrackLib辞書と照らし合わせます。(passwordDictCheck)"
 ],
 "Dictionary Check": [
  null,
  "辞書の確認をする"
 ],
 "Reject a password if it is a palindrome (passwordPalindrome).": [
  null,
  "新しいパスワードに回文が含まれる場合、パスワードを拒否します。(passwordPalindrome)"
 ],
 "Reject Palindromes": [
  null,
  "回文を拒否する"
 ],
 "Hide Temporary Password Settings": [
  null,
  "一時パスワード設定を非表示"
 ],
 "Show Temporary Password Settings": [
  null,
  "一時パスワード設定を表示"
 ],
 "\"User Must Change Password After Reset\" must be enabled in General Settings to activate TPR.": [
  null,
  "TPRを有効にするには、一般設定で「User Must Change Password After Reset」を有効にする必要があります。"
 ],
 "Number of times the temporary password can be used to authenticate (passwordTPRMaxUse).": [
  null,
  "一時パスワードが認証に使用できる回数を設定します。(passwordTPRMaxUse)"
 ],
 "Password Max Use": [
  null,
  "パスワードの最大使用回数"
 ],
 "Number of seconds before the temporary password expires (passwordTPRDelayExpireAt).": [
  null,
  "一時パスワードの有効期限が切れるまでの秒数を設定します。(passwordTPRDelayExpireAt)"
 ],
 "Password Expires In": [
  null,
  "パスワードの有効期限"
 ],
 "Number of seconds after which temporary password starts to be valid for authentication (passwordTPRDelayValidFrom).": [
  null,
  "一時パスワードが認証に有効な状態になるまでの秒数を設定します。(passwordTPRDelayValidFrom)"
 ],
 "Password Valid From": [
  null,
  "パスワードの有効開始日"
 ],
 "Create New Policy": [
  null,
  "ポリシーを作成"
 ],
 "Successfully created new password policy": [
  null,
  "新しいパスワードポリシーの作成に成功しました"
 ],
 "Error creating password policy - $0": [
  null,
  "パスワードポリシーの作成中にエラーが発生しました - $0"
 ],
 "Successfully updated password policy configuration": [
  null,
  "パスワードポリシー設定の更新に成功しました"
 ],
 "Error updating password policy configuration - $0": [
  null,
  "パスワードポリシー設定の更新中にエラーが発生しました - $0"
 ],
 "Error deleting local password policy - $0": [
  null,
  "パスワードポリシーの削除中にエラーが発生しました - $0"
 ],
 "Error loading global password storage scheme - $0": [
  null,
  "グローバルパスワードスキームの読み込み中にエラーが発生しました - $0"
 ],
 "Error loading local password policy - $0": [
  null,
  "ローカルパスワードポリシーの読み込み中にエラーが発生しました - $0"
 ],
 "Maximum Sequences": [
  null,
  "必要な配列数"
 ],
 "The maximum number of allowed monotonic characters sequences that can appear more than once (passwordMaxSeqSets).": [
  null,
  "複数回出現する単調な文字列の許容最大数を設定します。(passwordMaxSeqSets)"
 ],
 "The maximum number of consecutive characters from the same character class/category (passwordMaxClassChars).": [
  null,
  "同じ文字クラス/カテゴリから連続した文字の最大数を設定します。(passwordMaxClassChars)"
 ],
 "Check if the password is a palindrome (passwordPalindrome).": [
  null,
  "パスワードが回文であるかどうかをチェックします。(passwordPalindrome)"
 ],
 "Please choose a policy from the ": [
  null,
  "次のリンクからポリシーを選択してください:"
 ],
 "Local Policy Table": [
  null,
  "ローカルポリシーテーブル"
 ],
 "General Settings": [
  null,
  "一般設定"
 ],
 "Expiration": [
  null,
  "有効期限"
 ],
 "Account Lockout": [
  null,
  "アカウントロックアウト"
 ],
 "Syntax Checking": [
  null,
  "構文チェック"
 ],
 "Temporary Password Rules": [
  null,
  "一時的なパスワードのルール"
 ],
 "Create A Policy": [
  null,
  "ポリシーの作成"
 ],
 "Local Password Policies ": [
  null,
  "ローカルパスワードポリシー"
 ],
 "Refresh the local password policies": [
  null,
  "ローカルパスワードポリシーを更新する"
 ],
 "Delete Local Password Policy": [
  null,
  "ローカルパスワードポリシーを削除する"
 ],
 "Are you sure you want to delete this local password policy?": [
  null,
  "本当にこのローカルパスワードポリシーを削除しますか？"
 ],
 "Deleting local password policy ...": [
  null,
  "ローカルパスワードポリシーを削除しています ..."
 ],
 "Delete Policy": [
  null,
  "ポリシーの削除"
 ],
 "Successfully updated chaining configuration": [
  null,
  "チェイニング設定の更新に成功しました"
 ],
 "Error updating chaining configuration - $0": [
  null,
  "チェイニング設定の更新中にエラーが発生しました - $0"
 ],
 "Successfully updated chaining controls": [
  null,
  "チェイニングコントロールの更新に成功しました"
 ],
 "Error updating chaining controls - $0,": [
  null,
  "チェイニングコントロールの更新中にエラーが発生しました - $0"
 ],
 "Successfully removed chaining controls": [
  null,
  "チェイニングコントロールの削除に成功しました"
 ],
 "Error removing chaining controls - $0": [
  null,
  "チェイニングコントロールの削除中にエラーが発生しました - $0"
 ],
 "Successfully updated chaining components": [
  null,
  "チェイニングコンポーネントの更新に成功しました"
 ],
 "Error updating chaining components - $0": [
  null,
  "チェイニングコンポーネントの更新中にエラーが発生しました - $0"
 ],
 "Successfully removed chaining components": [
  null,
  "チェイニングコンポーネントの削除に成功しました"
 ],
 "Error removing chaining components - $0": [
  null,
  "チェイニングコンポーネントの削除中にエラーが発生しました - $0"
 ],
 "Save Settings": [
  null,
  "設定を保存"
 ],
 "Saving settings ...": [
  null,
  "設定を保存しています ..."
 ],
 "Database Chaining Settings": [
  null,
  "データベースチェイニング設定"
 ],
 "Default Creation Settings": [
  null,
  "デフォルトの作成設定"
 ],
 "The size limit of entries returned over a database link (nsslapd-sizelimit).": [
  null,
  "データベースリンクを介して返されるエントリのサイズ制限を設定します。(nsslapd-sizelimit)"
 ],
 "Size Limit": [
  null,
  "サイズ制限"
 ],
 "The maximum number of operations per connections. (nsconcurrentoperationslimit).": [
  null,
  "接続ごとの最大操作数を設定します。(nsconcurrentoperationslimit)"
 ],
 "Max Operations Per Conn": [
  null,
  "接続ごとの最大操作数"
 ],
 "The time limit of an operation over a database link (nsslapd-timelimit).": [
  null,
  "データベースリンク上の操作のタイムリミットを設定します。(nsslapd-timelimit)"
 ],
 "Time Limit": [
  null,
  "タイムリミット"
 ],
 "Connection Lifetime": [
  null,
  "コネクションライフタイム"
 ],
 "The maximum number of TCP connections the database link establishes with the remote server.  (nsbindconnectionslimit).": [
  null,
  "データベースリンクがリモートサーバと確立するTCP接続の最大数を設定します。(nsbindconnectionslimit)"
 ],
 "Max TCP Connections": [
  null,
  "最大TCP接続数"
 ],
 "The maximum number of connections allowed over the database link.  (nsoperationconnectionslimit).": [
  null,
  "データベースリンクを介して許可される最大接続数を設定します。(nsoperationconnectionslimit)"
 ],
 "Max LDAP Connections": [
  null,
  "最大LDAP接続数"
 ],
 "The number of seconds that pass before the server checks for abandoned operations.  (nsabandonedsearchcheckinterval).": [
  null,
  "サーバが放棄された操作をチェックするまでの経過時間を秒単位で設定します。(nsabandonedsearchcheckinterval)"
 ],
 "Abandoned Op Check Interval": [
  null,
  "放棄された操作のチェック間隔"
 ],
 "Max Binds Per Connection": [
  null,
  "接続ごとの最大バインド数"
 ],
 "The maximum number of times a request can be forwarded from one database link to another.  (nshoplimit).": [
  null,
  "1つのデータベースリンクから別のリンクにリクエストを転送する最大回数を設定します。(nshoplimit)"
 ],
 "Database Link Hop Limit": [
  null,
  "データベースリンクのホップ制限数"
 ],
 "The amount of time before the bind attempt times out. (nsbindtimeout).": [
  null,
  "バインド試行のタイムアウトまでの時間を設定します。(nsbindtimeout)"
 ],
 "Bind Timeout": [
  null,
  "バインドタイムアウト"
 ],
 "The number of times the database link tries to bind with the remote server after a connection failure. (nsbindretrylimit).": [
  null,
  "接続の失敗後、データベースリンクがリモートサーバにバインドしようと試みる回数を設定します。(nsbindretrylimit)"
 ],
 "Bind Retry Limit": [
  null,
  "バインド再試行回数制限"
 ],
 "Sets whether ACIs are evaluated on the database link as well as the remote data server (nschecklocalaci).": [
  null,
  "データベースリンクとリモートデータサーバの両方でACIが評価されるかどうかを設定します。(nschecklocalaci)"
 ],
 "Check Local ACIs": [
  null,
  "ローカルACIを確認する"
 ],
 "Sets whether referrals are returned by scoped searches (meaning 'one-level' or 'subtree' scoped searches). (nsreferralonscopedsearch).": [
  null,
  "スコープ付き検索('one-level'または'subtree'のスコープ付き検索)でリファラルが返されるかどうかを設定します。(nsreferralonscopedsearch)"
 ],
 "Send Referral On Scoped Search": [
  null,
  "スコープ付き検索時にリファラルを送信する"
 ],
 "Sets whether proxied authentication is allowed. (nsproxiedauthorization).": [
  null,
  "プロキシ認証を許可するかどうかを設定します。(nsproxiedauthorization)"
 ],
 "Allow Proxied Authentication": [
  null,
  "プロキシ認証を許可する"
 ],
 "Use StartTLS for connections to remote server. (nsusestarttls).": [
  null,
  "リモートサーバへの接続にStartTLSを使用するかどうかを設定します。(nsusestarttls)"
 ],
 "Controls & Components": [
  null,
  "コントロール＆コンポーネント"
 ],
 "A list of LDAP control OIDs to be forwarded through chaining.": [
  null,
  "チェイニングを介して転送されるLDAP制御OIDのリスト"
 ],
 "Forwarded LDAP Controls": [
  null,
  "転送されたLDAPコントロール"
 ],
 "A list of components to go through chaining": [
  null,
  "チェイニングを経由するコンポーネントのリスト"
 ],
 "Components to Chain": [
  null,
  "チェイニングするコンポーネント"
 ],
 "Remove Chaining OID": [
  null,
  "チェイニングのOIDを削除します"
 ],
 "Are you sure you want to delete this OID?": [
  null,
  "本当にこのOIDを削除しますか？"
 ],
 "Remove Chaining Component": [
  null,
  "チェイニングコンポーネントを削除します"
 ],
 "Are you sure you want to delete this component?": [
  null,
  "本当にこのコンポーネントを削除しますか？"
 ],
 "Missing Remote Server LDAP URL": [
  null,
  "リモートサーバのLDAP URLが見つかりません"
 ],
 "Missing Remote Bind DN": [
  null,
  "リモートバインドDNが見つかりません"
 ],
 "Missing Remote Bind DN Password": [
  null,
  "リモートバインドDNのパスワードが見つかりません"
 ],
 "Missing Remote Bind DN Password Confirmation": [
  null,
  "リモートバインドDNのパスワードの確認がありません"
 ],
 "Remote Bind DN Password Do Not Match": [
  null,
  "リモートバインドDNのパスワードが一致しません"
 ],
 "Successfully Updated Link Configuration": [
  null,
  "リンクの設定の更新に成功しました"
 ],
 "Failed to update link configuration": [
  null,
  "リンク設定の更新に失敗しました"
 ],
 "Save chaining link config": [
  null,
  "チェイニングリンクの設定を保存します"
 ],
 "Successfully Deleted Database Link": [
  null,
  "データベースリンクの削除に成功しました"
 ],
 "Failed to delete database link - $0": [
  null,
  "データベースリンクの削除中にエラーが発生しました - $0"
 ],
 "Refresh database link": [
  null,
  "データベースリンクを更新"
 ],
 "Delete Link": [
  null,
  "リンクの削除"
 ],
 "The LDAP URL for the remote server.  Add additional failure server URLs by separating them with a space. (nsfarmserverurl).": [
  null,
  "リモートサーバのLDAP URLです。複数の障害サーバのURLを追加する場合は、スペースで区切って追加してください。(nsfarmserverurl)"
 ],
 "Remote Server LDAP URL": [
  null,
  "リモートサーバのLDAP URL"
 ],
 "The distinguished name (DN) of the entry to authenticate to the remote server. (nsmultiplexorbinddn).": [
  null,
  "リモートサーバに認証するためのエントリの識別名(DN)を設定します。(nsmultiplexorbinddn)"
 ],
 "The password for the authenticating entry. (nsmultiplexorcredentials).": [
  null,
  "認証エントリのパスワードです。(nsmultiplexorcredentials)"
 ],
 "Bind DN Password": [
  null,
  "バインドDNのパスワード"
 ],
 "Confirm the password for the authenticating entry. (nsmultiplexorcredentials).": [
  null,
  "認証エントリのパスワードを確認します。(nsmultiplexorcredentials)"
 ],
 "The authentication mechanism.  Simple (user name and password), SASL/DIGEST-MD5, or SASL>GSSAPI. (nsbindmechanism).": [
  null,
  "認証メカニズムです。Simple(ユーザ名とパスワード)、SASL/DIGEST-MD5、またはSASL>GSSAPIのいずれかです。(nsbindmechanism)"
 ],
 "Use StartTLS for connections to the remote server. (nsusestarttls).": [
  null,
  "リモートサーバへの接続にStartTLSを使用します。(nsusestarttls)"
 ],
 "Hide Advanced Settings": [
  null,
  "詳細設定を非表示"
 ],
 "Show Advanced Settings": [
  null,
  "詳細設定を表示"
 ],
 "The maximum number of concurrent bind operations per TCP connection. (nsconcurrentbindlimit).": [
  null,
  "TCP接続ごとの同時バインド操作の最大数"
 ],
 "Max Operations Per Connection": [
  null,
  "接続ごとの最大操作数"
 ],
 "The life of a database link connection to the remote server in seconds.  0 is unlimited  (nsconnectionlife).": [
  null,
  "リモートサーバへのデータベースリンク接続の有効期間を秒単位で設定します。0は無制限を意味します。(nsconnectionlife)"
 ],
 "Hop Limit": [
  null,
  "ホップリミット"
 ],
 "Allow proxied authentication to the remote server. (nsproxiedauthorization).": [
  null,
  "リモートサーバへのプロキシ認証を許可します。(nsproxiedauthorization)"
 ],
 "Delete Database Link": [
  null,
  "データベースリンクの削除"
 ],
 "Are you really sure you want to delete this database link?": [
  null,
  "本当にこのデータベースリンクを削除しますか？"
 ],
 "Deleting Database Link...": [
  null,
  "データベースリンクを削除しています ..."
 ],
 "Add New Controls": [
  null,
  "新しいコントロールの追加"
 ],
 "Saving Controls ...": [
  null,
  "コントロールを保存しています ..."
 ],
 "Chaining LDAP Controls": [
  null,
  "LDAPチェイニングコントロール"
 ],
 "A list of LDAP control OIDs to be forwarded through chaining": [
  null,
  "チェイニングを介して転送されるLDAP制御OIDのリスト。"
 ],
 "Available LDAP Controls": [
  null,
  "利用可能なLDAPコントロール"
 ],
 "Add New Components": [
  null,
  "新しいコンポーネントを追加する"
 ],
 "Saving Components ...": [
  null,
  "コンポーネントを保存しています ..."
 ],
 "Chaining Components": [
  null,
  "チェイニングコンポーネント"
 ],
 "A list of LDAP control components": [
  null,
  "LDAPコントロールコンポーネントのリスト"
 ],
 "Available Components": [
  null,
  "使用可能なコンポーネント"
 ],
 "$0. You must restart the Directory Server for these changes to take effect.": [
  null,
  "$0。これらの変更を有効にするためには、ディレクトリサーバを再起動する必要があります。"
 ],
 "Error updating configuration - $0": [
  null,
  "設定の更新中にエラーが発生しました - $0"
 ],
 "Sets the DB lock exhaustion value in percentage (valid range is 70-95). If too many locks are acquired, the server will abort the searches while the number of locks are not decreased. It helps to avoid DB corruption and long recovery. (nsslapd-db-locks-monitoring-threshold).": [
  null,
  "DBロック枯渇値をパーセントで設定します。(有効範囲: 70～95)ロック取得数が多すぎる場合、ロック数が減らない間はサーバが検索を中止します。DBの破損や復旧に時間がかかるのを防ぐのに役立ちます。(nsslapd-db-locks-monitoring-threshold)"
 ],
 "DB Locks Threshold Percentage": [
  null,
  "データベースをロックする閾値"
 ],
 "Sets the amount of time (milliseconds) that the monitoring thread spends waiting between checks. (nsslapd-db-locks-monitoring-pause).": [
  null,
  "監視スレッドがチェックの間に待機する時間をミリ秒単位で設定します。(nsslapd-db-locks-monitoring-pause)"
 ],
 "DB Locks Pause Milliseconds": [
  null,
  "データベースロックの待機時間(ミリ秒)"
 ],
 "Enable database and entry cache auto-tuning using a percentage of the system's current resources (nsslapd-cache-autosize). If 0 is set, the default value is used instead.": [
  null,
  "システムの現在のリソースの割合を使用してデータベースとエントリキャッシュの自動チューニングを有効にします。0を設定すると、代わりにデフォルト値が使用されます。(nsslapd-cache-autosize)"
 ],
 "Memory Percentage": [
  null,
  "メモリ使用率"
 ],
 "Sets the percentage of memory that is used for the database cache. The remaining percentage is used for the entry cache (nsslapd-cache-autosize-split). If 0 is set, the default value is used instead.": [
  null,
  "データベースキャッシュに使用するメモリの割合を設定します。残りの割合は、エントリキャッシュに使用されます。0を設定すると、代わりにデフォルト値が使用されます。(nsslapd-cache-autosize-split)"
 ],
 "DB Cache Percentage": [
  null,
  "データベースキャッシュの使用率"
 ],
 "Specifies the database index cache size in bytes (nsslapd-dbcachesize).": [
  null,
  "データベースインデックスキャッシュサイズをバイト単位で指定します。(nsslapd-dbcachesize)"
 ],
 "Database Cache Size": [
  null,
  "データベースキャッシュサイズ"
 ],
 "Enter '-1' to use 50% of available memory, '0' to disable autotuning, or enter the percentage of available memory to use.  Value range -1 through 100, default is '-1' (nsslapd-import-cache-autosize).": [
  null,
  "使用可能なメモリの50％を使用する場合は「-1」、オートチューニングを無効にする場合は「0」、または 使用するメモリの割合を入力します。 値の範囲は -1 から 100、デフォルトは「-1」です。(nsslapd-import-cache-autosize)"
 ],
 "Import Cache Autosize": [
  null,
  "インポートするキャッシュサイズの自動調整"
 ],
 "The size of the database cache in bytes used in the bulk import process. (nsslapd-import-cachesize).": [
  null,
  "バルクインポート処理で使用されるデータベースキャッシュのサイズをバイト単位で設定します。(nsslapd-import-cachesize)"
 ],
 "Import Cache Size": [
  null,
  "インポートするキャッシュサイズ"
 ],
 "Saving config ...": [
  null,
  "設定を保存しています ..."
 ],
 "Global Database Configuration": [
  null,
  "グローバルデータベース設定"
 ],
 "Limits": [
  null,
  "制限"
 ],
 "The maximum number of entries that the Directory Server will check when examining candidate entries in response to a search request (nsslapd-lookthrough-limit).": [
  null,
  "ディレクトリサーバが検索要求に応じて候補エントリを調べる際に確認するエントリーの最大数を設定します。(nsslapd-lookthrough-limit)"
 ],
 "Database Look Through Limit": [
  null,
  "データベース検索数の上限"
 ],
 "The number of entry IDs that are searched during a search operation (nsslapd-idlistscanlimit).": [
  null,
  "検索操作中に検索されるエントリIDの数を設定します。(nsslapd-idlistscanlimit)"
 ],
 "ID List Scan Limit": [
  null,
  "IDリスト検索数の上限"
 ],
 "The maximum number of entries that the Directory Server will check when examining candidate entries for a search which uses the simple paged results control (nsslapd-pagedlookthroughlimit).": [
  null,
  "単純ページング結果制御を使用する検索で、ディレクトリサーバが候補エントリを調べる際の最大エントリ数を設定します。(nsslapd-pagedlookthroughlimit)"
 ],
 "Paged Search Look Through Limit": [
  null,
  "ページング検索数の上限"
 ],
 "The number of entry IDs that are searched, specifically, for a search operation using the simple paged results control (nsslapd-pagedidlistscanlimit).": [
  null,
  "単純ページング結果制御を使用した検索操作で特に検索されるエントリIDの数を設定します。(nsslapd-pagedidlistscanlimit)"
 ],
 "Paged Search ID List Scan Limit": [
  null,
  "ページング検索IDリストのスキャン制限"
 ],
 "The maximum number of entries that the Directory Server will check when examining candidate entries in response to a range search request (nsslapd-rangelookthroughlimit).": [
  null,
  "範囲検索要求に応答して候補エントリを調べる際に、ディレクトリサーバが確認するエントリの最大数を設定します。(nsslapd-rangelookthroughlimit)"
 ],
 "Range Search Look Through Limit": [
  null,
  "範囲検索数の上限"
 ],
 "Automatic Cache Tuning": [
  null,
  "キャッシュの自動チューニング"
 ],
 "Import Cache": [
  null,
  "キャッシュのインポート"
 ],
 "Automatic Import Cache Tuning": [
  null,
  "インポートキャッシュの自動チューニング"
 ],
 "Set import cache to be set automatically": [
  null,
  "インポートキャッシュを自動で設定する"
 ],
 "NDN Cache": [
  null,
  "NDNキャッシュ"
 ],
 "Set the maximum size in bytes for the Normalized DN Cache (nsslapd-ndn-cache-max-size).": [
  null,
  "正規化DNキャッシュの最大サイズをバイト数で設定します。(nsslapd-ndn-cache-max-size)"
 ],
 "Normalized DN Cache Max Size": [
  null,
  "正規化されたDNキャッシュの最大サイズ"
 ],
 "Database Locks": [
  null,
  "データベースロック"
 ],
 "The number of database locks (nsslapd-db-locks).": [
  null,
  "データベースロックの数を設定します。(nsslapd-db-locks)"
 ],
 "Enable DB Lock Monitoring": [
  null,
  "データベースロック監視を有効にする"
 ],
 "Database lock monitoring checks if the database locks are about to be exhausted, and if they are the server will abort all the current searches in order to prevent database corruption.": [
  null,
  "データベースロック監視は、データベースのロックが切れかかっているかどうかをチェックし、切れかかっている場合は、データベースの破損を防ぐために、サーバは現在の検索をすべて中止します。"
 ],
 "Advanced Settings": [
  null,
  "詳細設定"
 ],
 "Database Transaction Log Location (nsslapd-db-logdirectory).": [
  null,
  "データベーストランザクションログの場所を設定します。(nsslapd-db-logdirectory)"
 ],
 "Transaction Logs Directory": [
  null,
  "トランザクションログディレクトリ"
 ],
 "Location for database memory mapped files.  You must specify a subdirectory of a tempfs type filesystem (nsslapd-db-home-directory).": [
  null,
  "データベースのメモリーマップドファイルが置かれる場所を設定します。 tempfsタイプのファイルシステムのサブディレクトリを指定する必要があります。(nsslapd-db-home-directory)"
 ],
 "Database Home Directory": [
  null,
  "データベースホームディレクトリ"
 ],
 "The Time Of Day to perform the database compaction after the compact interval has been met.  Uses the format: 'HH:MM' and defaults to '23:59'. (nsslapd-db-compactdb-time)": [
  null,
  "圧縮のインターバルが満たされた後、データベースの圧縮を実行する日時を設定します。 HH:MM形式を使用し、デフォルトは「23:59」です。(nsslapd-db-compactdb-time)"
 ],
 "Database Compaction Time": [
  null,
  "データベース圧縮時間"
 ],
 "The interval in seconds when the database is compacted (nsslapd-db-compactdb-interval). The default is 30 days at midnight. 0 is no compaction.": [
  null,
  "データベースを圧縮する間隔を秒単位で設定します。デフォルトは30日後の午前0時です。0の場合、圧縮しません。(nsslapd-db-compactdb-interval)"
 ],
 "Database Compaction Interval": [
  null,
  "データベース圧縮間隔"
 ],
 "Amount of time in seconds after which the Directory Server sends a checkpoint entry to the database transaction log (nsslapd-db-checkpoint-interval).": [
  null,
  "ディレクトリサーバがデータベーストランザクションログにチェックポイントエントリを送信するまでの時間を秒単位で設定します。(nsslapd-db-checkpoint-interval)"
 ],
 "Database Checkpoint Interval": [
  null,
  "データベースチェックポイント間隔"
 ],
 "Failed to get matching rules - $0": [
  null,
  "マッチングルールの取得に失敗しました - $0"
 ],
 "Successfully created new index": [
  null,
  "新しいインデックスを作成しました"
 ],
 "Error creating index - $0": [
  null,
  "インデックスの作成に失敗しました - $0"
 ],
 "Error indexing attribute $0 - $1": [
  null,
  "属性 $0 のインデックス作成中にエラーが発生しました - $1"
 ],
 "Successfully edited index": [
  null,
  "インデックスの編集に成功しました"
 ],
 "Error editing index - $0": [
  null,
  "インデックス編集中にエラーが発生しました - $0"
 ],
 "Error deleting index - $0": [
  null,
  "インデックスの削除中にエラーが発生しました - $0"
 ],
 "Database Indexes ": [
  null,
  "データベースインデックス"
 ],
 "Add Index": [
  null,
  "インデックスを追加"
 ],
 "System Indexes ": [
  null,
  "システムインデックス"
 ],
 "Reindex Attribute": [
  null,
  "属性の再インデックス"
 ],
 "Are you sure you want to reindex this attribute?": [
  null,
  "本当にこの属性を再インデックスしますか？"
 ],
 "Reindexing ...": [
  null,
  "再インデックスしています ..."
 ],
 "Are you sure you want to delete this index?": [
  null,
  "本当にこのインデックスを削除しますか？"
 ],
 "Create Index": [
  null,
  "インデックスを作成する"
 ],
 "Add Database Index": [
  null,
  "データベースのインデックスを追加する"
 ],
 "Select an attribute to index": [
  null,
  "インデックス化する属性を選択してください"
 ],
 "Select An Attribute": [
  null,
  "属性を選択してください"
 ],
 "Type a attribute name to index": [
  null,
  "インデックス化する属性名を入力してください"
 ],
 "Type a attribute name to index..": [
  null,
  "インデックス化する属性名を入力してください.."
 ],
 "Index Types": [
  null,
  "インデックスタイプ"
 ],
 "Equality Indexing": [
  null,
  "等価性インデックス化"
 ],
 "Presence Indexing": [
  null,
  "プレゼンスインデックス化"
 ],
 "Substring Indexing": [
  null,
  "部分文字列インデックス化"
 ],
 "Approximate Indexing": [
  null,
  "近似インデックス化"
 ],
 "List of matching rules separated by a 'space'": [
  null,
  "マッチングルールを「スペース」で区切って並べたリスト"
 ],
 "Type a matching rule name...": [
  null,
  "マッチングルール名を入力してください..."
 ],
 "Index attribute after creation": [
  null,
  "作成した属性をインデックス化"
 ],
 "Save Index": [
  null,
  "インデックスを保存する"
 ],
 "Saving index ...": [
  null,
  "インデックスを保存しています ..."
 ],
 "Edit Database Index (": [
  null,
  "データベースのインデックスを編集します("
 ],
 "Reindex Attribute After Saving": [
  null,
  "保存後に属性の再インデックスを実行する"
 ],
 "Referral successfully deleted": [
  null,
  "リファラルの削除に成功しました"
 ],
 "Failure deleting referral - $0": [
  null,
  "リファラルの削除に失敗しました - $0"
 ],
 "Referral successfully created": [
  null,
  "リファラルの作成に成功しました"
 ],
 "Failure creating referral - $0": [
  null,
  "リファラルの作成に失敗しました - $0"
 ],
 "Create Referral": [
  null,
  "リファラルを作成"
 ],
 "Are you sure you want to delete this referral?": [
  null,
  "本当にこのリファラルを削除しますか？"
 ],
 "Add Database Referral": [
  null,
  "データベースリファラルの追加"
 ],
 "Protocol": [
  null,
  "プロトコル"
 ],
 "Host Name": [
  null,
  "ホスト名"
 ],
 "Port Number": [
  null,
  "ポート番号"
 ],
 "Comma separated list of attributes to return": [
  null,
  "属性のリストを入力する場合は、コンマで区切る必要があります"
 ],
 "Attributes": [
  null,
  "属性"
 ],
 "Computed Referral": [
  null,
  "作成されるリファラル"
 ],
 "Import successfully initiated": [
  null,
  "インポートが成功しました"
 ],
 "Error importing LDIF file - $0": [
  null,
  "LDIFファイルのインポート中にエラーが発生しました - $0"
 ],
 "Database has successfully been reindexed": [
  null,
  "データベースの再インデックスに成功しました"
 ],
 "Failed to reindex database - $0": [
  null,
  "データベースの再インデックスに失敗しました - $0"
 ],
 "Successfully created new sub-suffix": [
  null,
  "サブサフィックスの新規作成に成功しました"
 ],
 "Error creating sub-suffix - $0": [
  null,
  "サブサフィックスの作成中にエラーが発生しました - $0"
 ],
 "Successfully created database link": [
  null,
  "データベースリンクの作成に成功しました"
 ],
 "Error creating database link - $0": [
  null,
  "データベースリンクの作成中にエラーが発生しました - $0"
 ],
 "Successfully deleted database": [
  null,
  "データベースの削除に成功しました"
 ],
 "Error deleting database - $0": [
  null,
  "データベースの削除中にエラーが発生しました - $0"
 ],
 "You must restart the Directory Server for these changes to take effect.": [
  null,
  "変更を反映するためには、ディレクトリサーバの再起動が必要です。"
 ],
 "Error updating suffix configuration - $0": [
  null,
  "サフィックスの設定の更新中にエラーが発生しました - $0"
 ],
 "Import an LDIF file to initialize the database": [
  null,
  "LDIFファイルをインポートしてデータベースを初期化する"
 ],
 "Initialize Suffix": [
  null,
  "サフィックスの初期化"
 ],
 "Export the database to an LDIF file": [
  null,
  "データベースをLDIFファイルへエクスポート"
 ],
 "Export Suffix": [
  null,
  "サフィックスのエクスポート"
 ],
 "Reindex the entire database": [
  null,
  "データベースをすべて再インデックスする"
 ],
 "Reindex Suffix": [
  null,
  "サフィックスの再インデックス"
 ],
 "Create a sub-suffix under this suffix": [
  null,
  "このサフィックスの下にサブサフィックスを作成する"
 ],
 "Create a database chaining link subtree": [
  null,
  "データベースチェインリンクのサブツリーを作成する"
 ],
 "This will permanently delete the database": [
  null,
  "データベースは完全に削除されます"
 ],
 "Delete Suffix": [
  null,
  "サフィックスの削除"
 ],
 "Refresh suffix": [
  null,
  "サフィックスを更新する"
 ],
 "Suffix Tasks": [
  null,
  "サフィックスのタスク"
 ],
 "Settings": [
  null,
  "設定"
 ],
 "Encrypted Attributes": [
  null,
  "暗号化された属性"
 ],
 "Are you really sure you want to delete this suffix?": [
  null,
  "本当にこのサフィックスを削除しますか？"
 ],
 "Deleting suffix ...": [
  null,
  "サフィックスを削除しています ..."
 ],
 "Initialize Database From LDIF": [
  null,
  "LDIFからデータベースを初期化する"
 ],
 "Are you sure you want to initialize the database (it will permanently overwrite the current database)?": [
  null,
  "本当にデータベースを初期化しますか？(現在のデータベースは完全に上書きされます。)"
 ],
 "Initializing Database ...": [
  null,
  "データベースを初期化しています ..."
 ],
 "Initialize Database": [
  null,
  "データベースの初期化"
 ],
 "Reindex All Attributes": [
  null,
  "すべての属性の再インデックス"
 ],
 "Are you sure you want to reindex all the attribute indexes?": [
  null,
  "本当にすべての属性を再インデックスしますか？"
 ],
 "Reindexing Database ...": [
  null,
  "データベースを再インデックスしています ..."
 ],
 "The entry cache size in bytes setting is being auto-sized and is read-only - see Global Database Configuration": [
  null,
  "エントリキャッシュのサイズをバイト単位で設定します。エントリキャッシュは自動サイズ化され、読み取り専用です。詳細は、グローバルデータベースの設定を参照してください。"
 ],
 "Entry Cache Size": [
  null,
  "エントリキャッシュサイズ"
 ],
 "The entry cache max entries setting is being auto-sized and is read-only - see Global Database Configuration": [
  null,
  "エントリキャッシュの最大エントリ数の設定は自動サイズ化され、読み取り専用です。詳細は、グローバルデータベースの設定を参照してください。"
 ],
 "Entry Cache Max Entries": [
  null,
  "エントリキャッシュの最大エントリ数"
 ],
 "The available memory space in bytes for the DN cache. The DN cache is similar to the entry cache for a database, only its table stores only the entry ID and the entry DN (nsslapd-dncachememsize).": [
  null,
  "DNキャッシュのための利用可能なメモリスペースをバイト単位で設定します。DNキャッシュはデータベースのエントリキャッシュに似ていますが、そのテーブルにはエントリIDとエントリDNのみが格納されます。(nsslapd-dncachememsize)"
 ],
 "DN Cache Size": [
  null,
  "DNキャッシュサイズ"
 ],
 "The size for the available memory space in bytes for the entry cache (nsslapd-cachememsize).": [
  null,
  "エントリキャッシュのための利用可能なメモリスペースのサイズをバイト単位で設定します。(nsslapd-cachememsize)"
 ],
 "The number of entries to keep in the entry cache, use'-1' for unlimited (nsslapd-cachesize).": [
  null,
  "エントリキャッシュに保持するエントリの数です。制限なしの場合は「-1」を使用します。(nsslapd-cachesize)"
 ],
 "the available memory space in bytes for the DN cache. The DN cache is similar to the entry cache for a database, only its table stores only the entry ID and the entry DN (nsslapd-dncachememsize).": [
  null,
  "DNキャッシュのための利用可能なメモリスペースのサイズをバイト単位で設定します。DNキャッシュはデータベースのエントリキャッシュに類似していますが、そのテーブルにはエントリIDとエントリDNのみが格納されます。(nsslapd-dncachememsize)"
 ],
 "Save Configuration": [
  null,
  "設定を保存"
 ],
 "saving": [
  null,
  "保存中"
 ],
 "Set the backend type.  Warning, changing this setting could lead to unexpected database behavior.": [
  null,
  "バックエンドのタイプを設定してください。警告:この設定を変更すると、予期しないデータベースの動作が発生する可能性があります。"
 ],
 "Backend State": [
  null,
  "バックエンドの状態"
 ],
 "Backend": [
  null,
  "バックエンド"
 ],
 "Referral On Update": [
  null,
  "更新時のリファラル設定"
 ],
 "Put database in Read-Only mode (nsslapd-readonly).": [
  null,
  "データベースを読み取り専用モードに設定します。(nsslapd-readonly)"
 ],
 "Database Read-Only Mode": [
  null,
  "データベースを読み取り専用モードにする"
 ],
 "Block unindexed searches on this suffix (nsslapd-require-index).": [
  null,
  "このサフィックスでインデックスのない検索をブロックします。(nsslapd-require-index)"
 ],
 "Block Unindexed Searches": [
  null,
  "インデックスされていない検索をブロックする"
 ],
 "Successfully added VLV sort index": [
  null,
  "VLVソートインデックスの追加に成功しました"
 ],
 "Failed to add VLV index entry - $0": [
  null,
  "VLVソートインデックスの追加中にエラーが発生しました - $0"
 ],
 "Successfully removed VLV sort index": [
  null,
  "VLVソートインデックスの削除に成功しました"
 ],
 "Failed to delete VLV sort index - $0": [
  null,
  "VLVソートインデックスの削除中にエラーが発生しました - $0"
 ],
 "Successfully added VLV search: ": [
  null,
  "VLVサーチの追加に成功しました: "
 ],
 "Failed create VLV search entry - $0": [
  null,
  "VLVサーチエントリの追加中にエラーが発生しました - $0"
 ],
 "Successfully deleted VLV index": [
  null,
  "VLVインデックスの削除に成功しました"
 ],
 "Failed to deletre VLV index - $0": [
  null,
  "VLVインデックスの削除中にエラーが発生しました - $0"
 ],
 "Successfully completed VLV indexing": [
  null,
  "VLVのインデックス作成に成功しました"
 ],
 "Failed to index VLV index - $0": [
  null,
  "VLVのインデックス作成中にエラーが発生しました - $0"
 ],
 "Create VLV Index": [
  null,
  "VLVインデックスを作成する"
 ],
 "Delete VLV Index": [
  null,
  "VLVインデックスを削除する"
 ],
 "Are you sure you want to delete this VLV index??": [
  null,
  "本当にこのVLVインデックスを削除しますか？"
 ],
 "Deleting Index ...": [
  null,
  "インデックスを削除しています ..."
 ],
 "Reindex VLV Index": [
  null,
  "VLVインデックスを再インデックス"
 ],
 "Are you sure you want to reindex this VLV index??": [
  null,
  "本当にこのVLVインデックスを再インデックスしますか？"
 ],
 "Reindex ...": [
  null,
  "再インデックスしています ..."
 ],
 "Delete VLV Sort Index": [
  null,
  "VLVソートインデックスの削除"
 ],
 "Are you really sure you want to delete this sorting index?": [
  null,
  "本当にこのソートインデックスを削除しますか？"
 ],
 "Creating Index ...": [
  null,
  "インデックスを作成しています ..."
 ],
 "Create VLV Sort Index": [
  null,
  "VLVソートインデックスを作成"
 ],
 "Build a list of attributes to form the \"Sort\" index": [
  null,
  "「Sort」インデックスを形成するための属性のリスト"
 ],
 "Type an attribute names to create a sort index": [
  null,
  "ソートインデックスを作成する属性名を入力してください"
 ],
 "Reindex After Saving": [
  null,
  "保存後に再インデックス"
 ],
 "Save VLV Index": [
  null,
  "VLVインデックスを保存"
 ],
 "Saving Index ...": [
  null,
  "インデックスを保存しています ..."
 ],
 "Create VLV Search Index": [
  null,
  "VLV検索インデックスを作成"
 ],
 "VLV Index Name": [
  null,
  "VLVインデックス名"
 ],
 "Search Scope": [
  null,
  "検索範囲"
 ],
 "After creating this VLV Search entry you can go to the table and add VLV Sort Indexes to this VLV Search.  After adding the Sort Indexes you will need to reindex the VLV Index to make it active.": [
  null,
  "このVLV検索エントリを作成した後、テーブルに移動してこのVLV検索にVLVソートインデックスを追加できます。ソートインデックスを追加した後は、VLVインデックスを再インデックスする必要があります。これにより、VLVインデックスが有効になります。"
 ],
 "Error loading global password policy - $0": [
  null,
  "グローバルパスワードポリシーの読み込み中にエラーが発生しました - $0"
 ],
 "Allow subtree/user defined local password policies (nsslapd-pwpolicy-local).": [
  null,
  "サブツリー/ユーザ定義のローカルパスワードポリシーを許可します。(nsslapd-pwpolicy-local)"
 ],
 "Allow Local Password Policies": [
  null,
  "ローカルパスワードポリシーを許可する"
 ],
 "If a local password policy does not defined any syntax rules then inherit the local policy syntax (nsslapd-pwpolicy-inherit-global).": [
  null,
  "ローカルパスワードポリシーに構文ルールが定義されていない場合は、ローカルポリシーの構文を継承します。(nsslapd-pwpolicy-inherit-global)"
 ],
 "Local Policies Inherit Global Policy": [
  null,
  "ローカルポリシーはグローバルポリシーを継承する"
 ],
 "Allow anyone to add a prehashed password (nsslapd-allow-hashed-passwords).": [
  null,
  "誰でもプレハッシュドパスワードを追加できるようにします。(nsslapd-allow-hashed-passwords)。"
 ],
 "Allow Adding Pre-Hashed Passwords": [
  null,
  "プレハッシュドパスワードの追加を許可する"
 ],
 "Allow password policy state attributes to replicate (passwordIsGlobalPolicy).": [
  null,
  "パスワードポリシーの状態属性の複製を許可します。(passwordIsGlobalPolicy)"
 ],
 "Replicate Password Policy State Attributes": [
  null,
  "パスワードポリシーの状態属性を複製する"
 ],
 "Global Password Policy": [
  null,
  "グローバルパスワードポリシー"
 ],
 "Refresh global password policy settings": [
  null,
  "グローバルパスワードポリシーの設定を更新します"
 ],
 "Failed to get agreement information for: \"$0\" - $1": [
  null,
  "合意情報の取得に失敗しました : \"$0\" - $1"
 ],
 "Successfully updated replication agreement": [
  null,
  "レプリカ合意を更新しました"
 ],
 "Failed to update replication agreement - $0": [
  null,
  "レプリカ合意の更新に失敗しました - $0"
 ],
 "Successfully poked replication agreement": [
  null,
  "レプリカ合意の検証に成功しました"
 ],
 "Failed to poke replication agreement - $0": [
  null,
  "レプリカ合意の検証に失敗しました - $0"
 ],
 "Failed to initialize replication agreement - $0": [
  null,
  "レプリカ合意の初期化に失敗しました - $0"
 ],
 "Successfully enabled replication agreement": [
  null,
  "レプリカ合意を有効化しました"
 ],
 "Failed to enabled replication agreement - $0": [
  null,
  "レプリカ合意の有効化に失敗しました - $0"
 ],
 "Successfully disabled replication agreement": [
  null,
  "レプリカ合意を無効化しました"
 ],
 "Failed to disable replication agreement - $0": [
  null,
  "レプリカ合意の無効化に失敗しました - $0"
 ],
 "Successfully deleted replication agreement": [
  null,
  "レプリカ合意を削除しました"
 ],
 "Failed to delete replication agreement - $0": [
  null,
  "レプリカ合意の削除に失敗しました - $0"
 ],
 "Successfully created replication agreement": [
  null,
  "レプリカ合意を作成しました"
 ],
 "Failed to create replication agreement - $0": [
  null,
  "レプリカ合意の作成に失敗しました - $0"
 ],
 "Create Agreement": [
  null,
  "合意の作成"
 ],
 "Refresh Agreements": [
  null,
  "合意の更新"
 ],
 "Delete Replication Agreement": [
  null,
  "レプリカ合意の削除"
 ],
 "Are you sure you want to delete this replication agreement?": [
  null,
  "レプリカ合意を削除しますか？"
 ],
 "Deleting Replication Agreement ...": [
  null,
  "レプリカ合意を削除しています..."
 ],
 "Delete Agreement": [
  null,
  "合意の削除"
 ],
 "Initialize Replication Agreement": [
  null,
  "レプリカ合意の初期化"
 ],
 "Are you sure you want to initialize this replication agreement?": [
  null,
  "レプリカ合意を初期化しますか？"
 ],
 "Initializing Replication Agreement ...": [
  null,
  "レプリカ合意を初期化しています"
 ],
 "Initialize Agreement": [
  null,
  "合意の初期化"
 ],
 "Enable Replication Agreement": [
  null,
  "レプリカ合意を有効にする"
 ],
 "Are you sure you want to enable this replication agreement?": [
  null,
  "レプリカ合意を有効化しますか？"
 ],
 "Enabling ...": [
  null,
  "有効化しています..."
 ],
 "Enable Agreement": [
  null,
  "合意を有効にする"
 ],
 "Disable Replication Agreement": [
  null,
  "レプリカ合意を無効化"
 ],
 "Are you sure you want to disable this replication agreement?": [
  null,
  "レプリカ合意を無効化しますか？"
 ],
 "Disabling ...": [
  null,
  "無効化しています..."
 ],
 "Disable Agreement": [
  null,
  "合意の無効化"
 ],
 "  This requires a server restart to take effect": [
  null,
  "これを有効にするにはサーバを再起動する必要があります"
 ],
 "Error updating changelog configuration - $0": [
  null,
  "変更ログ設定の更新に失敗しました - $0"
 ],
 "Failed to reload changelog for \"$0\" - $1": [
  null,
  "\"$0\"の変更ログのリロードに失敗しました　 - $1"
 ],
 "Loading Changelog Configuration ...": [
  null,
  "変更ログ設定を読み込んでいます..."
 ],
 "Changelog trimming parameter.  Set the maximum number of changelog entries allowed in the database (nsslapd-changelogmaxentries).": [
  null,
  "変更ログの削除パラメータです。データベースに許可される変更ログエントリの最大数を設定します。 (nsslapd-changelogmaxentries)"
 ],
 "Changelog Maximum Entries": [
  null,
  "変更ログの最大エントリ数"
 ],
 "Changelog trimming parameter.  This set the maximum age of a changelog entry (nsslapd-changelogmaxage).": [
  null,
  "変更ログの削除パラメータです。変更ログエントリの最大保存期間を設定します。 (nsslapd-changelogmaxentries)"
 ],
 "Changelog Maximum Age": [
  null,
  "変更ログ最大保存期間"
 ],
 "Seconds": [
  null,
  "秒"
 ],
 "The changelog trimming interval.  Set how often the changelog checks if there are entries that can be purged from the changelog based on the trimming parameters (nsslapd-changelogtrim-interval).": [
  null,
  "変更ログの削除間隔です。削除パラメータに基づいて、変更ログから削除できるエントリがあるかどうかをチェックする頻度を設定します。 (nsslapd-changelogtrim-interval)"
 ],
 "Changelog Trimming Interval": [
  null,
  "変更ログ削除間隔"
 ],
 "Changelog Encryption": [
  null,
  "変更ログを暗号化する"
 ],
 "Changelog encryption requires that the server must already be configured for security/TLS.  This setting also requires that you export and import the changelog which must be done while the database is in read-only mode.  So first put the database into read-only mode, then export the changelog, enable changelog encryption, restart the server, import the changelog, and finally unset the database read-only mode.": [
  null,
  "変更ログを暗号化するには、サーバがすでにセキュリティ/TLSに設定されている必要があります。この設定では、データベースが読み取り専用モードの時に、変更ログをエクスポートおよびインポートする必要があります。まず、データベースを読み取り専用モードに設定し、次に変更ログをエクスポートして、変更ログの暗号化を有効にし、サーバを再起動し、変更ログをインポートし、最後にデータベースの読み取り専用モードを解除します。"
 ],
 "A Supplier replica requires a unique numerical identifier.  Please enter an ID between 1 and 65534": [
  null,
  "サプライヤーレプリカには固有の数値識別子が必要です。1から65534の間のIDを入力してください"
 ],
 "Successfully $0 replica to a $1": [
  null,
  "$1にレプリカを$0しました"
 ],
 "Failed to $0 replica - $1": [
  null,
  "レプリカの$0に失敗しました - $1"
 ],
 "Invalid DN for the Replication Manager: $0": [
  null,
  "レプリケーションマネージャのDNが無効です: $0"
 ],
 "You must provide a password for the Replication Manager": [
  null,
  "レプリケーションマネージャのパスワードを入力してください"
 ],
 "Passwords do not match": [
  null,
  "パスワードが一致しません"
 ],
 "Replication Manager": [
  null,
  "レプリケーションマネージャー"
 ],
 "Successfully added Replication Manager": [
  null,
  "レプリケーションマネージャーを追加しました"
 ],
 "Failure adding Replication Manager": [
  null,
  "レプリケーションマネージャーの追加に失敗しました"
 ],
 "Adding Replication Manager": [
  null,
  "レプリケーションマネージャーの追加中"
 ],
 "Successfully removed Replication Manager": [
  null,
  "レプリケーションマネージャーを削除しました"
 ],
 "Failure removing Replication Manager - $0": [
  null,
  "レプリケーションマネージャーの削除に失敗しました - $0"
 ],
 "Successfully updated replication configuration.": [
  null,
  "レプリケーション構成を更新しました"
 ],
 "Error updating replication configuration - $0": [
  null,
  "レプリケーション構成の更新に失敗しました - $0"
 ],
 "Demote this Supplier replica to a Hub or Consumer": [
  null,
  "このサプライヤーレプリカをハブまたはコンシューマーに降格します"
 ],
 "Change Role": [
  null,
  "ロールの変更"
 ],
 "Promote or Demote this Hub replica to a Supplier or Consumer": [
  null,
  "このハブレプリカをサプライヤーまたはコンシューマーに昇格または降格する"
 ],
 "Promote this Consumer replica to a Supplier or Hub": [
  null,
  "このコンシューマーレプリカをサプライヤーまたはハブに昇格する"
 ],
 "Replica Role": [
  null,
  "レプリカロール"
 ],
 "Replication Managers": [
  null,
  "レプリケーションマネージャ"
 ],
 "Add Replication Manager": [
  null,
  "レプリケーションマネージャーの追加"
 ],
 "The DN of the replication manager group": [
  null,
  "レプリケーションマネージャグループのDNを設定します。"
 ],
 "Bind DN Group": [
  null,
  "バインドDNグループ"
 ],
 "The interval to check for any changes in the group membership specified in the Bind DN Group and automatically rebuilds the list for the replication managers accordingly.  (nsds5replicabinddngroupcheckinterval).": [
  null,
  "バインドDNグループで指定されたグループメンバーシップの変更をチェックし、レプリケーションマネージャのリストを自動的に再構築する間隔を設定します。(nsds5replicabinddngroupcheckinterval)"
 ],
 "Bind DN Group Check Interval": [
  null,
  "バインドDNグループチェック間隔"
 ],
 "This controls the maximum age of deleted entries (tombstone entries), and entry state information.  (nsds5replicapurgedelay).": [
  null,
  "これは削除されたエントリ(tombstoneエントリ)とエントリの状態情報の最大保存期間を制御します。 (nsds5replicapurgedelay)"
 ],
 "Purge Delay": [
  null,
  "完全削除の遅延設定"
 ],
 "This attribute specifies the time interval in seconds between purge operation cycles.  (nsds5replicatombstonepurgeinterval).": [
  null,
  "この属性は、完全に削除するまでの時間間隔を秒単位で指定します。(nsds5replicatombstonepurgeinterval)"
 ],
 "Tombstone Purge Interval": [
  null,
  "Tombstoneの削除間隔"
 ],
 "A time limit (in seconds) that tells a replication session to yield if other replicas are trying to acquire this one (nsds5replicareleasetimeout).": [
  null,
  "別のレプリカでのレプリケーションセッション中は、同時に使用できないためロックされます。この設定は、他のレプリカがこのレプリカを取得しようとした場合に、レプリケーションセッションをリリースするまでの制限時間を秒単位で設定します。(nsds5replicareleasetimeout)"
 ],
 "Replica Release Timeout": [
  null,
  "レプリケーションリリースタイムアウト"
 ],
 "A timeout on how long to wait before stopping a replication session when the server is being stopped, replication is being disabled, or when removing a replication agreement. (nsds5replicaprotocoltimeout).": [
  null,
  "サーバが停止しレプリケーションが無効になっている場合、またはレプリケーション合意を削除する場合に、レプリケーションセッションを停止するまでのタイムアウトを設定します。(nsds5replicaprotocoltimeout)"
 ],
 "Replication Timeout": [
  null,
  "レプリケーションタイムアウト"
 ],
 "This is the minimum amount of time in seconds that a replication session will go into a backoff state  (nsds5replicabackoffmin).": [
  null,
  "レプリケーションセッションがバックオフ状態に入る最小時間を秒単位で設定します。(nsds5replicabackoffmin)"
 ],
 "Back Off Minimum": [
  null,
  "バックオフの最小時間"
 ],
 "This is the maximum amount of time in seconds that a replication session will go into a backoff state  (nsds5replicabackoffmax).": [
  null,
  "レプリケーションセッションがバックオフ状態に入る最大時間を秒単位で設定します。 (nsds5replicabackoffmax)"
 ],
 "Back Off Maximum": [
  null,
  "バックオフの最大時間"
 ],
 "The interval in seconds that the server will apply an internal update to get the RUV from getting stale. (nsds5replicakeepaliveupdateinterval).": [
  null,
  "RUVを最新の状態に保つために、サーバが内部的な更新を適用する間隔を秒単位で設定します。 (nsds5replicakeepaliveupdateinterval)"
 ],
 "Refresh RUV Interval": [
  null,
  "RUV更新間隔"
 ],
 "Enables faster tombstone purging (nsds5replicaprecisetombstonepurging).": [
  null,
  "tombstoneの削除を高速化します。(nsds5replicaprecisetombstonepurging)"
 ],
 "Fast Tombstone Purging": [
  null,
  "Tombstone削除を高速化する"
 ],
 "Remove Replication Manager": [
  null,
  "レプリケーションマネージャを削除"
 ],
 "Are you sure you want to remove this Replication Manager?": [
  null,
  "レプリケーションマネージャを削除しますか？"
 ],
 "Removing Manager ...": [
  null,
  "マネージャを削除しています..."
 ],
 "Remove Manager": [
  null,
  "マネージャーの削除"
 ],
 "Successfully started CleanAllRUV task": [
  null,
  "CleanAllRUVタスクを開始しました"
 ],
 "Failed to start CleanAllRUV task - $0": [
  null,
  "CleanAllRUVタスクの開始に失敗しました - $0"
 ],
 "Changelog was successfully initialized": [
  null,
  "変更ログを初期化しました。"
 ],
 "Error importing changelog LDIF - $0": [
  null,
  "変更ログLDIFファイルのインポートに失敗しました - $0"
 ],
 "Changelog was successfully exported": [
  null,
  "変更ログをエクスポートしました。"
 ],
 "LDAP URL": [
  null,
  "LDAP URL"
 ],
 "Min CSN": [
  null,
  "最小CSN"
 ],
 "There is no local RUV, the database might not have been initialized yet.": [
  null,
  "ローカルRUVが存在しないため、データベースがまだ初期化されていない可能性があります。"
 ],
 "Local RUV": [
  null,
  "ローカルRUV"
 ],
 "Refresh the RUV for this suffix": [
  null,
  "このサフィックスのRUVを更新します"
 ],
 "Remote RUV's": [
  null,
  "リモートRUV"
 ],
 "Refresh the remote RUVs for this suffix": [
  null,
  "このサフィックスのリモートRUVを更新します"
 ],
 "Replication Change Log Tasks": [
  null,
  "レプリケーション変更ログタスク"
 ],
 "Export the changelog to an LDIF file.  Typically used for changelog encryption purposes, or debugging.": [
  null,
  "変更ログをLDIFファイルにエクスポートします。通常は変更ログの暗号化やデバッグに使用されます。"
 ],
 "Export Changelog": [
  null,
  "変更ログのエクスポート"
 ],
 "Export the replication changelog to a LDIF file.  Used for preparing to encrypt the changelog, or simply for debugging.": [
  null,
  "レプリケーション変更ログをLDIFファイルにエクスポートします。変更ログの暗号化のための準備や単純なデバッグに使用されます。"
 ],
 "Initialize the changelog with an LDIF file for changelog encryption purposes.": [
  null,
  "変更ログの暗号化のために、LDIFファイルで変更ログを初期化します。"
 ],
 "Import Changelog": [
  null,
  "変更ログのインポート"
 ],
 "Initialize the replication changelog from an LDIF file.  Used to initialize the change log after encryption has been enabled.": [
  null,
  "LDIFファイルからレプリケーション変更ログを初期化します。暗号化が有効になった後に変更ログを初期化するために使用されます。"
 ],
 "Remove RUV Element (CleanAllRUV)": [
  null,
  "RUV要素の削除(CleanAllRUV)"
 ],
 "Are you sure you want to attempt to clean this Replica ID from the suffix?": [
  null,
  "このレプリカIDをサフィックスから削除しますか？"
 ],
 "Starting cleaning task (CleanAllRUV) ...": [
  null,
  "クリーニングタスク(CleanAllRUV)の開始しました..."
 ],
 "Remove RUV Element": [
  null,
  "RUV要素の削除"
 ],
 "Initialize Replication Changelog From LDIF": [
  null,
  "LDIFからレプリケーション変更ログを初期化"
 ],
 "Are you sure you want to attempt to initialize the changelog from LDIF?  This will reject all operations during during the initialization.": [
  null,
  "LDIFから変更ログを初期化しますか？初期化中はすべての操作が拒否されます。"
 ],
 "Initialzing Replication Change Log ...": [
  null,
  "レプリケーション変更ログを初期化しています..."
 ],
 "Import Changelog LDIF": [
  null,
  "変更ログのLDIFインポート"
 ],
 "Successfully updated winsync agreement": [
  null,
  "winsync合意を更新しました"
 ],
 "Failed to update winsync agreement": [
  null,
  "winsync合意の更新に失敗しました"
 ],
 "Successfully poked winsync agreement": [
  null,
  "winsync合意の検証に成功しました"
 ],
 "Failed to poke winsync agreement - $0": [
  null,
  "winsync合意の検証に失敗しました - $0"
 ],
 "Failed to initialize winsync agreement - $0": [
  null,
  "winsync合意の初期化に失敗しました - $0"
 ],
 "Successfully enabled winsync agreement": [
  null,
  "winsync合意を有効化しました"
 ],
 "Failed to enabled winsync agreement - $0": [
  null,
  "winsync合意の有効化に失敗しました - $0"
 ],
 "Successfully disabled winsync agreement": [
  null,
  "winsync合意を無効化しました"
 ],
 "Failed to disable winsync agreement - $0": [
  null,
  "winsync合意の無効化に失敗しました - $0"
 ],
 "Successfully deleted winsync agreement": [
  null,
  "winsync合意を削除しました"
 ],
 "Failed to delete winsync agreement - $0": [
  null,
  "winsync合意の削除に失敗しました - $0"
 ],
 "Successfully created winsync agreement": [
  null,
  "winsync合意を作成しました"
 ],
 "Failed to create winsync agreement": [
  null,
  "winsync合意の作成に失敗しました"
 ],
 "Create winsync agreement": [
  null,
  "winsync合意を作成"
 ],
 "Delete Winsync Agreement": [
  null,
  "Winsync合意を削除"
 ],
 "Are you sure you want to delete this winsync agreement": [
  null,
  "Winsync合意を削除しますか？"
 ],
 "Deleting Winsync Agreement ...": [
  null,
  "Winsync合意を削除しています..."
 ],
 "Initialize Winsync Agreement": [
  null,
  "Winsync合意の初期化"
 ],
 "Are you sure you want to initialize this winsync agreement?": [
  null,
  "Winsync合意を初期化しますか？"
 ],
 "Initializing Winsync Agreement ...": [
  null,
  "Winsync合意を初期化しています..."
 ],
 "Enable Winsync Agreement": [
  null,
  "Winsync合意を有効にする"
 ],
 "Are you sure you want to enable this winsync agreement?": [
  null,
  "Winsync合意を有効化しますか？"
 ],
 "Enabling Winsync Agreement ...": [
  null,
  "Winsync合意を有効化しています..."
 ],
 "Disable Winsync Agreement": [
  null,
  "Winsync合意の無効化"
 ],
 "Are you sure you want to disable this winsync agreement?": [
  null,
  "Winsync合意を無効化しますか？"
 ],
 "Disabling Winsync Agreement ...": [
  null,
  "Winsync合意を無効化しています..."
 ],
 "Host": [
  null,
  "ホスト"
 ],
 "State": [
  null,
  "状態"
 ],
 "Last Init Status": [
  null,
  "前回の初期化ステータス"
 ],
 "Edit Agreement": [
  null,
  "合意の編集"
 ],
 "Poke Agreement": [
  null,
  "合意の検証"
 ],
 "Disable/Enable Agreement": [
  null,
  "合意の無効化/有効化"
 ],
 "Initialized": [
  null,
  "初期化しました"
 ],
 "Not Initialized": [
  null,
  "初期化されていません"
 ],
 "Initializing": [
  null,
  "初期化中"
 ],
 "Search agreements": [
  null,
  "合意を検索"
 ],
 "No Replication Managers": [
  null,
  "レプリケーションマネージャーがいません"
 ],
 "Delete Replication Manager": [
  null,
  "レプリケーションマネージャーを削除"
 ],
 "Replica LDAP URL": [
  null,
  "レプリカLDAP URL"
 ],
 "No RUV's": [
  null,
  "RUVがありません"
 ],
 "Remove this RUV/Replica ID from all the Replicas.": [
  null,
  "このRUV/レプリカIDをすべてのレプリカから削除します。"
 ],
 "Clean RUV": [
  null,
  "RUVの一掃"
 ],
 "Create": [
  null,
  "作成"
 ],
 "Save Agreement": [
  null,
  "合意を保存"
 ],
 "Saving Agreement ...": [
  null,
  "合意を保存しています..."
 ],
 "Consumer Initialization": [
  null,
  "コンシューマの初期化"
 ],
 "Do Not Initialize": [
  null,
  "初期化しない"
 ],
 "Do Online Initialization": [
  null,
  "オンライン初期化を行う"
 ],
 "Days To Send Synchronization Updates": [
  null,
  "同期/更新の実施日"
 ],
 "Time to start initiating replication sessions": [
  null,
  "レプリケーションセッションの開始時間を設定します"
 ],
 "Replication Start Time": [
  null,
  "レプリケーション開始時間"
 ],
 "Start time must be before the End time": [
  null,
  "開始時間は終了時間より前でなければなりません"
 ],
 "Time to initiating replication sessions": [
  null,
  "レプリケーションセッションの終了時間を設定します"
 ],
 "Replication End Time": [
  null,
  "レプリケーション終了時間"
 ],
 "End time must be after the Start time": [
  null,
  "終了時間は開始時間より後でなければなりません"
 ],
 "$0 Winsync Agreement": [
  null,
  "Winsync合意を$0"
 ],
 "Main Settings": [
  null,
  "メイン設定"
 ],
 "Windows AD Host": [
  null,
  "WindowsADホスト"
 ],
 "Windows AD Port": [
  null,
  "WindowsADポート"
 ],
 "Bind Password": [
  null,
  "バインドパスワード"
 ],
 "Passwords must match": [
  null,
  "パスワードが一致しません"
 ],
 "Domain & Content": [
  null,
  "ドメイン&コンテンツ"
 ],
 "Windows Domain Name": [
  null,
  "Windowsドメイン名"
 ],
 "The Active Directory subtree to synchronize": [
  null,
  "同期するActiveDirectoryサブツリーを設定します"
 ],
 "Windows Subtree": [
  null,
  "Windowsサブツリー"
 ],
 "e.g. cn=Users,dc=domain,dc=com": [
  null,
  "例: cn=Users,dc=domain,dc=com"
 ],
 "Directory Server subtree to synchronize": [
  null,
  "同期するディレクトリサーバのサブツリーを設定します"
 ],
 "DS Subtree": [
  null,
  "DSサブツリー"
 ],
 "e.g. ou=People,dc=domain,dc=com": [
  null,
  "例: ou=People,dc=domain,dc=com"
 ],
 "Connection Protocol": [
  null,
  "接続プロトコル"
 ],
 "LDAPS": [
  null,
  "LDAPS"
 ],
 "StartTLS": [
  null,
  "StartTLS"
 ],
 "Synchronization Direction": [
  null,
  "同期方向"
 ],
 "Synchronization in both directions (default behavior).": [
  null,
  "双方向に同期します。(デフォルトの動作)"
 ],
 "both": [
  null,
  "both"
 ],
 "Only synchronize Directory Server updates to Windows.": [
  null,
  "ディレクトリサーバの更新のみをWindowsに同期します。"
 ],
 "toWindows": [
  null,
  "toWindows"
 ],
 "Only synchronize Windows updates to Directory Server.": [
  null,
  "Windowsの更新のみをディレクトリサーバに同期します。"
 ],
 "fromWindows": [
  null,
  "fromWindows"
 ],
 "The interval to check for updates on Windows.  Default is 300 seconds": [
  null,
  "Windowsの更新を確認する間隔を設定します。デフォルトは300秒です。"
 ],
 "Synchronization Interval": [
  null,
  "同期間隔"
 ],
 "Attribute to exclude from replication": [
  null,
  "レプリケーションから除外する属性を指定します。"
 ],
 "Exclude Attributes": [
  null,
  "除外属性"
 ],
 "Start typing an attribute...": [
  null,
  "属性を入力してください..."
 ],
 "Synchronize New Windows Groups": [
  null,
  "新しいWindowsグループを同期する"
 ],
 "Synchronize New Windows Users": [
  null,
  "新しいWindowsユーザを同期する"
 ],
 "Scheduling": [
  null,
  "スケジュール"
 ],
 "By default replication updates are sent to the replica as soon as possible, but if there is a need for replication updates to only be sent on certain days and within certain windows of time then you can setup a custom replication schedule.": [
  null,
  "通常、レプリケーションの更新はできるだけ早くレプリカに送信されますが、特定の日や時間帯にのみレプリケーション更新を送信する必要がある場合は、カスタムレプリケーションスケジュールを設定できます。"
 ],
 "Use A Custom Schedule": [
  null,
  "カスタムスケジュールを使用する"
 ],
 "If you are using Bind Group's on the consumer replica you can configure bootstrap credentials that can be used to do online initializations, or bootstrap a session if the bind groups get out of synchronization": [
  null,
  "コンシューマーレプリカでバインドグループを使用している場合、オンライン初期化を行うために使用できるブートストラップ認証情報を設定することができます。また、バインドグループが同期しなくなった場合にセッションをブートストラップすることもできます。"
 ],
 "The Bind DN the agreement can use to bootstrap initialization": [
  null,
  "ブートストラップ初期化のために合意が使用できるバインドDNを設定します"
 ],
 "The Bind DN password for bootstrap initialization": [
  null,
  "ブートストラップ初期化のためのバインドDNのパスワードを設定します"
 ],
 "Password must match": [
  null,
  "パスワードが一致している必要があります"
 ],
 "Confirm the Bind DN password for bootstrap initialization": [
  null,
  "ブートストラップ初期化のためのバインドDNパスワードを再入力してください"
 ],
 "The connection protocol for bootstrap initialization": [
  null,
  "ブートストラップ初期化のための接続プロトコルを設定します"
 ],
 "LDAP": [
  null,
  "LDAP"
 ],
 "STARTTLS": [
  null,
  "STARTTLS"
 ],
 "The authentication method for bootstrap initialization": [
  null,
  "ブートストラップ初期化のための認証方法を設定します"
 ],
 "Authentication Method": [
  null,
  "認証方式"
 ],
 "Days To Send Replication Updates": [
  null,
  "レプリケーションの更新を送信する日数"
 ],
 "$0 Replication Agreement": [
  null,
  "レプリカ合意を$0"
 ],
 "Consumer Host": [
  null,
  "コンシューマーホスト"
 ],
 "Consumer Port": [
  null,
  "コンシューマーポート"
 ],
 "Port must be between 1 and 65535": [
  null,
  "ポート番号は1から65535の範囲内で設定してください"
 ],
 "Fractional Settings": [
  null,
  "フラクショナル設定"
 ],
 "Attribute to exclude from replica Initializations": [
  null,
  "レプリカの初期化から除外する属性を設定します。"
 ],
 "Exclude Init Attributes": [
  null,
  "初期化から除外する属性"
 ],
 "Attributes to strip from a replication update": [
  null,
  "レプリケーションの更新から削除する属性を設定します。"
 ],
 "Strip Attributes": [
  null,
  "削除する属性"
 ],
 "Bootstrap Settings": [
  null,
  "ブートストラップ設定"
 ],
 "Enable Bootstrap Settings": [
  null,
  "ブートストラップ設定を有効にする"
 ],
 "Supplier Replica Identifier.  This must be unique across all the Supplier replicas in your environment": [
  null,
  "サプライヤーレプリカの識別子。環境内のすべてのサプライヤーレプリカで一意である必要があります。"
 ],
 "Change Replica Role": [
  null,
  "レプリカロールの変更"
 ],
 "Please choose the new replication role you would like for this suffix": [
  null,
  "このサフィックスの新しいレプリカロールを選択してください。"
 ],
 "New Role": [
  null,
  "新しいロール"
 ],
 "Adding Replication Manager ...": [
  null,
  "レプリケーションマネージャーを追加しています..."
 ],
 "Create a Replication Manager entry, and add it to the replication configuration for this suffix.  If the entry already exists it will be overwritten with the new credentials.": [
  null,
  "レプリケーションマネージャーエントリを作成し、このサフィックスのレプリケーション構成に追加します。エントリがすでに存在する場合は、新しい認証情報で上書きされます。"
 ],
 "The DN of the replication manager": [
  null,
  "レプリケーションマネージャのDNを設定します"
 ],
 "Replication Manager DN": [
  null,
  "レプリケーションマネージャDN"
 ],
 "Replication Manager password": [
  null,
  "レプリケーションマネージャパスワードを設定します"
 ],
 "Enable Replication": [
  null,
  "レプリケーションを有効にする"
 ],
 "Enabling Replication ...": [
  null,
  "レプリケーションを有効化しています..."
 ],
 "Choose the replication role for this suffix.  If it is a Supplier replica then you must pick a unique ID to identify it among the other Supplier replicas in your environment.  The replication changelog will also automatically be created for you.": [
  null,
  "このサフィックスのレプリケーションロールを選択ししてくださ。サプライヤーレプリカの場合は、環境内の他のサプライヤーレプリカと区別するために一意のIDを選択する必要があります。レプリケーション変更ログも自動的に作成されます。:"
 ],
 "Replication Role": [
  null,
  "レプリケーションロール"
 ],
 "Hub": [
  null,
  "ハブ"
 ],
 "Consumer": [
  null,
  "コンシューマー"
 ],
 "You can optionally define the authentication information for this replicated suffix.  Either a Manager DN and Password, a Bind Group DN, or both, can be provided.  The Manager DN should be an entry under \"cn=config\" and if it does not exist it will be created, while the Bind Group DN is usually an existing group located in the database suffix.  Typically, just the Manager DN and Password are used when enabling replication for a suffix.": [
  null,
  "このレプリケーションされたサフィックスの認証情報をオプションで定義できます。マネージャーDNとパスワード、バインドグループDN、または両方を指定できます。マネージャーDNは\"cn=config\"の下にあるエントリである必要があり、存在しない場合は作成されます。一方、バインドグループDNは通常、データベースサフィックス内に存在する既存のグループです。サフィックスのレプリケーションを有効にする場合には、通常、マネージャーDNとパスワードだけが使用されます。"
 ],
 "The DN of the replication manager.  If you supply a password the entry will be created in the server (it will also overwrite the entry is it already exists).": [
  null,
  "レプリケーションマネージャのDNを設定します。パスワードを指定すると、サーバにエントリが作成されます。(すでに存在する場合は上書きされます。)"
 ],
 "Confirm the Replication Manager password": [
  null,
  "レプリケーションマネージャのパスワードを再入力してください"
 ],
 "The DN of a group that contains users that can perform replication updates": [
  null,
  "レプリケーション更新を実行できるユーザが含まれるグループのDNを設定します"
 ],
 "Bind Group DN": [
  null,
  "バインドグループDN"
 ],
 "This will export the changelog to the server's LDIF directory.  This is the only LDIF file that can be imported into the server for enabling changelog encryption.  Do not edit or rename the file.": [
  null,
  "この操作により、変更ログがサーバのLDIFディレクトリにエクスポートされます。変更ログの暗号化を有効にするためにサーバにインポートできる唯一のLDIFファイルです。ファイルを編集したり名前を変更しないでください。"
 ],
 "The LDIF file that is generated should <b>not</b> be used to initialize the Replication Changelog.  It is only meant for debugging/investigative purposes.": [
  null,
  "生成されたLDIFファイルは、レプリケーション変更ログの初期化には使用しないでください。デバッグや調査のみを目的として使用することができます。"
 ],
 "Decode base64 changes": [
  null,
  "Base64の変更をデコード"
 ],
 "Only Export CSN's": [
  null,
  "CSNのみをエクスポート"
 ],
 "Create Replication Change Log LDIF File": [
  null,
  "レプリケーション変更ログLDIFファイルの作成"
 ],
 "Export to LDIF For Reinitializing The Changelog": [
  null,
  "変更ログを再初期化するためのLDIFファイルにエクスポート"
 ],
 "Export to LDIF For Debugging": [
  null,
  "デバッグのためのLDIFファイルにエクスポート"
 ],
 "The Bind DN is not a valid DN (Distinguished Name) $0": [
  null,
  "バインドDNが有効なDN(Distinguished Name)ではありません $0"
 ],
 "The Group DN is not a valid DN (Distinguished Name)": [
  null,
  "グループDNが有効なDN(Distinguished Name)ではありません"
 ],
 "The Bind DN passwords do not match": [
  null,
  "バインドDNパスワードが一致しません"
 ],
 "The Replica ID is not in the valid range of 1 - 65534": [
  null,
  "レプリカIDは1から65534の有効な範囲内で設定してください"
 ],
 "Successfully enabled replication for $0": [
  null,
  "$0のレプリケーションを有効にしました"
 ],
 "Failed to enable replication for $0": [
  null,
  "$0のレプリケーションの有効化に失敗しました"
 ],
 "Successfully disabled replication for $0": [
  null,
  "$0のレプリケーションを無効にしました"
 ],
 "Failed to disable replication for $0 - $1": [
  null,
  "$0のレプリケーションの無効化に失敗しました - $1"
 ],
 "Configuration": [
  null,
  "設定"
 ],
 "Agreements ": [
  null,
  "合意"
 ],
 "Winsync Agreements ": [
  null,
  "Winsync合意"
 ],
 "Change Log": [
  null,
  "変更ログ"
 ],
 "RUV's & Tasks": [
  null,
  "RUV&タスク"
 ],
 "Disable replication, and remove all replication agreements.": [
  null,
  "レプリケーションを無効にし、すべてのレプリカ合意を削除します。"
 ],
 "Disable": [
  null,
  "無効"
 ],
 "Replication is not enabled for this suffix": [
  null,
  "このサフィックスでレプリケーションが有効になっていません"
 ],
 "Refresh replication settings for this suffix": [
  null,
  "このサフィックスのレプリケーション設定を更新します"
 ],
 "Disable Replication": [
  null,
  "レプリケーションを無効にする"
 ],
 "Are you sure you want to disable replication for this suffix?": [
  null,
  "このサフィックスのレプリケーションを無効にしますか？"
 ],
 "Disabling Replication ...": [
  null,
  "レプリケーションを無効にしています..."
 ],
 "Loading ...": [
  null,
  "読み込み中 ..."
 ],
 "Entry DN": [
  null,
  "エントリDN"
 ],
 "Child Entries": [
  null,
  "子エントリ"
 ],
 "Last Modified": [
  null,
  "最終更新"
 ],
 "This is a hidden LDAP subentry": [
  null,
  "これは非表示のLDAPサブエントリです"
 ],
 "Error searching the database: $0": [
  null,
  "データベースの検索中にエラーが発生しました: $0"
 ],
 "Lock ...": [
  null,
  "ロック ..."
 ],
 "Unlock ...": [
  null,
  "ロック解除 ..."
 ],
 "Search ...": [
  null,
  "検索 ..."
 ],
 "Edit ...": [
  null,
  "編集 ..."
 ],
 "New ...": [
  null,
  "新規作成 ..."
 ],
 "ACIs ...": [
  null,
  "ACIs"
 ],
 "Class of Service ...": [
  null,
  "サービスクラス ..."
 ],
 "No Search Results": [
  null,
  "検索結果はありません"
 ],
 "Search Database": [
  null,
  "データベースを検索"
 ],
 "No databases": [
  null,
  "データベースがありません"
 ],
 "Text that will be used with pre-selected attributes to find matching entries.": [
  null,
  "事前に選択された属性と共に使用されるテキストで、一致するエントリを見つけるために使用されます。"
 ],
 "Text": [
  null,
  "テキスト"
 ],
 "Specific LDAP search filter for finding entries.": [
  null,
  "エントリを検索するための特定のLDAP検索フィルタです。"
 ],
 "Enter search text ...": [
  null,
  "テキストでエントリを検索"
 ],
 "Enter an LDAP search filter ...": [
  null,
  "フィルタでLDAPエントリを検索"
 ],
 "Hide Search Criteria": [
  null,
  "検索条件を非表示"
 ],
 "Show Search Criteria": [
  null,
  "検索条件を表示"
 ],
 "searchBase": [
  null,
  "検索ベース"
 ],
 "Invalid DN syntax": [
  null,
  "無効なDN構文"
 ],
 "Search for entries starting at the search base, and including all its child entries": [
  null,
  "検索ベースから始まり、その子エントリを含むエントリを検索します。"
 ],
 "One Level": [
  null,
  "One Level"
 ],
 "Search for entries starting at the search base, and include only the first level of child entries": [
  null,
  "検索ベースから始まり、子エントリの最初のレベルのみを含むエントリを検索します。"
 ],
 "Base": [
  null,
  "Base"
 ],
 "Search for an exact entry (search base). This does not include child entries.": [
  null,
  "検索ベースと完全に一致するエントリ(子エントリを除く)を検索します。"
 ],
 "Search timeout in seconds": [
  null,
  "検索のタイムアウトを秒単位で指定します。"
 ],
 "Check if the search result entries are locked, and add Lock/Unlock options to the dropdown. This setting vastly impacts the search's performance. Use only when needed.": [
  null,
  "検索結果エントリがロックされているかどうかをチェックし、ドロップダウンにロック/アンロックのオプションを追加します。この設定は検索のパフォーマンスに大きく影響しますので、必要な場合にのみ使用してください。"
 ],
 "Show Locking": [
  null,
  "ロックの表示"
 ],
 "Only used for Search Text based queries.  The selected attributes will use Search Text as the attribute value in the search filter": [
  null,
  "Search Textベースのクエリにのみ使用されます。選択した属性は、検索フィルタ内の検索テキストを属性値として使用します。"
 ],
 "Search Attributes": [
  null,
  "属性を検索"
 ],
 "cn": [
  null,
  "cn"
 ],
 "uid": [
  null,
  "uid"
 ],
 "sn": [
  null,
  "sn"
 ],
 "givenName": [
  null,
  "givenName"
 ],
 "mail": [
  null,
  "mail"
 ],
 "displayName": [
  null,
  "displayName"
 ],
 "legalName": [
  null,
  "legalName"
 ],
 "memberOf": [
  null,
  "memberOf"
 ],
 "member": [
  null,
  "メンバー"
 ],
 "uniqueMember": [
  null,
  "uniqueMember"
 ],
 "Enter space or comma separated list of attributes to search.": [
  null,
  "検索する属性のリストをスペースまたはカンマで区切って入力してください。"
 ],
 "Type attributes to include in the filter ...": [
  null,
  "フィルタに含める属性を入力してください ..."
 ],
 "Searching": [
  null,
  "検索"
 ],
 "Results:": [
  null,
  "結果:"
 ],
 "Are you sure you want to $0 the $1?": [
  null,
  "本当に$1を$0しますか？"
 ],
 "Confirm": [
  null,
  "確認"
 ],
 "No entry is selected": [
  null,
  "エントリが選択されていません"
 ],
 "Select an entry to see its details.": [
  null,
  "エントリの詳細を表示するためにエントリを選択してください。"
 ],
 "Value": [
  null,
  "値"
 ],
 "MESSAGE": [
  null,
  "メッセージ"
 ],
 ":ENTRY TOO LARGE - OUTPUT TRUNCATED!": [
  null,
  ": エントリが大きすぎるため、出力が切り捨てられました。"
 ],
 "Rename ...": [
  null,
  "名前変更 ..."
 ],
 "Delete ...": [
  null,
  "削除 ..."
 ],
 "Suffix Trees": [
  null,
  "サフィックスツリー"
 ],
 "Suffix Tree": [
  null,
  "サフィックスツリー"
 ],
 "Refresh all the suffixes": [
  null,
  "すべてのサフィックスを更新する"
 ],
 "Suffix creation": [
  null,
  "サフィックスを作成"
 ],
 "Entry Details": [
  null,
  "エントリの詳細"
 ],
 "Reload the LDAP entry.": [
  null,
  "LDAPエントリを更新"
 ],
 "Entry is too large - Table is truncated.": [
  null,
  "エントリが大きすぎるため、テーブルが切り捨てられました。"
 ],
 "Empty suffix!": [
  null,
  "サフィックスが空です！"
 ],
 "The suffix is configured, but it has no entries.": [
  null,
  "サフィックスは設定されていますが、エントリがありません。"
 ],
 "Create the root suffix entry": [
  null,
  "ルートサフィックスのエントリを作成"
 ],
 "Last Modified Time": [
  null,
  "最終更新日時"
 ],
 "Last Refresh at": [
  null,
  "最後の更新"
 ],
 "required": [
  null,
  "必須"
 ],
 "This field is required": [
  null,
  "この項目は必須です"
 ],
 "Value is too large to display": [
  null,
  "値が大きすぎて表示できません"
 ],
 "Empty value!": [
  null,
  "値が空です！"
 ],
 "Naming Attribute": [
  null,
  "命名属性"
 ],
 "This attribute and value are part of the entry's DN and can not be changed except by doing a Rename (modrdn) operation on the entry.": [
  null,
  "この属性と値はエントリのDNの一部であり、エントリをリネーム(modrdn)操作を行わない限り、変更することはできません。"
 ],
 "Set as Naming Attribute": [
  null,
  "命名属性として設定"
 ],
 "Add Another Value": [
  null,
  "別の値を追加する"
 ],
 "Remove this Row": [
  null,
  "この列を削除する"
 ],
 "Required Attribute": [
  null,
  "必須属性"
 ],
 "DN Renaming": [
  null,
  "DNをリネームする"
 ],
 "You cannot modify the RDN ( Relative Distinguished Name ) in the edit mode.": [
  null,
  "クイック更新モードでは、RDN(相対識別名)を変更することはできません。"
 ],
 "If you want to modify the RDN, please use the <Rename action option": [
  null,
  "RDNを変更する場合は、クイック更新アクションオプションを使用してください。"
 ],
 "Set Password": [
  null,
  "パスワードの設定"
 ],
 "passwordField": [
  null,
  "パスワードフィールド"
 ],
 "Upload a file local to the browser.": [
  null,
  "ローカルファイルをブラウザにアップロードします。"
 ],
 "Select a file from the machine on which the browser was launched.": [
  null,
  "ブラウザを起動したマシンから、ファイルを選択します。"
 ],
 "Write the complete file path.": [
  null,
  "完全なファイルパスを記入してください。"
 ],
 "Type the full path of the file on the server host the LDAP server.": [
  null,
  "LDAPサーバがホストされているサーバホスト上のファイルの完全なパスを入力してください。"
 ],
 "There was an issue with the uploaded file ( incorrect type? ).": [
  null,
  "アップロードされたファイルに問題がありました(不正なタイプ？)"
 ],
 "The file size larger than 128 MB. Use the file path option.": [
  null,
  "ファイルサイズが128 MBを超えています。ファイルパスオプションを使用してください。"
 ],
 "The certificate must be stored in the Distinguished Encoding Rules (DER) format.": [
  null,
  "証明書は、Distinguished Encoding Rules(DER)形式で保存する必要があります。"
 ],
 "Choose Binary File": [
  null,
  "バイナリファイルを選択"
 ],
 "selected": [
  null,
  "選択済み"
 ],
 "Loading LDAP entries...": [
  null,
  "LDAPエントリを読み込んでいます ..."
 ],
 "Error searching database - $0": [
  null,
  "データベースの検索中にエラーが発生しました - $0"
 ],
 "Encountered an error, unable to display child entries": [
  null,
  "エラーが発生し、子エントリを表示することができません"
 ],
 "Loading...": [
  null,
  "読み込み中..."
 ],
 "No Databases": [
  null,
  "データベースがありません"
 ],
 "The suffix $0 was successfully created.": [
  null,
  "サフィックス$0の作成に成功しました。"
 ],
 "Failed to create the suffix $0.": [
  null,
  "サフィックス$0の作成に失敗しました。"
 ],
 "Create Root Entry": [
  null,
  "ルートエントリの作成"
 ],
 "LDIF data to create the suffix": [
  null,
  "サフィックスを作成するためのLDIFデータ"
 ],
 "Failed to generate the LDIF data": [
  null,
  "LDIFデータの作成に失敗しました"
 ],
 "Size limit of $0 was exceeded. The child entries of $1 have been truncated.": [
  null,
  "$0のサイズ制限を超過しました。$1の子エントリは切り捨てられました。"
 ],
 "Size limit of $0 was exceeded.  The child entries of $1 have been truncated. Use the 'Search' feature if you want to adjust the size limit and retrieve more entries": [
  null,
  "$0のサイズ制限を超過しました。$1の子エントリは切り捨てられました。さらにエントリを取得するためにサイズ制限を調整したい場合は、「検索」機能を使用してください。"
 ],
 "Get Started": [
  null,
  "開始"
 ],
 "Requirement for a root entry creation": [
  null,
  "ルートエントリ作成時の条件"
 ],
 "The object class must contain the attribute used to name the suffix.": [
  null,
  "オブジェクトクラスには、サフィックスの命名に使用される属性が含まれていなければなりません。"
 ],
 "For example, if the entry corresponds to the suffix <code>ou=people,dc=example,dc=com</code>, then choose the<code>organizationalUnit</code>object class or another object class that allows the <code>ou</code> attribute.": [
  null,
  "例えば、エントリがサフィックスou=people, dc=example,dc=comに対応している場合、organizationalUnitオブジェクトクラスまたはou属性を許可する他のオブジェクトクラスを選択してください。"
 ],
 "The root entry to create is $0": [
  null,
  "作成するルートエントリは $0 です。"
 ],
 "Make sure to select an <strong>ObjectClass</strong>that allows or requires the attribute": [
  null,
  "必要な属性を許可または要求するObjectClassを選択してください。"
 ],
 "Create a new User": [
  null,
  "新しいユーザを作成"
 ],
 "Add a new User (inetOrgPerson objectClass)": [
  null,
  "新しいユーザを追加(inetOrgPersonオブジェクトクラス)"
 ],
 "Create a new Group": [
  null,
  "新しいグループを作成"
 ],
 "Add a new Group (GroupOfNames/GroupOfUniqueNames objectClass)": [
  null,
  "新しいグループを追加(GroupOfNames/GroupOfUniqueNamesオブジェクトクラス)。"
 ],
 "Create a new Organizational Unit": [
  null,
  "新しい組織ユニットを作成"
 ],
 "Add a new Organizational Unit": [
  null,
  "新しい組織ユニットを追加"
 ],
 "Create a new Role": [
  null,
  "新しいロールを作成"
 ],
 "Add a new Role (Filtered / Managed / Nested)": [
  null,
  "新しいロールを追加(フィルタリングされたロール/管理されたロール/ネストされたロール)"
 ],
 "Create a new custom Entry": [
  null,
  "新しいカスタムエントリを作成"
 ],
 "Add a new entry by selecting ObjectClasses and Attributes": [
  null,
  "オブジェクトクラスと属性を選択して新しいエントリを追加"
 ],
 "ACI Name": [
  null,
  "ACI名"
 ],
 "ACI Rule Type": [
  null,
  "ACIルールタイプ"
 ],
 "Edit ACI": [
  null,
  "ACI編集"
 ],
 "Remove ACI": [
  null,
  "ACI削除"
 ],
 "Successfully replace ACI": [
  null,
  "ACIの置き換えに成功しました"
 ],
 "Failed to update ACI: error $0 - $1": [
  null,
  "ACIの置き換えに失敗しました: エラー $0 - $1"
 ],
 "Successfully removed ACI": [
  null,
  "ACIの削除に成功しました"
 ],
 "Failed to remove ACI: error $0 - $1": [
  null,
  "ACIの削除に失敗しました: エラー $0 - $1"
 ],
 "Successfully added ACI": [
  null,
  "ACIの追加に成功しました"
 ],
 "Failed to add ACI: error $0 - $1": [
  null,
  "ACIの追加に失敗しました: エラー $0 - $1"
 ],
 "Save ACI": [
  null,
  "ACIを保存"
 ],
 "Saving ACI ...": [
  null,
  "ACIを保存しています ..."
 ],
 "Manage ACI's For ": [
  null,
  "ACIの管理"
 ],
 "No ACI's": [
  null,
  "ACIがありません"
 ],
 "Access Control Instructions": [
  null,
  "アクセス制御命令"
 ],
 "Add ACI Wizard": [
  null,
  "ACI追加ウィザード"
 ],
 "Add ACI Manually": [
  null,
  "手動でACIを追加"
 ],
 "Delete ACI": [
  null,
  "ACIを削除"
 ],
 "Are you sure you want to delete this ACI?": [
  null,
  "本当にこのACIを削除しますか？"
 ],
 "Reset ACI": [
  null,
  "ACIをリセット"
 ],
 "Enter ACI ...": [
  null,
  "ACIを入力してください..."
 ],
 "Select CoS Type": [
  null,
  "CoSタイプを選択"
 ],
 "A class of service definition (CoS) shares attributes between entries in a way that is transparent to applications. CoS simplifies entry management and reduces storage requirements.": [
  null,
  "クラスオブサービス定義(CoS)は、エントリ間で属性を共有し、アプリケーションにとって透明な方法で動作します。CoSはエントリの管理を簡素化し、ストレージ要件を削減します。"
 ],
 "Clients of the Directory Server read the attributes in a user's entry. With CoS, some attribute values may not be stored within the entry itself. Instead, these attribute values are generated by class of service logic as the entry is sent to the client application.": [
  null,
  "ディレクトリサーバのクライアントは、ユーザのエントリ内の属性を読み取ります。しかし、CoSでは、一部の属性値はエントリ自体に格納されていない場合があります。代わりに、これらの属性値はクラスオブサービスのロジックによって生成され、エントリがクライアントアプリケーションに送信される際に提供されます。"
 ],
 "Create a new CoS Definition": [
  null,
  "新しいCoSの定義を作成します。"
 ],
 "The CoS definition entry identifies the type of CoS used. The CoS definition entry is below the branch at which it is effective.": [
  null,
  "CoSの定義エントリは、使用されるCoSのタイプを識別します。CoSの定義エントリは、それが有効な枝の下に配置されます。"
 ],
 "Create a new CoS Template": [
  null,
  "新しいCoSテンプレートを作成します。"
 ],
 "The CoS template entry contains a list of the shared attribute values. Changes to the template entry attribute values are automatically applied to all the entries within the scope of the CoS.": [
  null,
  "CoSテンプレートエントリには、共有属性値のリストが含まれています。テンプレートエントリの属性値の変更は、CoSの範囲内のすべてのエントリに自動的に適用されます。"
 ],
 "Right": [
  null,
  "権限"
 ],
 "OID": [
  null,
  "OID"
 ],
 "read": [
  null,
  "read"
 ],
 "See the values of targeted attributes": [
  null,
  "対象となる属性の値を確認します"
 ],
 "compare": [
  null,
  "compare"
 ],
 "Compare targeted attribute values": [
  null,
  "対象となる属性値を比較します"
 ],
 "search": [
  null,
  "search"
 ],
 "Determine if targeted attributes exist": [
  null,
  "対象となる属性が存在するかどうかを判断します"
 ],
 "selfwrite": [
  null,
  "selfwrite"
 ],
 "Add one's own DN to the target": [
  null,
  "自分自身のDNを対象に追加します"
 ],
 "write": [
  null,
  "write"
 ],
 "Modify targeted attributes": [
  null,
  "対象の属性を変更します"
 ],
 "delete": [
  null,
  "delete"
 ],
 "Remove targeted entries": [
  null,
  "対象の属性を削除します"
 ],
 "add": [
  null,
  "add"
 ],
 "Add targeted entries": [
  null,
  "対象のエントリを追加します"
 ],
 "moddn": [
  null,
  "moddn"
 ],
 "Move an entry from one subtree to another": [
  null,
  "他のサブツリーにエントリを移動します"
 ],
 "proxy": [
  null,
  "proxy"
 ],
 "Authenticate as another user": [
  null,
  "別のユーザとして認証します"
 ],
 "timeofday": [
  null,
  "日時"
 ],
 " and timeofday": [
  null,
  "と日時"
 ],
 "Successfully added ACI!": [
  null,
  "ACIの追加に成功しました！"
 ],
 "Failed to add ACI, error: ": [
  null,
  "ACIの追加に失敗しました、エラー: "
 ],
 "Choose a name for this ACI, and set the Target": [
  null,
  "このACIの名前を選択し、Targetを設定します。"
 ],
 "Target": [
  null,
  "対象"
 ],
 "LDAP Tree": [
  null,
  "LDAPツリー"
 ],
 "Available Entries": [
  null,
  "利用可能なエントリ"
 ],
 "Chosen Entries": [
  null,
  "選択したエントリ"
 ],
 "Searching database ...": [
  null,
  "データベースを検索しています ..."
 ],
 "Search for entries ...": [
  null,
  "エントリを検索"
 ],
 "Define the Bind Rules": [
  null,
  "バインドルールを定義"
 ],
 "The bind rules in an ACI define the required bind parameters that must meet so that Directory Server applies the ACI. For example user-based access uses the 'userdn' bind rule, while group access uses 'groupdn' bind rule.  There are other combinations that are possible, as well as defining the authentication method allowed or the connection security level. Please see the Administration Guide for more information.": [
  null,
  "ACI内のバインドルールは、ディレクトリサーバがACIを適用するために必要なバインドパラメータを定義します。たとえば、ユーザベースのアクセスでは「userdn」バインドルールを使用し、グループアクセスでは「groupdn」バインドルールを使用します。他の組み合わせも可能であり、許可される認証メソッドや接続セキュリティレベルの定義もできます。詳細については、管理ガイドを参照してください。"
 ],
 "Add Bind Rule": [
  null,
  "バインドルールの追加"
 ],
 "This section defines what permissions and rights are allowed/denied by this ACI.": [
  null,
  "このセクションでは、このアクセス制御設定によって許可または拒否される権限とアクセス権が定義されます。"
 ],
 "Choose the Target Attributes": [
  null,
  "ターゲット属性を選択"
 ],
 "You can control what entry attributes are affected by an ACI. If no target attributes are defined then all attributes are impacted by the ACI.  It is strongly discouraged to use the comparison operator ": [
  null,
  "ACIによって影響を受けるエントリ属性を制御することができます。ターゲット属性が定義されていない場合、すべての属性がACIの影響を受けます。比較演算子を使用することは強く推奨されません。"
 ],
 "Comparison Operator": [
  null,
  "比較演算子"
 ],
 "Target Filter ": [
  null,
  "ターゲットフィルタ"
 ],
 "Optionally set an LDAP search filter to further refine which entries this ACI will be applied to. Warning, using targetFilters can impact the server's performance and should be used sparingly.": [
  null,
  "必要に応じて、このACIが適用されるエントリをさらに絞り込むためのLDAP検索フィルタを設定することができます。ただし、targetFiltersを使用するとサーバのパフォーマンスに影響を与える可能性があり、慎重に使用する必要があります。"
 ],
 "Target Filter": [
  null,
  "ターゲットフィルタ"
 ],
 "The filter must be enclosed with parentheses": [
  null,
  "フィルタは、括弧で囲む必要があります"
 ],
 "Moving Entries": [
  null,
  "エントリの移動"
 ],
 "If you need to control access for moving an entry between specific subtrees then use the target_from and target_to keywords. These keywords use LDAP URL's to specify the database resources/locations: ldap ldap:": [
  null,
  "特定のサブツリー間でエントリの移動のアクセスを制御する必要がある場合は、target_fromおよびtarget_toキーワードを使用します。これらのキーワードは、LDAP URLを使用してデータベースのリソース/場所を指定します: ldap ldap:"
 ],
 "Specifies the entries that can be moved.  Example: ldap:///uid=*,cn=staging,dc=example,dc=com": [
  null,
  "移動できるエントリを指定します。例: ldap:///uid=*,cn=staging,dc=example,dc=com"
 ],
 "Target From": [
  null,
  "ターゲットの位置"
 ],
 "The LDAP URL must start with 'ldap:///'": [
  null,
  "LDAP URLは「ldap:///」で始まる必要があります。"
 ],
 "Specifies where the entries can be moved to.  Example: ldap:///ou=people,dc=example,dc=com": [
  null,
  "エントリを移動できる場所を指定します。例えば、ldap:///ou=people,dc=example,dc=comとなります。"
 ],
 "These bind rules control access based off of specific days or times. If no days are selected then all the days of the week are assumed.": [
  null,
  "これらのバインドルールは、特定の日や時間に基づいてアクセスを制御します。日が選択されていない場合、週のすべての日が適用されます。"
 ],
 "Choose Days To Allow Access": [
  null,
  "アクセスを許可する日数を選択"
 ],
 "Time Range Restrictions": [
  null,
  "時間範囲の制限"
 ],
 "First part of the time range": [
  null,
  "時間範囲の最初の部分"
 ],
 "Time Start": [
  null,
  "開始時刻"
 ],
 "Second part of the time range": [
  null,
  "時間範囲の後半部分"
 ],
 "Time End": [
  null,
  "終了時刻"
 ],
 "Review the new ACI, and make any edits as needed, but once you edit the ACI you can not go back to the previous wizard steps without undoing your changes.": [
  null,
  "新しいACIを確認し、必要に応じて編集を行ってください。ただし、ACIを編集した後は、変更を元に戻さない限り、以前のウィザードのステップに戻ることはできません。"
 ],
 "Undo Changes": [
  null,
  "変更を元に戻す"
 ],
 "Result for ACI addition": [
  null,
  "ACI追加の結果"
 ],
 "Adding ACI ...": [
  null,
  "ACIを追加しています ..."
 ],
 "ACI Value": [
  null,
  "ACIの値"
 ],
 "ACI Name & Target": [
  null,
  "ACI名とターゲット"
 ],
 "Target Attributes": [
  null,
  "ターゲット属性"
 ],
 "Access Rights": [
  null,
  "アクセス権"
 ],
 "Bind Rules": [
  null,
  "バインドルール"
 ],
 "Days & Times": [
  null,
  "日時制限"
 ],
 "Review & Edit": [
  null,
  "確認と編集"
 ],
 "Add ACI": [
  null,
  "ACIの追加"
 ],
 "Result": [
  null,
  "結果"
 ],
 "Finish": [
  null,
  "終了"
 ],
 "ACI: ": [
  null,
  "ACI:"
 ],
 "Add a new Access Control Instruction": [
  null,
  "新しいアクセス制御命令を追加します。"
 ],
 "Choose Bind Rule": [
  null,
  "バインドルールの選択"
 ],
 "User DN (userdn)": [
  null,
  "ユーザDN"
 ],
 "Bind rules for user entries": [
  null,
  "ユーザエントリのバインドルールを設定"
 ],
 "Group DN (groupdn)": [
  null,
  "グループDN"
 ],
 "Bind rules for groups": [
  null,
  "グループDNのバインドルール"
 ],
 "Role DN (roledn)": [
  null,
  "ロールDN"
 ],
 "Bind rules for Roles": [
  null,
  "ロールのバインドDN"
 ],
 "User DN Aliases (userdn)": [
  null,
  "ユーザDNエイリアス"
 ],
 "Special bind rules for user DN categories": [
  null,
  "ユーザDNカテゴリに特別なバインドルールを設定"
 ],
 "User Attribute (userattr)": [
  null,
  "ユーザ属性"
 ],
 "Bind rule to specify which attribute must match between the entry used to bind to the directory and the targeted entry": [
  null,
  "ディレクトリにバインドするために使用されるエントリと対象のエントリの間で一致する必要がある属性を指定するバインドルールです。"
 ],
 "Authentication Method (authmethod)": [
  null,
  "認証方式"
 ],
 "Specify the authentication methods to restrict": [
  null,
  "制限する認証方式を指定します"
 ],
 "Connection Security (ssf)": [
  null,
  "接続のセキュリティ"
 ],
 "Specify the connection security strength factor (ssf)": [
  null,
  "接続セキュリティ強度係数(ssf)を指定します"
 ],
 "IP Address (ip)": [
  null,
  "IPアドレス"
 ],
 "Specify an IP address or range to restrict": [
  null,
  "制限するIPアドレスまたは範囲を指定します"
 ],
 "Hostname (dns)": [
  null,
  "ホスト名"
 ],
 "Specify a hostname or domain to restrict": [
  null,
  "制限するホスト名やドメインを指定します"
 ],
 "Select Alias URL": [
  null,
  "エイリアスURLを選択"
 ],
 "ldap:///anyone": [
  null,
  "ldap:///anyone"
 ],
 "Grants Anonymous Access": [
  null,
  "匿名でのアクセスを許可"
 ],
 "ldap:///all": [
  null,
  "ldap:///all"
 ],
 "Grants Access to Authenticated Users": [
  null,
  "認証されたユーザに対してアクセスを許可"
 ],
 "ldap:///self": [
  null,
  "ldap:///self"
 ],
 "Enables Users to Access Their Own Entries": [
  null,
  "ユーザが自分のエントリにアクセスできるようにします"
 ],
 "ldap:///parent": [
  null,
  "ldap:///parent"
 ],
 "Grants Access for Child Entries of a User": [
  null,
  "ユーザの子エントリにアクセス権を付与します"
 ],
 "Choose Comparator": [
  null,
  "比較演算子を選択"
 ],
 "Invalid format for IP address": [
  null,
  "不正な形式のIPアドレス"
 ],
 "Invalid format for hostname": [
  null,
  "不正な形式のホスト名"
 ],
 "Security Strength Factor (ssf) - encryption key strength": [
  null,
  "セキュリティ強度ファクター(SSF)- 暗号化キーの強度"
 ],
 "Connection Security": [
  null,
  "接続のセキュリティ"
 ],
 "No connection security restrictions": [
  null,
  "接続セキュリティの制約がありません"
 ],
 "User Attribute": [
  null,
  "ユーザ属性"
 ],
 "Select an attribute": [
  null,
  "属性を選択"
 ],
 "USERDN": [
  null,
  "ユーザDN"
 ],
 "GROUPDN": [
  null,
  "グループDN"
 ],
 "ROLEDN": [
  null,
  "ロールDN"
 ],
 "SELFDN": [
  null,
  "セルフDN"
 ],
 "LDAPURL": [
  null,
  "LDAPURL"
 ],
 "Extend the application of the ACI several levels below the targeted entry. This is possible by specifying the number of levels below the target that should inherit the ACI.": [
  null,
  "対象エントリの下に複数のレベルにわたってACIを適用する。これは、対象からの継承レベル数を指定することで実現できます。"
 ],
 "Inheritance Levels": [
  null,
  "継承レベル"
 ],
 "Default value, this userattr bind rule only applies to the target entry (no child entries).": [
  null,
  "デフォルト値では、この'userattr'バインドルールは対象エントリのみに適用されます(子エントリは含まれません)。"
 ],
 "Bind Rule": [
  null,
  "バインドルール"
 ],
 "Comparator": [
  null,
  "コンパレータ"
 ],
 "LDAP URLs": [
  null,
  "LDAP URLs"
 ],
 "Remove Bind Rule": [
  null,
  "バインドルールの削除"
 ],
 "No bind rules": [
  null,
  "バインドルールがありません"
 ],
 "Attribute Name": [
  null,
  "属性名"
 ],
 "From ObjectClass": [
  null,
  "上位のオブジェクトクラス"
 ],
 "Invalid LDIF": [
  null,
  "無効なLDIF"
 ],
 "CoS Definition successfully created!": [
  null,
  "CoS の定義の作成に成功しました！"
 ],
 "Failed to create cos: ": [
  null,
  "cosの作成に失敗しました: "
 ],
 "Select Name And CoS Type": [
  null,
  "名前とCoSタイプを選択"
 ],
 "CoS Name": [
  null,
  "CoS名"
 ],
 "Choose CoS Type": [
  null,
  "CoSタイプの選択"
 ],
 "Pointer": [
  null,
  "ポインタ"
 ],
 "Identifies the template entry using the template DN only.": [
  null,
  "テンプレートDNのみを使用して、テンプレートエントリを識別する。"
 ],
 "Indirect": [
  null,
  "インダイレクト"
 ],
 "Identifies the template entry using the value of one of the target entry's attributes.": [
  null,
  "ターゲットエントリの属性の値を使用して、テンプレートエントリを特定します。"
 ],
 "Classic": [
  null,
  "クラシック"
 ],
 "Identifies the template entry using a combination of the template entry's base DN and the value of one of the target entry's attributes.": [
  null,
  "テンプレートエントリのベース DN とターゲットエントリの属性の値の組み合わせを使用して、テンプレートエントリを特定します。"
 ],
 "Select CoS Template": [
  null,
  "CoS テンプレートを選択"
 ],
 "The CoS template entry contains the value or values of the attributes generated by the CoS logic. The CoS template entry contains a general object class of cosTemplate The CoS template entries for a given CoS are stored in the directory tree along with the CoS definition.": [
  null,
  "CoSテンプレートエントリには、CoSロジックによって生成された属性の値または値が含まれています。CoSテンプレートエントリは、cosTemplateという一般的なオブジェクトクラスを持っています。特定のCoSに対するCoSテンプレートエントリは、CoSの定義とともにディレクトリツリーに格納されます。"
 ],
 "Find CoS Template...": [
  null,
  "CoSテンプレートを検索"
 ],
 "CoS Template Selected:": [
  null,
  "選択されたCoSテンプレート: "
 ],
 "Choose The New CoS Template Parent DN": [
  null,
  "新しいCoSテンプレートの親DNを選択してください"
 ],
 "Leaving CoS Definition Creation Wizard": [
  null,
  "CoS定義作成ウィザードを終了します"
 ],
 "You are about to leave CoS Definition creation wizard. After you click 'Confirm', you'll appear in CoS Template creation wizard and you won't able to return from there until the process is finished. Then you'll be able to use the created entry in the CoS definition creation. It'll be preselected for you automatically.": [
  null,
  "CoS定義作成ウィザードを終了し、CoSテンプレート作成ウィザードに移動します。移動後はプロセスが完了するまで戻ることはできません。作成されたエントリはCoS定義作成時に自動的に事前選択されます。「確認」をクリックすると進みます。"
 ],
 "Choose A Parent DN": [
  null,
  "親DNを選択する"
 ],
 "Done": [
  null,
  "完了"
 ],
 "Choose CoS Attributes": [
  null,
  "CoS属性の選択"
 ],
 "The cosAttribute contains the name of the attribute for which to generate a value for the CoS. There can be more than one cosAttribute value specified.": [
  null,
  "cosAttributeフィールドには、CoSの値を生成するための属性の名前が含まれています。複数のcosAttribute値を指定することもできます。"
 ],
 "Configure CoS Attributes": [
  null,
  "CoS 属性の設定"
 ],
 "The CoS attribute contains the name of another attribute which is governed by the class of service. This attribute allows an override qualifier after the attribute value which sets how the CoS handles existing attribute values on entries when it generates attribute values.": [
  null,
  "CoS属性には、クラス・オブ・サービスによって管理される別の属性の名前が含まれています。この属性には、属性値の後にオーバーライド修飾子が付加されることがあります。オーバーライド修飾子は、CoSが属性値を生成する際にエントリの既存の属性値の処理方法を設定します。"
 ],
 "Please, consult official documentation for more information.": [
  null,
  "詳しくは、公式ドキュメントをご覧ください。"
 ],
 "If no qualifier is set, default is assumed.": [
  null,
  "もし修飾子が設定されていない場合、デフォルトが適用されます。"
 ],
 "Override Qualifiers": [
  null,
  "オーバーライド修飾子"
 ],
 "Only returns a generated value if there is no corresponding attribute value stored with the entry.": [
  null,
  "対応する属性値がエントリに格納されていない場合にのみ、生成された値を返します。"
 ],
 "Default": [
  null,
  "Default"
 ],
 "Always returns the value generated by the CoS, even when there is a value stored with the entry.": [
  null,
  "エントリに値が格納されていても、常にCoSによって生成された値を返します。"
 ],
 "Override": [
  null,
  "Override"
 ],
 "Returns a generated attribute only if it is explicitly requested in the search. Operational attributes do not need to pass a schema check in order to be returned. When operational is used, it also overrides any existing attribute values.": [
  null,
  "検索時に明示的に要求された場合のみ、生成された属性を返します。オペレーション属性は、返されるためにスキーマチェックをパスする必要はありません。'operational'が使用されると、既存の属性値も上書きされます。"
 ],
 "Operational": [
  null,
  "Operational"
 ],
 "Only returns a generated value if there is no corresponding attribute value stored with the entry and if it is explicitly requested in the search.": [
  null,
  "検索時に明示的にリクエストされた場合、エントリに対応する属性値が保存されていない場合にのみ、生成された値を返します。"
 ],
 "Operational-Default": [
  null,
  "Operational-Default"
 ],
 "Using the merge-schemes qualifier tells the CoS that it will, or can, generate multiple values for the managed attribute.": [
  null,
  "マージスキーム修飾子を使用すると、CoSは管理された属性の複数の値を生成することができることを示します。"
 ],
 "Merge-Schemes": [
  null,
  "Merge-Schemes"
 ],
 "Select CoS Specifier": [
  null,
  "CoS指定子を選択"
 ],
 "Indirect type of CoS identifies the template entry based on the value of one of the target entry's attributes, as specified in the cosIndirectSpecifier attribute.": [
  null,
  "CoSの間接タイプは、cosIndirectSpecifier属性で指定された対象エントリの属性の値に基づいて、テンプレートエントリを特定します。"
 ],
 "Classic type of CoS identifies the template entry using both the template entry's DN (which you will assign later) and the value of one of the target entry's attributes (set in the cosSpecifier attribute).": [
  null,
  "CoSのクラシックタイプは、テンプレートエントリのDN(後で指定する)と対象エントリの属性の値(cosSpecifier属性で設定)の両方を使用して、テンプレートエントリを特定します。"
 ],
 "LDIF Content for CoS Creation": [
  null,
  "CoS作成のためのLDIFコンテンツ"
 ],
 "Result for CoS Creation": [
  null,
  "CoS作成結果"
 ],
 "Adding CoS definition ...": [
  null,
  "CoSの定義を追加しています ..."
 ],
 "LDIF Data": [
  null,
  "LDIFデータ"
 ],
 "Create CoS Template": [
  null,
  "CoSテンプレートの作成"
 ],
 "You've chosen 'Classic' CoS type. cosTemplateDN attribute is set to $0.": [
  null,
  "あなたは「Classic」 CoSタイプを選択しました。cosTemplateDN属性は$0に設定されています。"
 ],
 "Do you want to create a CoS template now?": [
  null,
  "今、CoSのテンプレートを作成しますか？"
 ],
 "It will be added as a child to this entry: $0": [
  null,
  "このエントリの子として追加されます: $0"
 ],
 "Select Type": [
  null,
  "タイプを選択"
 ],
 "Select CoS Attributes": [
  null,
  "CoS属性の選択"
 ],
 "Create CoS": [
  null,
  "CoSを作成する"
 ],
 "Review Result": [
  null,
  "結果を確認"
 ],
 "Create Templates": [
  null,
  "テンプレートの作成"
 ],
 "Parent DN:": [
  null,
  "親DN:"
 ],
 "Add A CoS Definition": [
  null,
  "CoS定義の追加"
 ],
 "Successfully added user!": [
  null,
  "ユーザの追加に成功しました！"
 ],
 "Failed to add user, error: ": [
  null,
  "ユーザの追加に失敗しました、エラー: "
 ],
 "Invalid Naming Attribute - Empty Value!": [
  null,
  "無効な命名属性 - 値が空です！"
 ],
 "DN ( Distinguished Name )": [
  null,
  "DN"
 ],
 "Select Entry Type": [
  null,
  "エントリタイプを選択"
 ],
 "Basic Account": [
  null,
  "ベーシックアカウント"
 ],
 " - This type of user entry uses a common set of objectclasses (nsPerson, nsAccount, and nsOrgPerson).": [
  null,
  "- このタイプのユーザエントリは、共通のオブジェクトクラス(nsPerson、nsAccount、nsOrgPerson)を使用します。"
 ],
 "Posix Account": [
  null,
  "ポジックスアカウント"
 ],
 " - This type of user entry uses a similar set of objectclasses as the <i>Basic Account</i>(nsPerson, nsAccount, nsOrgPerson, and posixAccount), but it includes POSIX attributes like:": [
  null,
  "- このタイプのユーザエントリは、ベーシックアカウント(nsPerson、nsAccount、nsOrgPerson、posixAccount)と同様のオブジェクトクラスセットを使用しますが、以下のようなPOSIX属性も含まれています:"
 ],
 "uidNumber, gidNumber, homeDirectory, loginShell, and gecos": [
  null,
  "uidNumber, gidNumber, homeDirectory, loginShell, and gecos"
 ],
 "Service Account": [
  null,
  "サービスアカウント"
 ],
 " - This type of entry uses a bare minimum of objectclasses (nsAccount, and applicationProcess) and attributes to create a simple object used to represent a service (not a user identity).": [
  null,
  "このタイプのエントリは、最小限のオブジェクトクラス(nsAccount、applicationProcess)と属性を使用して、ユーザのアイデンティティではなく、サービスを表すための単純なオブジェクトを作成します。"
 ],
 "Select Entry Attributes": [
  null,
  "エントリ属性を選択"
 ],
 "Entry DN:": [
  null,
  "エントリDN:"
 ],
 "Set Attribute Values": [
  null,
  "属性の値を設定"
 ],
 "LDIF Content for User Creation": [
  null,
  "ユーザ作成用LDIFコンテンツ"
 ],
 "Result for User Creation": [
  null,
  "ユーザ作成結果"
 ],
 "Adding user ...": [
  null,
  "ユーザを追加しています ..."
 ],
 "Choose User Type": [
  null,
  "ユーザタイプを選択"
 ],
 "Select Attributes": [
  null,
  "属性を選択"
 ],
 "Set Values": [
  null,
  "値を設定"
 ],
 "Create User": [
  null,
  "ユーザの作成"
 ],
 "Parent DN: ": [
  null,
  "親DN: "
 ],
 "Add A User": [
  null,
  "ユーザの追加"
 ],
 "Successfully renamed entry": [
  null,
  "エントリ名の変更に成功しました"
 ],
 "Successfully renamed entry!": [
  null,
  "エントリ名の変更に成功しました！"
 ],
 "Failed to rename entry, error: ": [
  null,
  "エントリ名の変更に失敗しました、エラー: "
 ],
 "Can not select a subtree that is the current entry, or a child of the current entry.": [
  null,
  "現在のエントリまたは現在のエントリの子エントリを選択することはできません。"
 ],
 "Select The Naming Attribute And Value": [
  null,
  "命名属性と値を選択"
 ],
 "New Name": [
  null,
  "新しい名前"
 ],
 "Removes the old RDN attribute from the entry, otherwise this attribute remains in the entry and references the old RDN.": [
  null,
  "古いRDN属性をエントリから削除します。そうしない場合、この属性はエントリに残り、古いRDNを参照します。"
 ],
 "Delete the old RDN attribute from the entry": [
  null,
  "エントリから古いRDN属性を削除する"
 ],
 "Select The Entry Location": [
  null,
  "エントリの場所を選択"
 ],
 "The superior DN where the entry will be located under.": [
  null,
  "エントリが配置される上位のDNです。"
 ],
 "Parent Subtree": [
  null,
  "親のサブツリー"
 ],
 "Review Changes": [
  null,
  "変更を確認"
 ],
 "Original DN": [
  null,
  "元のDN"
 ],
 "New DN": [
  null,
  "新しいDN"
 ],
 "This attribute will be removed from the entry so that there is no trace of the old RDN in the entry.": [
  null,
  "この属性はエントリから削除され、エントリに古いRDNの痕跡が残らないようになります。"
 ],
 "Remove Old Rdn": [
  null,
  "古いRDNを削除"
 ],
 "LDIF Statements": [
  null,
  "LDIFステートメント"
 ],
 "Result for Entry Modification": [
  null,
  "エントリの変更結果"
 ],
 "Renaming entry ...": [
  null,
  "エントリの名前を変更しています ..."
 ],
 "Set Naming Attribute": [
  null,
  "命名属性を設定"
 ],
 "Set Entry Location": [
  null,
  "エントリの場所を設定"
 ],
 "Change Entry Name": [
  null,
  "エントリ名を変更"
 ],
 "Old DN": [
  null,
  "古いDN"
 ],
 "Rename LDAP Entry": [
  null,
  "LDAPエントリ名を変更"
 ],
 "Group successfully created!": [
  null,
  "グループの作成に成功しました！"
 ],
 "Failed to create group, error: ": [
  null,
  "グループの作成に失敗しました、エラー: "
 ],
 "Select Group Type": [
  null,
  "グループのタイプを選択"
 ],
 "Basic Group": [
  null,
  "ベーシックグループ"
 ],
 "- This type of group can use membership attributes: member, or uniqueMember common set of objectclasses and attributes.": [
  null,
  "- このタイプのグループでは、メンバーシップ属性として、memberまたはuniqueMemberを使用できます。共通のオブジェクトクラスと属性を持ちます。"
 ],
 "Posix Group": [
  null,
  "ポジックスグループ"
 ],
 "- This type of group uses the 'GroupOfNames' and 'PosixGroup' objectclasses which allows for attributes like <i>gidNumber</i>, etc": [
  null,
  "- このタイプのグループは、「GroupOfNames」および「PosixGroup」のオブジェクトクラスを使用し、gidNumberなどの属性を設定することができます。"
 ],
 "Select Name": [
  null,
  "名前を選択"
 ],
 "Group Name": [
  null,
  "グループ名"
 ],
 "Optional description for this group.": [
  null,
  "このグループのオプションの説明"
 ],
 "Group Description": [
  null,
  "グループの概要"
 ],
 "Choose The Membership": [
  null,
  "メンバーシップを選択"
 ],
 "This group uses objectclass 'GroupOfNames'.": [
  null,
  "このグループは、オブジェクトクラス「GroupOfNames」を使用しています。"
 ],
 "uniquemember": [
  null,
  "ユニークメンバー"
 ],
 "This group uses objectclass 'GroupOfUniqueNames'.": [
  null,
  "このグループは、オブジェクトクラス「GroupOfUniqueNames」を使用しています。"
 ],
 "This setting corresponds to the attribute: gidNumber.  The gid must be greater than zero for this attribute to be set.": [
  null,
  "この設定は属性「gidNumber」に対応しています。gidNumberの値はゼロより大きい必要があります。"
 ],
 "Group ID Number": [
  null,
  "グループID番号"
 ],
 "Add Members To The Group": [
  null,
  "グループにメンバーを追加する"
 ],
 "Find members...": [
  null,
  "メンバーを検索"
 ],
 "Available Members": [
  null,
  "利用可能なメンバー"
 ],
 "Chosen Members": [
  null,
  "選択したメンバー"
 ],
 "Choose A Branch To Search": [
  null,
  "検索するブランチを選択"
 ],
 "LDIF Content for Group Creation": [
  null,
  "グループ作成用LDIFコンテンツ"
 ],
 "Result for Group Creation": [
  null,
  "グループ作成結果"
 ],
 "Adding group ...": [
  null,
  "グループを追加しています ..."
 ],
 "Add Members": [
  null,
  "メンバーを追加"
 ],
 "Create Group": [
  null,
  "グループを作成"
 ],
 "Add A Group": [
  null,
  "グループを追加"
 ],
 "Statement": [
  null,
  "ステートメント"
 ],
 "ObjectClass Name": [
  null,
  "オブジェクトクラス名"
 ],
 "Required Attributes": [
  null,
  "必要な属性"
 ],
 "Optional Attributes": [
  null,
  "オプション属性"
 ],
 "Successfully added entry!": [
  null,
  "エントリの追加に成功しました!"
 ],
 "Failed to add entry, error: ": [
  null,
  "エントリの追加に失敗しました、エラー: "
 ],
 "Select ObjectClasses": [
  null,
  "オブジェクトクラスを選択"
 ],
 "Loading ObjectClasses ...": [
  null,
  "オブジェクトクラスを読み込んでいます ..."
 ],
 "Search Objectclasses": [
  null,
  "オブジェクトクラスを検索"
 ],
 "Adding entry ...": [
  null,
  "エントリを追加しています ..."
 ],
 "Edit Values": [
  null,
  "値の編集"
 ],
 "Create Entry": [
  null,
  "エントリの作成"
 ],
 "Add An LDAP Entry": [
  null,
  "LDAPエントリを追加"
 ],
 "Successfully added $0 member $1": [
  null,
  "$0メンバー$1の追加に成功しました"
 ],
 "Failed to update group, error code: ": [
  null,
  "グループの更新に失敗しました、エラーコード: "
 ],
 "Successfully updated group": [
  null,
  "グループの更新に成功しました"
 ],
 "Switch To Generic Editor": [
  null,
  "一般のエディタに切り替える"
 ],
 "Current Members": [
  null,
  "現在のメンバー"
 ],
 "Find New Members": [
  null,
  "新メンバーを検索"
 ],
 "View Entry": [
  null,
  "エントリを表示"
 ],
 "Remove Member From Group": [
  null,
  "グループからメンバーを削除"
 ],
 "Are you sure you want to remove this member?": [
  null,
  "本当にこのメンバーを削除しますか？"
 ],
 "Remove Members From Group": [
  null,
  "グループからメンバーを削除"
 ],
 "Are you sure you want to remove these members?": [
  null,
  "本当にこれらのメンバーを削除しますか？"
 ],
 "Delete Members": [
  null,
  "メンバーの削除"
 ],
 "Successfully modified entry": [
  null,
  "エントリの変更に成功しました"
 ],
 "Successfully modified entry!": [
  null,
  "エントリの変更に成功しました！"
 ],
 "Failed to modify entry, error: ": [
  null,
  "エントリの変更に失敗しました、エラー: "
 ],
 "Keep": [
  null,
  "保存"
 ],
 "Replace": [
  null,
  "置き換え"
 ],
 "old:": [
  null,
  "古い"
 ],
 "new:": [
  null,
  "新しい"
 ],
 "'Loading...": [
  null,
  "読み込み中..."
 ],
 "Edit Attribute Values": [
  null,
  "属性の値を編集"
 ],
 "Modifying entry ...": [
  null,
  "エントリを変更しています ..."
 ],
 "View Changes": [
  null,
  "変更を確認"
 ],
 "Modify Entry": [
  null,
  "エントリを変更"
 ],
 "Entry DN: ": [
  null,
  "エントリDN: "
 ],
 "Edit An LDAP Entry": [
  null,
  "LDAPエントリの編集"
 ],
 "Invalid RDN - Empty Value!": [
  null,
  "無効なRDN - 値が空です！"
 ],
 "The full DN of the entry will be: ": [
  null,
  "エントリのフルDNは、このようになります: "
 ],
 "LDIF Content for Entry Creation": [
  null,
  "エントリ作成用のLDIFコンテンツ"
 ],
 "Result for Entry Creation": [
  null,
  "エントリの作成結果は以下の通りです。"
 ],
 "Create Organizational Unit": [
  null,
  "組織ユニットを作成"
 ],
 "Create LDAP Entry": [
  null,
  "LDAPエントリを作成"
 ],
 "Add A New User": [
  null,
  "新しいユーザを追加"
 ],
 "Add An Organizational Unit": [
  null,
  "組織ユニットを追加"
 ],
 "Add A New LDAP Entry": [
  null,
  "新しいLDAPエントリを追加"
 ],
 "Member DN": [
  null,
  "メンバーDN"
 ],
 "No Members": [
  null,
  "メンバーが存在しません"
 ],
 "Edit Entry": [
  null,
  "エントリを編集"
 ],
 "Remove Entry": [
  null,
  "エントリを削除"
 ],
 "Remove Selected Members": [
  null,
  "選択したメンバーを削除"
 ],
 "Updating group ...": [
  null,
  "グループを更新しています ..."
 ],
 "Updating": [
  null,
  "更新中"
 ],
 "Search Members": [
  null,
  "メンバーを検索"
 ],
 "Role successfully created!": [
  null,
  "ロールの作成に成功しました！"
 ],
 "Failed to create role, error: ": [
  null,
  "ロールの作成に失敗しました、エラー: "
 ],
 "Select Name And Role Type": [
  null,
  "名前とロールタイプを選択"
 ],
 "Role Name": [
  null,
  "ロール名"
 ],
 "Choose Role Type": [
  null,
  "ロールタイプを選択"
 ],
 "Managed": [
  null,
  "マネージド"
 ],
 "This attribute uses objectclass 'RoleOfNames'": [
  null,
  "この属性は、オブジェクトクラス'RoleOfNames'を使用します。"
 ],
 "Filtered": [
  null,
  "フィルタ"
 ],
 "This attribute uses objectclass 'RoleOfUniqueNames'": [
  null,
  "この属性は、オブジェクトクラス'RoleOfUniqueNames'を使用します。"
 ],
 "Nested": [
  null,
  "ネスト"
 ],
 "Add Roles to the Nested Role": [
  null,
  "ネストされたロールにロールを追加する"
 ],
 "Find roles...": [
  null,
  "ロールを検索"
 ],
 "Available Roles": [
  null,
  "利用可能なロール"
 ],
 "Chosen Roles": [
  null,
  "選択したロール"
 ],
 "LDIF Content for Role Creation": [
  null,
  "ロール作成のためのLDIFコンテンツ"
 ],
 "Result for Role Creation": [
  null,
  "ロール作成結果"
 ],
 "Adding Role ...": [
  null,
  "ロールを追加しています ..."
 ],
 "Select Name & Type": [
  null,
  "名前とタイプを選択"
 ],
 "Add Nested Roles": [
  null,
  "ネストされたロールの追加"
 ],
 "Create Role": [
  null,
  "ロールを作成"
 ],
 "Parent DN": [
  null,
  "親DN"
 ],
 "Add A Role": [
  null,
  "ロールを追加"
 ],
 "Select Name And CoS Priority": [
  null,
  "名前とCoSの優先順位を選択"
 ],
 "The CoS template entry contains the value or values of the attributes generated by the CoS logic. The CoS template entry contains a general object class of cosTemplate. The CoS template entries for a given CoS are stored in the directory tree along with the CoS definition.": [
  null,
  "CoSテンプレートエントリには、CoSロジックによって生成された属性の値または値が含まれています。CoSテンプレートエントリには、cosTemplateという一般的なオブジェクトクラスが含まれています。特定のCoSに対するCoSテンプレートエントリは、CoSの定義とともにディレクトリツリーに格納されます。"
 ],
 "The relative distinguished name (RDN) of the template entry is determined by one of the following:": [
  null,
  "テンプレートエントリの相対識別名(RDN)は、次のいずれかによって決定されます:"
 ],
 "The DN of the template entry alone. This type of template is associated with a pointer CoS definition.": [
  null,
  "テンプレートエントリのDNのみです。このタイプのテンプレートは、ポインタCoS定義と関連付けられています。"
 ],
 "The value of one of the target entry's attributes. The attribute used to provide the relative DN to the template entry is specified in the CoS definition entry using the cosIndirectSpecifier attribute. This type of template is associated with an indirect CoS definition.": [
  null,
  "対象エントリの属性の値です。相対DNをテンプレートエントリに提供するために使用される属性は、CoS定義エントリでcosIndirectSpecifier属性を使用して指定されます。このタイプのテンプレートは間接的なCoS定義と関連付けられています。"
 ],
 "By a combination of the DN of the subtree where the CoS performs a one level search for templates and the value of one of the target entry's attributes. This type of template is associated with a classic CoS definition.": [
  null,
  "CoSがテンプレートを1レベルの検索で実行するサブツリーのDNと、対象エントリの属性の値の組み合わせで決まります。このタイプのテンプレートはクラシックなCoS定義と関連付けられています。"
 ],
 "Please, consult official documentation for more information and examplesa.": [
  null,
  "公式のドキュメントを参照して詳細や例を確認してください。"
 ],
 "CoS Template Name": [
  null,
  "CoSテンプレート名"
 ],
 "CoS Priority": [
  null,
  "CoSの優先順位"
 ],
 "Specifies which template provides the attribute value when CoS templates compete to provide an attribute value. This attribute represents the global priority of a template. A priority of zero is the highest priority.": [
  null,
  "優先度は、CoS テンプレートが属性値を提供する際に競合する場合に、どのテンプレートが属性値を提供するかを指定します。この属性は、テンプレートのグローバルな優先度を表します。優先度が0の場合、最も高い優先度となります。"
 ],
 "Create More Templates": [
  null,
  "さらにテンプレートを作成する"
 ],
 "Back to CoS Definition": [
  null,
  "CoS定義へ戻る"
 ],
 "Select Name & Priority": [
  null,
  "名前と優先順位を選択"
 ],
 "Add CoS Template": [
  null,
  "CoSテンプレートの追加"
 ],
 "Successfully deleted entry": [
  null,
  "エントリの削除が成功しました"
 ],
 "Failed to delete entry, error: ": [
  null,
  "エントリの削除に失敗しました、エラー: "
 ],
 "It has child entries which will also be recursively deleted.": [
  null,
  "子エントリも再帰的に削除されます。"
 ],
 "Delete Acknowledgement": [
  null,
  "削除の確認"
 ],
 "You are about to delete this entry. $0": [
  null,
  "このエントリを削除しようとしています。$0"
 ],
 "LDIF Entry To Be Deleted": [
  null,
  "削除されるLDIFエントリ"
 ],
 "Run this command to retrieve the LDAP entries that will be deleted": [
  null,
  "このコマンドを実行して、削除されるLDAPエントリを取得します。"
 ],
 "Copy": [
  null,
  "コピーする"
 ],
 "Copied": [
  null,
  "コピーしました"
 ],
 "This entry, as well as all its child entries,": [
  null,
  "このエントリとすべての子エントリ"
 ],
 "This entry": [
  null,
  "このエントリ"
 ],
 "This is an irreversible operation!": [
  null,
  "これは取り消しできない操作です！"
 ],
 "Confirm Entry Deletion": [
  null,
  "エントリの削除を確認"
 ],
 "$0 will be deleted.": [
  null,
  "$0は削除されます。"
 ],
 "Yes, I'm sure.": [
  null,
  "はい、削除します。"
 ],
 "No, don't delete.": [
  null,
  "いいえ、削除しません。"
 ],
 "The entries were": [
  null,
  "このエントリは"
 ],
 "The entry was": [
  null,
  "このエントリは"
 ],
 " successfully deleted.": [
  null,
  "削除に成功しました。"
 ],
 "There was an error during the deletion.": [
  null,
  "削除中にエラーが発生しました。"
 ],
 "Acknowledgement": [
  null,
  "確認"
 ],
 "Deletion": [
  null,
  "削除"
 ],
 "Review": [
  null,
  "確認"
 ],
 "Delete LDAP Entry": [
  null,
  "LDAPエントリの削除"
 ],
 "Delete LDAP Entries": [
  null,
  "LDAPエントリの削除"
 ],
 "Objectclass Name": [
  null,
  "オブジェクトクラス名"
 ],
 "No description": [
  null,
  "詳細がありません"
 ],
 "No Objectclasses": [
  null,
  "オブジェクトクラスがありません"
 ],
 "Objectclasses": [
  null,
  "オブジェクトクラス"
 ],
 "Edit Objectclass": [
  null,
  "オブジェクトクラスの編集"
 ],
 "Delete Objectclass": [
  null,
  "オブジェクトクラスの削除"
 ],
 "Loading Objectclasses ...": [
  null,
  "オブジェクトクラスを読み込んでいます..."
 ],
 "Syntax": [
  null,
  "構文"
 ],
 "Description:": [
  null,
  "説明:"
 ],
 "X-Origin:": [
  null,
  "X-Origin:"
 ],
 "Aliases:": [
  null,
  "エイリアス:"
 ],
 "Parent Attribute:": [
  null,
  "親属性:"
 ],
 "Read Only:": [
  null,
  "読み込み専用:"
 ],
 "Multivalued:": [
  null,
  "複数値:"
 ],
 "Usage:": [
  null,
  "使い方:"
 ],
 "Equality Matching Rules:": [
  null,
  "等価マッチングルール:"
 ],
 "Substring Matching Rules:": [
  null,
  "部分文字列に一致するルール:"
 ],
 "Ordering Matching Rules:": [
  null,
  "順序に一致するルール:"
 ],
 "Edit Attribute": [
  null,
  "属性の編集"
 ],
 "Loading Attributes ...": [
  null,
  "属性を読み込んでいます..."
 ],
 "Matching Rule": [
  null,
  "一致するルール"
 ],
 "No Matching Rules": [
  null,
  "一致するルールがありません"
 ],
 "No Name": [
  null,
  "名前がありません"
 ],
 "Search Matching Rules": [
  null,
  "一致するルールの検索"
 ],
 "$0 ObjectClass": [
  null,
  "オブジェクトクラスを$0"
 ],
 "Saving...": [
  null,
  "保存しています..."
 ],
 "Name of the objectClass": [
  null,
  "オブジェクトクラスの名前を設定します"
 ],
 "Describes what is the objectClass's purpose": [
  null,
  "オブジェクトクラスの目的を記載します"
 ],
 "Object identifier": [
  null,
  "オブジェクト識別子を設定します"
 ],
 "OID (optional)": [
  null,
  "OID(オプション)"
 ],
 "An objectClass's parent": [
  null,
  "オブジェクトクラスの親を設定します"
 ],
 "Parent Objectclass": [
  null,
  "親オブジェクトクラス"
 ],
 "An objectClass's kind": [
  null,
  "オブジェクトクラスの種類を設定します"
 ],
 "Objectclass Kind": [
  null,
  "オブジェクトクラスの種類"
 ],
 "A must attribute name": [
  null,
  "必須とする属性名を設定します"
 ],
 "A may attribute name": [
  null,
  "許可する属性名を設定します"
 ],
 "Allowed Attributes": [
  null,
  "許可属性"
 ],
 "View Attribute - $0": [
  null,
  "属性を表示 - $0"
 ],
 "$0 Attribute": [
  null,
  "属性を$0"
 ],
 "Name of the attribute": [
  null,
  "属性の名前を設定します"
 ],
 "Describes what is the attribute's purpose": [
  null,
  "属性の目的を記載します"
 ],
 "An attribute's parent/superior": [
  null,
  "属性の親/上位を設定します"
 ],
 "Parent Attribute": [
  null,
  "親属性"
 ],
 "An attribute's syntax": [
  null,
  "属性の構文を設定します"
 ],
 "Syntax Name": [
  null,
  "構文名"
 ],
 "An attribute's usage purpose": [
  null,
  "属性の使用目的を記載します"
 ],
 "Attribute Usage": [
  null,
  "属性の使い方"
 ],
 "If attribute can have a multiple values": [
  null,
  "属性が複数の値を持つことができる場合に有効にする"
 ],
 "Multivalued Attribute": [
  null,
  "複数値属性を有効にする"
 ],
 "If attribute is not modifiable by a client application": [
  null,
  "属性がクライアントアプリケーションによって変更できないように設定します。"
 ],
 "Not Modifiable By A User": [
  null,
  "ユーザによって変更できないようにする"
 ],
 "An alias name for the attribute": [
  null,
  "属性のエイリアス名を設定します"
 ],
 "Alias Names": [
  null,
  "エイリアス名"
 ],
 "Type an alias name...": [
  null,
  "エイリアス名を入力してください..."
 ],
 "An equality matching rule": [
  null,
  "等価マッチングルールを設定します"
 ],
 "Equality Matching Rule": [
  null,
  "等価マッチングルール"
 ],
 "Type an Equality matching rule...": [
  null,
  "等価マッチングルールを入力してください..."
 ],
 "There are no matching rules": [
  null,
  "一致するルールがありません"
 ],
 "An order matching rule": [
  null,
  "順序に一致するルールを設定します"
 ],
 "Order Matching Rule": [
  null,
  "順序に一致するルール"
 ],
 "Type an Ordering matching rule..": [
  null,
  "順序に一致するルールを入力してください..."
 ],
 "A substring matching rule": [
  null,
  "部分文字列に一致するルールを設定します"
 ],
 "Substring Matching Rule": [
  null,
  "部分文字列に一致するルール"
 ],
 "Type a Substring matching rule...": [
  null,
  "部分文字列に一致するルールを入力してください..."
 ],
 "Export Certificate": [
  null,
  "証明書のエクスポート"
 ],
 "Exporting Certificate ...": [
  null,
  "証明書をエクスポートしています..."
 ],
 "Exporting": [
  null,
  "エクスポート中"
 ],
 "Export Certificate:": [
  null,
  "証明書のエクスポート:"
 ],
 "Enter the full path and file name, if the path portion is omitted the certificate is written to the server's certificate directory ": [
  null,
  "ファイル名を含むフルパスを入力してください。パス部分を省略した場合、証明書はサーバの証明書ディレクトリに書き込まれます。"
 ],
 "Certificate File Name": [
  null,
  "証明書ファイル名"
 ],
 "Enter full path to and and including certificate file name": [
  null,
  "証明書ファイル名を含むフルパスを入力してください"
 ],
 "Export certificate in its binary format.  Otherwise a PEM file is created.": [
  null,
  "証明書をバイナリ形式でエクスポートします。それ以外の場合は、PEM ファイルが作成されます。"
 ],
 "Export Certificate In Binary/DER Format": [
  null,
  "証明書をバイナリ/DER 形式でエクスポート"
 ],
 "Add Certificate": [
  null,
  "証明書を追加"
 ],
 "Adding Certificate ...": [
  null,
  "証明書を追加しています..."
 ],
 "Paste the base64 encoded certificate that starts with \"-----BEGIN CERTIFICATE-----\" and ends with \"-----END CERTIFICATE-----\".  Make sure there are no special carriage return characters after each line.": [
  null,
  "base64でエンコードされた \"-----BEGIN CERTIFICATE-----\" で始まり、\"-----END CERTIFICATE-----\" で終わる証明書を貼り付けます。各行の後に特別な改行文字がないことを確認してください。"
 ],
 "Upload PEM File, or Certificate Text": [
  null,
  "PEMファイルまたは証明書のテキストをアップロード"
 ],
 "Add Server Certificate": [
  null,
  "サーバ証明書を追加"
 ],
 "Add a Server Certificate to the security database.": [
  null,
  "サーバ証明書をセキュリティデータベースに追加します。"
 ],
 "Add Certificate Authority": [
  null,
  "認証局を追加"
 ],
 "Add a CA Certificate to the security database.": [
  null,
  "CA証明書をセキュリティデータベースに追加します。"
 ],
 "Enter name/nickname of the certificate": [
  null,
  "証明書の名前/ニックネームを入力してください"
 ],
 "Certificate Nickname": [
  null,
  "証明書のニックネーム"
 ],
 "Upload the contents of a PEM file from the client's system.": [
  null,
  "クライアントシステムからPEMファイルの内容をアップロードしてください"
 ],
 "Drag and drop a file, or upload one": [
  null,
  "ファイルをドラッグ＆ドロップするか、アップロードしてください"
 ],
 "Upload PEM File": [
  null,
  "PEMファイルをアップロード"
 ],
 "Choose a certificate from the server's certificate directory": [
  null,
  "サーバの証明書ディレクトリから証明書を選択してください"
 ],
 "Choose Certificate From Server": [
  null,
  "サーバから証明書を選択"
 ],
 "No certificates present": [
  null,
  "証明書が存在しません"
 ],
 "Enter the full path on the server to and including the certificate file name": [
  null,
  "サーバ上の証明書ファイルのパス(ファイル名を含む)を入力してください"
 ],
 "Certificate File Location": [
  null,
  "証明書ファイルの場所"
 ],
 "Create Certificate Signing Request": [
  null,
  "証明書署名リクエスト(CSR)を作成"
 ],
 "Creating Certificate Signing Request ...": [
  null,
  "証明書署名リクエスト(CSR)を作成しています..."
 ],
 "CSR Name": [
  null,
  "CSR名を設定します"
 ],
 "Name used to identify a CSR": [
  null,
  "CSRを識別するための名前を入力してください"
 ],
 "CSR Subject alternative host names": [
  null,
  "CSRサブジェクトの別名(SAN)を入力してください"
 ],
 "Subject Alternative Names": [
  null,
  "サブジェクトの別名"
 ],
 "Type an alternative host name": [
  null,
  "サブジェクトの別名を入力してください"
 ],
 "Invalid host names: ": [
  null,
  "無効なホスト名です:"
 ],
 "CSR Subject: Common Name": [
  null,
  "CSRサブジェクト: コモンネーム(CN)を設定します"
 ],
 "Common Name (CN)": [
  null,
  "コモンネーム(CN)"
 ],
 "The fully qualified domain name (FQDN) of your server": [
  null,
  "サーバの完全修飾ドメイン名(FQDN)を入力してください"
 ],
 "CSR Subject: Organisation": [
  null,
  "CSRサブジェクト: 組織名(O)を設定します"
 ],
 "Organization (O)": [
  null,
  "組織名(O)"
 ],
 "The legal name of your organization": [
  null,
  "組織の正式名称を入力してください"
 ],
 "CSR Subject: Organisational Unit": [
  null,
  "CSRサブジェクト: 組織単位名(OU)を設定します"
 ],
 "Organizational Unit (OU)": [
  null,
  "組織単位名(OU)"
 ],
 "The division of your organization handling the certificate": [
  null,
  "証明書を扱う組織の部門を入力してください"
 ],
 "CSR Subject: City/Locality": [
  null,
  "CSRサブジェクト: 市/地域(L)を設定します"
 ],
 "City/Locality (L)": [
  null,
  "市/地域(L)"
 ],
 "The city where your organization is located": [
  null,
  "組織の所在地の市/地域を入力してください"
 ],
 "CSR Subject: State/Region": [
  null,
  "CSRサブジェクト: 州/地域(ST)を設定します"
 ],
 "State/County/Region (ST)": [
  null,
  "州/郡/地域(ST)"
 ],
 "The state/region where your organization is located": [
  null,
  "組織の所在地である州/地域を入力してください"
 ],
 "CSR Subject: Country Code": [
  null,
  "CSRサブジェクト: 国コード(C)を設定します"
 ],
 "Country Code (C)": [
  null,
  "国コード(C)"
 ],
 "Two-letter country code where organization is located": [
  null,
  "組織の所在地である2文字国コードを入力してください"
 ],
 "CSR Subject: Email Address": [
  null,
  "CSRサブジェクト:メールアドレスを設定します"
 ],
 "Email Address": [
  null,
  "メールアドレス"
 ],
 "Email address used to contact your organization": [
  null,
  "組織との連絡の取れるメールアドレスを入力してください"
 ],
 "Computed Subject": [
  null,
  "生成されるサブジェクト"
 ],
 "CSR content": [
  null,
  "CSRの内容"
 ],
 "Copy to clipboard": [
  null,
  "クリップボードにコピー"
 ],
 "Nothing to display": [
  null,
  "表示する内容はありません"
 ],
 "Enable Security": [
  null,
  "セキュリティを有効にする"
 ],
 "Enabling Security ...": [
  null,
  "セキュリティを有効にしています..."
 ],
 "You are choosing to enable security for the Directory Server which allows the server to accept incoming client TLS connections.  Please select which certificate the server should use.": [
  null,
  "ディレクトリサーバのセキュリティを有効にすることを選択しました。これにより、サーバはクライアントからのTLS接続を受け入れることができるようになります。サーバが使用する証明書を選択してください。"
 ],
 "The server certificates the Directory Server can use": [
  null,
  "ディレクトリサーバが使用できるサーバ証明書"
 ],
 "Available Certificates": [
  null,
  "利用可能な証明書"
 ],
 "Save Flags": [
  null,
  "フラグを保存"
 ],
 "Saving flags ...": [
  null,
  "フラグを保存しています..."
 ],
 "Edit Certificate Trust Flags (": [
  null,
  "CA信頼フラグの編集 ("
 ],
 "Flags": [
  null,
  "フラグ"
 ],
 "SSL": [
  null,
  "SSL"
 ],
 "Email": [
  null,
  "メール"
 ],
 "Object Signing": [
  null,
  "オブジェクト署名"
 ],
 "Trusted CA (flag 'C', also implies 'c' flag)": [
  null,
  "信頼されたCA (フラグ 'C'、'c'フラグも含まれます)"
 ],
 "(C) - Trusted CA": [
  null,
  "(C) - 信頼されたCA"
 ],
 "Trusted CA for client authentication (flag 'T')": [
  null,
  "クライアント認証のための信頼されたCA (フラグ 'T')"
 ],
 "(T) - Trusted CA Client Auth": [
  null,
  "(T) - 信頼されたCA クライアント認証"
 ],
 "Valid CA (flag 'c')": [
  null,
  "有効なCA (フラグ 'c')"
 ],
 "(c) - Valid CA": [
  null,
  "(c) - 有効なCA"
 ],
 "Trusted Peer (flag 'P', implies flag 'p')": [
  null,
  "信頼されたピア (フラグ 'P'、'p'フラグを含む)"
 ],
 "(P) - Trusted Peer": [
  null,
  "(P) - 信頼されたピア"
 ],
 "Valid Peer (flag 'p')": [
  null,
  "有効なピア (フラグ 'p')"
 ],
 "(p) - Valid Peer": [
  null,
  "(p) - 有効なピア"
 ],
 "A private key is associated with the certificate. This is a dynamic flag and you cannot adjust it.": [
  null,
  "証明書には秘密鍵が関連付けられています。これは動的なフラグであり、調整することはできません。"
 ],
 "(u) - Private Key": [
  null,
  "(u) - 秘密鍵"
 ],
 "Successfully exported certificate as: $0": [
  null,
  "証明書をエクスポートしました: $0"
 ],
 "Error exporting certificate - $0": [
  null,
  "証明書のエクスポートに失敗しました - $0"
 ],
 "Error deleting tmp certificate - $0": [
  null,
  "一時的な証明書の削除に失敗しました - $0"
 ],
 "Successfully added certificate": [
  null,
  "証明書を追加しました"
 ],
 "Error adding certificate - $0": [
  null,
  "証明書の追加に失敗しました - $0"
 ],
 "Faield to create temporary certificate file: $0": [
  null,
  "一時的な証明書ファイルの作成に失敗しました - $0"
 ],
 "Successfully created CSR": [
  null,
  "CSRを作成しました"
 ],
 "Error Improperly formatted subject": [
  null,
  "サブジェクトの形式が間違っています"
 ],
 "Error creating CSR - $0": [
  null,
  "CSRの作成に失敗しました - $0"
 ],
 "Missing CSR Name": [
  null,
  "CSR名がありません"
 ],
 "Error displaying CSR - $0": [
  null,
  "CSRの表示に失敗しました - $0"
 ],
 "Successfully deleted certificate": [
  null,
  "証明書を削除しました"
 ],
 "Error deleting certificate - $0": [
  null,
  "証明書の削除に失敗しました - $0"
 ],
 "Successfully deleted CSR": [
  null,
  "CSRを削除しました"
 ],
 "Error deleting CSR - $0": [
  null,
  "CSRの削除に失敗しました - $0"
 ],
 "Successfully deleted key": [
  null,
  "鍵を削除しました"
 ],
 "Error deleting key - $0": [
  null,
  "鍵の削除に失敗しました - $0"
 ],
 "Successfully changed certificate's trust flags": [
  null,
  "証明書の信頼フラグを変更しました"
 ],
 "Error setting trust flags - $0": [
  null,
  "信頼フラグの設定に失敗しました - $0"
 ],
 "Error loading server certificates - $0": [
  null,
  "サーバ証明書の読み込みに失敗しました - $0"
 ],
 "Error loading CSRs - $0": [
  null,
  "CSRの読み込みに失敗しました - $0"
 ],
 "Error loading Orphan Keys - $0": [
  null,
  "孤立鍵の読み込みに失敗しました - $0"
 ],
 "Error loading CA certificates - $0": [
  null,
  "CA証明書の読み込みに失敗しました - $0"
 ],
 "Loading Certificates ...": [
  null,
  "証明書を読み込んでいます..."
 ],
 "Trusted Certificate Authorities": [
  null,
  "信頼された認証局"
 ],
 "Add CA Certificate": [
  null,
  "CA証明書を追加"
 ],
 "TLS Certificates": [
  null,
  "TLS証明書"
 ],
 "Certificate Sigining Requests": [
  null,
  "証明書署名リクエスト"
 ],
 "Create Certificate Sigining Request": [
  null,
  "CSRを作成"
 ],
 "Orphan Keys": [
  null,
  "孤立鍵"
 ],
 "Delete Certificate": [
  null,
  "証明書を削除"
 ],
 "Are you sure you want to delete this certificate?": [
  null,
  "この証明書を削除しますか？"
 ],
 "Deleting Certificate ...": [
  null,
  "証明書を削除しています..."
 ],
 "Delete CSR": [
  null,
  "CSRを削除"
 ],
 "Are you sure you want to delete this CSR?": [
  null,
  "このCSRを削除しますか？"
 ],
 "Deleting CSR ...": [
  null,
  "CSRを削除しています..."
 ],
 "Delete Key": [
  null,
  "鍵を削除"
 ],
 "Are you sure you want to delete this Key?": [
  null,
  "この鍵を削除しますか？"
 ],
 "Deleting Key ...": [
  null,
  "鍵を削除しています..."
 ],
 "Warning - Altering CA Certificate Properties": [
  null,
  "警告 - CA証明書のプロパティを変更しています"
 ],
 "Removing the 'C' or 'T' flags from the SSL trust category could break all TLS connectivity to and from the server, are you sure you want to proceed?": [
  null,
  "SSL信頼カテゴリから 'C' または 'T' フラグを削除すると、サーバとのすべてのTLS接続が壊れる可能性があります。本当に続行しますか？"
 ],
 "Editing CA Certificate ...": [
  null,
  "CA証明書を変更しています..."
 ],
 "Change Trust Flags": [
  null,
  "信頼フラグを変更"
 ],
 "Successfully set cipher preferences.": [
  null,
  "暗号スイートの設定を行いました"
 ],
 "Error setting cipher preferences - $0": [
  null,
  "暗号スイートの設定に失敗しました - $0"
 ],
 "Saving Cipher Preferences ...": [
  null,
  "暗号スイートを保存しています..."
 ],
 "The current ciphers the server is accepting.  This list is only updated after a server restart.": [
  null,
  "現在サーバで使用できる暗号スイートです。このリストはサーバ再起動後にのみ更新されます。"
 ],
 "Enabled Ciphers": [
  null,
  "有効な暗号スイート"
 ],
 "enabled cipher list": [
  null,
  "有効な暗号スイートリスト"
 ],
 "The current ciphers the server supports": [
  null,
  "現在サーバがサポートしている暗号スイートです。"
 ],
 "Supported Ciphers": [
  null,
  "サポートされている暗号スイート"
 ],
 "Cipher Suite": [
  null,
  "暗号スイート"
 ],
 "Default Ciphers": [
  null,
  "デフォルトの暗号スイート"
 ],
 "All Ciphers": [
  null,
  "すべての暗号スイート"
 ],
 "No Ciphers": [
  null,
  "暗号スイートを使用しない"
 ],
 "Default cipher suite is chosen. It enables the default ciphers advertised by NSS except weak ciphers.": [
  null,
  "デフォルトの暗号スイートが選択されています。これにより、NSSによってデフォルトで提供される暗号スイートが有効になりますが、弱い暗号スイートは除外されます。"
 ],
 " Any data in the 'Allow Specific Ciphers' and 'Deny Specific Ciphers' fields will be cleaned after the restart.": [
  null,
  "'Allow Specific Ciphers'および'Deny Specific Ciphers'のフィールドにあるデータは、再起動後にクリアされます。"
 ],
 "Allow Specific Ciphers": [
  null,
  "特定の暗号スイートを許可"
 ],
 "Type a cipher": [
  null,
  "暗号スイートを入力"
 ],
 "Type a cipher...": [
  null,
  "暗号スイートを入力してください..."
 ],
 "Deny Specific Ciphers": [
  null,
  "特定の暗号スイートを拒否"
 ],
 "Cipher": [
  null,
  "暗号スイート"
 ],
 "Key Identifier": [
  null,
  "鍵識別子"
 ],
 "No Orphan keys": [
  null,
  "孤立鍵はありません"
 ],
 "Orphan keys": [
  null,
  "孤立鍵"
 ],
 "An orphan key is a private key in the NSS DB for which there is NO cert with the corresponding public key. An orphan key is created during CSR creation, when the certificate associated with a CSR has been imported into the NSS DB its orphan state will be removed.": [
  null,
  "孤立鍵とは、対応する公開鍵の証明書が存在しない、NSS DB 内の秘密鍵のことです。孤立鍵はCSRの作成時に作成され、CSRに関連する証明書がNSS DBにインポートされると、孤立状態は解除されます。"
 ],
 "Make sure an orphan key is not associated with a submitted CSR before you delete it.": [
  null,
  "削除する前に、孤立鍵が送信されたCSRと関連付けられていないことを確認してください。"
 ],
 "What is an orphan key?": [
  null,
  "孤立鍵とは？"
 ],
 "Subject DN": [
  null,
  "サブジェクトDN"
 ],
 "Modification Date": [
  null,
  "変更日"
 ],
 "No Certificate Signing Requests": [
  null,
  "証明書署名リクエストはありません"
 ],
 "Certificate Signing Requests": [
  null,
  "証明書署名リクエスト"
 ],
 "View CSR": [
  null,
  "CSRを表示"
 ],
 "Search CSRs": [
  null,
  "CSRを検索"
 ],
 "Nickname": [
  null,
  "ニックネーム"
 ],
 "Expiration Date": [
  null,
  "有効期限"
 ],
 "Issuer DN:": [
  null,
  "発行元DN:"
 ],
 "Trust Flags:": [
  null,
  "信頼フラグ:"
 ],
 "No Certificates": [
  null,
  "証明書がありません"
 ],
 "Certificates": [
  null,
  "証明書"
 ],
 "Edit Trust Flags": [
  null,
  "信頼フラグを編集"
 ],
 "Search Certificates": [
  null,
  "証明書を検索"
 ],
 "Issued By": [
  null,
  "発行元"
 ],
 "Effective Date": [
  null,
  "有効期限開始日"
 ],
 "Next Update": [
  null,
  "次回の更新日"
 ],
 "Type": [
  null,
  "タイプ"
 ],
 "View CRL": [
  null,
  "CRLを表示"
 ],
 "Delete CRL": [
  null,
  "CRLを削除"
 ],
 "Search CRL's": [
  null,
  "CRLを検索"
 ],
 "Successfully updated Audit Log settings": [
  null,
  "監査ログ設定を更新しました"
 ],
 "Error saving Audit Log settings - $0": [
  null,
  "監査ログ設定の保存に失敗しました - $0"
 ],
 "Error loading Audit Log configuration - $0": [
  null,
  "監査ログ設定の読み込みに失敗しました - $0"
 ],
 "Save Log Settings": [
  null,
  "ログ設定を保存"
 ],
 "Save Rotation Settings": [
  null,
  "ローテーション設定を保存"
 ],
 "Save Deletion Settings": [
  null,
  "削除設定を保存"
 ],
 "Enable audit logging (nsslapd-auditlog-logging-enabled).": [
  null,
  "監査ログを有効にするかどうかを設定します。(nsslapd-auditlog-logging-enabled)"
 ],
 "Enable Audit Logging": [
  null,
  "監査ログを有効にする"
 ],
 "Audit Log Location": [
  null,
  "監査ログのパス"
 ],
 "Enable audit logging (nsslapd-auditlog).": [
  null,
  "監査ログのパスおよびファイル名を設定します。(nsslapd-auditlog)"
 ],
 "Display Attributes": [
  null,
  "表示する属性"
 ],
 "Display attributes from the entry in the audit log (nsslapd-auditlog-display-attrs).": [
  null,
  "監査ログに表示する属性を設定します。(nsslapd-auditlog-display-attrs)"
 ],
 "Display all attributes from the entry in the audit log (nsslapd-auditlog-display-attrs).": [
  null,
  "監査ログにすべての属性を表示します。(nsslapd-auditlog-display-attrs)"
 ],
 "All Attributes": [
  null,
  "すべての属性を設定する"
 ],
 "Rotation Policy": [
  null,
  "ローテーションポリシー"
 ],
 "The maximum number of logs that are archived (nsslapd-auditlog-maxlogsperdir).": [
  null,
  "アーカイブされるログの最大数を設定します。(nsslapd-auditlog-maxlogsperdir)"
 ],
 "Maximum Number Of Logs": [
  null,
  "アーカイブされるログの最大数"
 ],
 "The maximum size of each log file in megabytes (nsslapd-auditlog-maxlogsize).": [
  null,
  "最大ログサイズをMB単位で設定します。(nsslapd-auditlog-maxlogsize)"
 ],
 "Maximum Log Size (in MB)": [
  null,
  "最大ログサイズ(MB)"
 ],
 "Rotate the log based this number of time units (nsslapd-auditlog-logrotationtime).": [
  null,
  "監査ログファイルのローテーションの間隔を設定します。(nsslapd-auditlog-logrotationtime)"
 ],
 "Create New Log Every ...": [
  null,
  "ログローテーション間隔"
 ],
 "minute": [
  null,
  "分"
 ],
 "hour": [
  null,
  "時間"
 ],
 "day": [
  null,
  "日"
 ],
 "week": [
  null,
  "週間"
 ],
 "month": [
  null,
  "ヵ月"
 ],
 "The time when the log should be rotated (nsslapd-auditlog-logrotationsynchour, nsslapd-auditlog-logrotationsyncmin).": [
  null,
  "監査ログをローテーションする時刻を設定します。(nsslapd-auditlog-logrotationsynchour, nsslapd-auditlog-logrotationsyncmin)"
 ],
 "Time Of Day": [
  null,
  "時刻"
 ],
 "Compress (gzip) the log after it's rotated.": [
  null,
  "ログをローテーション後にgz形式で圧縮します。"
 ],
 "Compress Rotated Logs": [
  null,
  "ローテーションされたログを圧縮"
 ],
 "Deletion Policy": [
  null,
  "削除ポリシー"
 ],
 "The server deletes the oldest archived log when the total of all the logs reaches this amount (nsslapd-auditlog-logmaxdiskspace).": [
  null,
  "監査ログが消費できる最大ディスク容量をMB単位で設定します。ログの容量が設定した容量を超えると、サーバはアーカイブされた最も古いログを削除します。(nsslapd-auditlog-logmaxdiskspace)"
 ],
 "Log Archive Exceeds (in MB)": [
  null,
  "ログの最大ディスク領域(MB)"
 ],
 "The server deletes the oldest archived log file when available disk space is less than this amount. (nsslapd-auditlog-logminfreediskspace).": [
  null,
  "許可される最小空きディスク容量をMB単位で設定します。空きディスク容量が設定した容量未満になると、十分なディスク領域が解放されるまでアーカイブされた最も古いログファイルを削除します。(nsslapd-auditlog-logminfreediskspace)"
 ],
 "Free Disk Space (in MB)": [
  null,
  "空きディスク領域(MB)"
 ],
 "Server deletes an old archived log file when it is older than the specified age. (nsslapd-auditlog-logexpirationtime).": [
  null,
  "指定された期間よりも古い場合、サーバはアーカイブされた古いログファイルを削除します。(nsslapd-auditlog-logexpirationtime)"
 ],
 "Log File is Older Than ...": [
  null,
  "ログの保存期間"
 ],
 "Loading Audit Log settings ...": [
  null,
  "監査ログの設定を読み込んでいます..."
 ],
 "Audit Log Settings": [
  null,
  "監査ログ設定"
 ],
 "Refresh log settings": [
  null,
  "ログ設定を更新します"
 ],
 "Successfully updated Audit Fail Log settings": [
  null,
  "監査失敗ログの設定を更新しました"
 ],
 "Error saving Audit Fail Log settings - $0": [
  null,
  "監査失敗ログ設定の保存に失敗しました - $0"
 ],
 "Error loading Audit Fail Log configuration - $0": [
  null,
  "監査失敗ログ設定の読み込みに失敗しました - $0"
 ],
 "Enable audit fail logging (nsslapd-auditfaillog-logging-enabled).": [
  null,
  "監査失敗ログを有効にするかどうかを設定します。(nsslapd-auditfaillog-logging-enabled)"
 ],
 "Enable Audit Fail Logging": [
  null,
  "監査失敗ログを有効にする"
 ],
 "Enable audit fail logging (nsslapd-auditfaillog).": [
  null,
  "監査失敗ログのパスおよびファイル名を設定します。(nsslapd-auditfaillog)"
 ],
 "Audit Fail Log Location": [
  null,
  "監査失敗ログのパス"
 ],
 "The maximum number of logs that are archived (nsslapd-auditfaillog-maxlogsperdir).": [
  null,
  "アーカイブされるログの最大数を設定します。(nsslapd-auditfaillog-maxlogsperdir)"
 ],
 "The maximum size of each log file in megabytes (nsslapd-auditfaillog-maxlogsize).": [
  null,
  "最大ログサイズをMB単位で設定します。(nsslapd-auditfaillog-maxlogsize)"
 ],
 "Rotate the log based this number of time units (nsslapd-auditfaillog-logrotationtime).": [
  null,
  "監査失敗ログファイルのローテーションの間隔を設定します。(nsslapd-auditfaillog-logrotationtime)"
 ],
 "The time when the log should be rotated (nsslapd-auditfaillog-logrotationsynchour, nsslapd-auditfaillog-logrotationsyncmin).": [
  null,
  "監査失敗ログをローテーションする時刻を設定します。(nsslapd-auditfaillog-logrotationsynchour, nsslapd-auditfaillog-logrotationsyncmin)"
 ],
 "The server deletes the oldest archived log when the total of all the logs reaches this amount (nsslapd-auditfaillog-logmaxdiskspace).": [
  null,
  "監査失敗ログが消費できる最大ディスク容量をMB単位で設定します。ログの容量が設定した容量を超えると、サーバはアーカイブされた最も古いログを削除します。(nsslapd-auditfaillog-logmaxdiskspace)"
 ],
 "The server deletes the oldest archived log file when available disk space is less than this amount. (nsslapd-auditfaillog-logminfreediskspace).": [
  null,
  "許可される最小空きディスク容量をMB単位で設定します。空きディスク容量が設定した容量未満になると、十分なディスク領域が解放されるまでアーカイブされた最も古いログファイルを削除します。(nsslapd-auditfaillog-logminfreediskspace)"
 ],
 "Server deletes an old archived log file when it is older than the specified age. (nsslapd-auditfaillog-logexpirationtime).": [
  null,
  "指定された期間よりも古い場合、サーバはアーカイブされた古いログファイルを削除します。(nsslapd-auditfaillog-logexpirationtime)"
 ],
 "Loading Audit Fail Log Settings ...": [
  null,
  "監査失敗ログの設定を読み込んでいます..."
 ],
 "Audit Fail Log Settings": [
  null,
  "監査失敗ログ設定"
 ],
 "Trace Function Calls": [
  null,
  "関数呼び出しのトレース"
 ],
 "(level 1)": [
  null,
  "(レベル 1)"
 ],
 "Packet Handling": [
  null,
  "パケット処理"
 ],
 "(level 2)": [
  null,
  "(レベル 2)"
 ],
 "Heavy Trace Output": [
  null,
  "詳細なトレース出力"
 ],
 "(level 4)": [
  null,
  "(レベル 4)"
 ],
 "Connection Management": [
  null,
  "接続管理"
 ],
 "(level 8)": [
  null,
  "(レベル 8)"
 ],
 "Packets Sent & Received": [
  null,
  "送信および受信したパケット"
 ],
 "(level 16)": [
  null,
  "(レベル 16)"
 ],
 "Search Filter Processing": [
  null,
  "検索フィルタの処理"
 ],
 "(level 32)": [
  null,
  "(レベル 32)"
 ],
 "Config File Processing": [
  null,
  "設定ファイルの処理"
 ],
 "(level 64)": [
  null,
  "(レベル 64)"
 ],
 "Access Control List Processing": [
  null,
  "アクセス制御リストの処理"
 ],
 "(level 128)": [
  null,
  "(レベル 128)"
 ],
 "Log Entry Parsing": [
  null,
  "ログエントリの解析"
 ],
 "(level 2048)": [
  null,
  "(レベル 2048)"
 ],
 "Housekeeping": [
  null,
  "メンテナンス"
 ],
 "(level 4096)": [
  null,
  "(レベル 4096)"
 ],
 "Replication": [
  null,
  "レプリケーション"
 ],
 "(level 8192)": [
  null,
  "(レベル 8192)"
 ],
 "(level 32768)": [
  null,
  "(レベル 32768)"
 ],
 "(level 65536)": [
  null,
  "(レベル 65536)"
 ],
 "Access Control Summary": [
  null,
  "アクセス制御の概要"
 ],
 "(level 262144)": [
  null,
  "(レベル 262144)"
 ],
 "Backend Database": [
  null,
  "データベースのバックエンド"
 ],
 "(level 524288)": [
  null,
  "(レベル 524288)"
 ],
 "Password Policy": [
  null,
  "パスワードポリシー"
 ],
 "(level 1048576)": [
  null,
  "(レベル 1048576)"
 ],
 "Logging Level": [
  null,
  "ログレベル"
 ],
 "Successfully updated Error Log settings": [
  null,
  "エラーログの設定を更新しました"
 ],
 "Error saving Error Log settings - $0": [
  null,
  "エラーログ設定の保存に失敗しました - $0"
 ],
 "Error loading Error Log configuration - $0": [
  null,
  "エラーログ設定の読み込みに失敗しました - $0"
 ],
 "Enable Error logging (nsslapd-errorlog-logging-enabled).": [
  null,
  "エラーログを有効にするかどうかを設定します。(nsslapd-errorlog-logging-enabled)"
 ],
 "Enable Error Logging": [
  null,
  "エラーログを有効にする"
 ],
 "Error Log Location": [
  null,
  "エラーログのパス"
 ],
 "Enable Error logging (nsslapd-errorlog).": [
  null,
  "エラーログのパスおよびファイル名を設定します。(nsslapd-errorlog)"
 ],
 "Hide Verbose Logging Levels": [
  null,
  "ログレベルの詳細を非表示"
 ],
 "Show Verbose Logging Levels": [
  null,
  "ログレベルの詳細を表示"
 ],
 "The maximum number of logs that are archived (nsslapd-errorlog-maxlogsperdir).": [
  null,
  "アーカイブされるログの最大数を設定します。(nsslapd-errorlog-maxlogsperdir)"
 ],
 "The maximum size of each log file in megabytes (nsslapd-errorlog-maxlogsize).": [
  null,
  "最大エラーログサイズをMB単位で設定します。(nsslapd-errorlog-maxlogsize)"
 ],
 "Rotate the log based this number of time units (nsslapd-errorlog-logrotationtime).": [
  null,
  "エラーログファイルのローテーションの間隔を設定します。(nsslapd-errorlog-logrotationtime)"
 ],
 "The time when the log should be rotated (nsslapd-errorlog-logrotationsynchour, nsslapd-errorlog-logrotationsyncmin).": [
  null,
  "エラーログをローテーションする時刻を設定します。(nsslapd-errorlog-logrotationsynchour, nsslapd-errorlog-logrotationsyncmin)"
 ],
 "The server deletes the oldest archived log when the total of all the logs reaches this amount (nsslapd-errorlog-logmaxdiskspace).": [
  null,
  "エラーログが消費できる最大ディスク容量をMB単位で設定します。ログの容量が設定した容量を超えると、サーバはアーカイブされた最も古いログを削除します。(nsslapd-errorlog-logmaxdiskspace)"
 ],
 "The server deletes the oldest archived log file when available disk space is less than this amount. (nsslapd-errorlog-logminfreediskspace).": [
  null,
  "許可される最小空きディスク容量をMB単位で設定します。空きディスク容量が設定した容量未満になると、十分なディスク領域が解放されるまでアーカイブされた最も古いログファイルを削除します。(nsslapd-errorlog-logminfreediskspace)"
 ],
 "Server deletes an old archived log file when it is older than the specified age. (nsslapd-errorlog-logexpirationtime).": [
  null,
  "指定された期間よりも古い場合、サーバはアーカイブされた古いログファイルを削除します。(nsslapd-errorlog-logexpirationtime)"
 ],
 "Loading Error Log Settings ...": [
  null,
  "エラーログ設定を読み込んでいます..."
 ],
 "Error Log Settings": [
  null,
  "エラーログ設定"
 ],
 "Error loading server configuration - $0": [
  null,
  "サーバ構成の読み込みに失敗しました - $0"
 ],
 "Successfully updated LDAPI configuration": [
  null,
  "LDAPI設定を更新しました"
 ],
 "Error updating LDAPI configuration - $0": [
  null,
  "LDAPI設定の更新に失敗しました - $0"
 ],
 "The Directory Server attribute to map system UIDs to user entries (nsslapd-ldapiuidnumbertype).": [
  null,
  "システムUIDをユーザエントリにマップするためのディレクトリサーバ属性を設定します。(nsslapd-ldapiuidnumbertype)"
 ],
 "LDAPI UID Number Attribute": [
  null,
  "LDAPI UID番号の属性"
 ],
 "The Directory Server attribute to map system GUIDs to user entries (nsslapd-ldapigidnumbertype).": [
  null,
  "システムGUIDsをユーザエントリにマップするためのディレクトリサーバ属性を設定します。(nsslapd-ldapiuidnumbertype)"
 ],
 "LDAPI GID Number Attribute": [
  null,
  "LDAPI GID番号の属性"
 ],
 "The subtree to search for user entries to use for autobind. (nsslapd-ldapientrysearchbase).": [
  null,
  "自動バインドに使用するユーザエントリを検索するサブツリーを指定します。(nsslapd-ldapientrysearchbase)"
 ],
 "LDAPI Entry Search Base": [
  null,
  "LDAPI エントリの検索ベース"
 ],
 "The Unix socket file (nsslapd-ldapifilepath).  The UI requires this exact path so it is a read-only setting.": [
  null,
  "UNIXソケットファイルのパスを設定します。UIでは正確なパスが必要なため、読み取り専用設定になります。(nsslapd-ldapifilepath)"
 ],
 "LDAPI Socket File Path": [
  null,
  "LDAPI ソケットのファイルパス"
 ],
 "Map the Unix root entry to this Directory Manager DN (nsslapd-ldapimaprootdn).  The UI requires this to be set to the current root DN so it is a read-only setting. The attribute is obsolete. Current root DN is used.": [
  null,
  "Unix ルートエントリにマップするディレクトリマネージャDNを設定します(nsslapd-ldapimaprootdn)UIでは、現在のルートDNに設定する必要があるため、読み取り専用設定になります。この属性は非推奨です。現在のルートDNが使用されます。"
 ],
 "LDAPI Map To Root DN": [
  null,
  "ルートDNのためのLDAPIマップ"
 ],
 "Map regular system users to Directory Server entries (nsslapd-ldapimaptoentries).": [
  null,
  "通常のシステムユーザに対してデータベースエントリへのマッピングを有効にするかどうかを設定します。(nsslapd-ldapimaptoentries)"
 ],
 "Map System Users to Database Entries": [
  null,
  "システムユーザをデータベースエントリにマッピングする"
 ],
 "Loading LDAPI configuration ...": [
  null,
  "LDAPI設定を読み込んでいます..."
 ],
 "LDAPI & AutoBind Settings": [
  null,
  "LDAPIと自動バインドの設定"
 ],
 "Refresh LDAPI settings": [
  null,
  "LDAPI設定を更新します"
 ],
 "Successfully updated Security Log settings": [
  null,
  "セキュリティログ設定を更新しました"
 ],
 "Error saving Security Log settings - $0": [
  null,
  "セキュリティログ設定の保存に失敗しました - $0"
 ],
 "Error loading Security Log configuration - $0": [
  null,
  "セキュリティログ設定の読み込みに失敗しました - $0"
 ],
 "Enable security logging (nsslapd-securitylog-logging-enabled).": [
  null,
  "セキュリティログを有効にするかどうかを設定します(nsslapd-securitylog-logging-enabled)"
 ],
 "Enable Security Logging": [
  null,
  "セキュリティログを有効にする"
 ],
 "Security Log Location": [
  null,
  "セキュリティログのパス"
 ],
 "Enable security logging (nsslapd-securitylog).": [
  null,
  "セキュリティメッセージを記録するために使用されるログのパスおよびファイル名を設定します。(nsslapd-securitylog)"
 ],
 "Disable security log buffering for faster troubleshooting, but this will impact server performance (nsslapd-securitylog-logbuffering).": [
  null,
  "トラブルシューティングを迅速に行うために、セキュリティログのバッファリングを無効にするかどうかを設定します。無効にした場合、サーバのパフォーマンスに影響を与える可能性があります。(nsslapd-securitylog-logbuffering)"
 ],
 "Security Log Buffering Enabled": [
  null,
  "セキュリティログのバッファリングを有効にする"
 ],
 "Hide Logging Levels": [
  null,
  "ログレベルを非表示"
 ],
 "Show Logging Levels": [
  null,
  "ログレベルを表示"
 ],
 "The maximum number of logs that are archived (nsslapd-securitylog-maxlogsperdir).": [
  null,
  "アーカイブされるログの最大数を設定します(nsslapd-securitylog-maxlogsperdir)"
 ],
 "The maximum size of each log file in megabytes (nsslapd-securitylog-maxlogsize).": [
  null,
  "最大ログサイズをMB単位で設定します。(nsslapd-securitylog-maxlogsize)"
 ],
 "Rotate the log based this number of time units (nsslapd-securitylog-logrotationtime).": [
  null,
  "セキュリティログファイルのローテーションの間隔を設定します。(nsslapd-securitylog-logrotationtime)"
 ],
 "The time when the log should be rotated (nsslapd-securitylog-logrotationsynchour, nsslapd-securitylog-logrotationsyncmin).": [
  null,
  "セキュリティログをローテーションする時刻を設定します。(nsslapd-securitylog-logrotationsynchour, nsslapd-securitylog-logrotationsyncmin)"
 ],
 "The server deletes the oldest archived log when the total of all the logs reaches this amount (nsslapd-securitylog-logmaxdiskspace).": [
  null,
  "セキュリティログが消費できる最大ディスク容量をMB単位で設定します。ログの容量が設定した容量を超えると、サーバはアーカイブされた最も古いログを削除します。(nsslapd-securitylog-logmaxdiskspace)"
 ],
 "The server deletes the oldest archived log file when available disk space is less than this amount. (nsslapd-securitylog-logminfreediskspace).": [
  null,
  "許可される最小空きディスク容量をMB単位で設定します。空きディスク容量が設定した容量未満になると、十分なディスク領域が解放されるまでアーカイブされた最も古いログファイルを削除します。(nsslapd-securitylog-logminfreediskspace)"
 ],
 "Server deletes an old archived log file when it is older than the specified age. (nsslapd-securitylog-logexpirationtime).": [
  null,
  "指定された期間よりも古い場合、サーバはアーカイブされた古いログファイルを削除します。(nsslapd-securitylog-logexpirationtime)"
 ],
 "Loading Security Log Settings ...": [
  null,
  "セキュリティログの設定を読み込んでいます..."
 ],
 "Security Log Settings": [
  null,
  "セキュリティログ設定"
 ],
 "Mapping Name": [
  null,
  "マッピング名"
 ],
 "Priority": [
  null,
  "プライオリティ"
 ],
 "Regular Expression:": [
  null,
  "正規表現:"
 ],
 "No SASL Mappings": [
  null,
  "SASLマッピングがありません"
 ],
 "SASL Mappings": [
  null,
  "SASLマッピング"
 ],
 "Edit Mapping": [
  null,
  "マッピングを編集"
 ],
 "Delete Mapping": [
  null,
  "マッピングを削除"
 ],
 "Search Mappings": [
  null,
  "マッピングの検索"
 ],
 "Successfully updated Directory Manager configuration": [
  null,
  "ディレクトリ管理者の設定を更新しました"
 ],
 "Error updating Directory Manager configuration - $0": [
  null,
  "ディレクトリ管理者の設定更新に失敗しました - $0"
 ],
 "Error reloading Directory Manager configuration - $0": [
  null,
  "ディレクトリ管理者設定の読み込みに失敗しました - $0"
 ],
 "Successfully updated Disk Monitoring configuration": [
  null,
  "ディスク監視設定を更新しました"
 ],
 "Error updating Disk Monitoring configuration - $0": [
  null,
  "ディスク監視設定の更新に失敗しました - $0"
 ],
 "Error reloading Disk Monitoring configuration - $0": [
  null,
  "ディスク監視設定の読み込みに失敗しました - $0"
 ],
 "Successfully updated Advanced configuration": [
  null,
  "詳細設定を更新しました"
 ],
 "Error updating Advanced configuration - $0": [
  null,
  "詳細設定の更新に失敗しました - $0"
 ],
 "Error loading Advanced configuration - $0": [
  null,
  "詳細設定の読み込みに失敗しました - $0"
 ],
 "Successfully updated server configuration.  These changes require the server to be restarted to take effect.": [
  null,
  "サーバ設定を更新しました。変更を反映するには、サーバを再起動してください。"
 ],
 "Error updating server configuration - $0": [
  null,
  "サーバ設定の更新に失敗しました - $0"
 ],
 "Error reloading server configuration - $0": [
  null,
  "サーバ設定の読み込みに失敗しました - $0"
 ],
 "The available disk space, in bytes, that will trigger the shutdown process. Default is 2mb. Once below half of the threshold then we enter the shutdown mode. Value range: 4096 - 9223372036854775807. (nsslapd-disk-monitoring-threshold)": [
  null,
  "サーバに十分なディスク領域があるかどうかを評価するために使用する閾値をバイト単位で設定します。領域がこの閾値の半分に達すると、サーバはシャットダウンプロセスを開始します。値は次の範囲内で設定してください: 4096 - 9223372036854775807。(nsslapd-disk-monitoring-threshold)"
 ],
 "Disk Monitoring Threshold": [
  null,
  "ディスク監視閾値"
 ],
 "Value must be greater than or equal to 4096": [
  null,
  "4096以上の値を設定してください"
 ],
 "How many minutes to wait to allow an admin to clean up disk space before shutting slapd down. The default is 60 minutes. (nsslapd-disk-monitoring-grace-period)": [
  null,
  "slapdをシャットダウンするまでに待機する分数を指定します。管理者がこの待機時間中にディスク容量をクリーンアップすることで、強制的なシャットダウンを回避することができます。デフォルトは60分です。(nsslapd-disk-monitoring-grace-period)"
 ],
 "Disk Monitoring Grace Period": [
  null,
  "ディスク監視の猶予時間"
 ],
 "When disk space gets critically low do not remove logs to free up disk space. (nsslapd-disk-monitoring-logging-critical)": [
  null,
  "ディスク容量が極端に低下した場合、ディスク容量を解放するためにログを削除しないよう設定します。(nsslapd-disk-monitoring-logging-critical)"
 ],
 "Preserve Logs Even If Disk Space Gets Low": [
  null,
  "ディスク容量が少なくなってもログを保持する"
 ],
 "Loading Server Settings ...": [
  null,
  "サーバ設定を読み込んでいます..."
 ],
 "Server Settings": [
  null,
  "サーバ設定"
 ],
 "Refresh configuration settings": [
  null,
  "サーバ設定を更新します"
 ],
 "The version of the Directory Server package": [
  null,
  "ディレクトリサーバパッケージのバージョンが表示されます"
 ],
 "Server Version": [
  null,
  "サーババージョン"
 ],
 "The server's local hostname (nsslapd-localhost).": [
  null,
  "サーバのホスト名を設定します。(nsslapd-localhost)"
 ],
 "Server Hostname": [
  null,
  "サーバホスト名"
 ],
 "The server's port number (nsslapd-port).": [
  null,
  "標準のLDAP通信に使用されるポート番号を設定します。(nsslapd-port)"
 ],
 "LDAP Port": [
  null,
  "LDAPポート"
 ],
 "The server's secure port number (nsslapd-secureport).": [
  null,
  "TLS通信に使用されるポート番号を設定します。(nsslapd-secureport)"
 ],
 "LDAPS Port": [
  null,
  "LDAPSポート"
 ],
 "This parameter can be used to restrict the Directory Server instance to a single IP interface (hostname, or IP address).  Requires restart. (nsslapd-listenhost).": [
  null,
  "ディレクトリサーバインスタンスを単一のIPインターフェース(ホスト名またはIPアドレス)に制限するために使用できます。設定を反映するには再起動が必要です。(nsslapd-listenhost)"
 ],
 "Listen Host Address": [
  null,
  "リッスンホストアドレス"
 ],
 "Invalid backup directory path!": [
  null,
  "バックアップディレクトリのパスが無効です！"
 ],
 "The location where database backups are stored (nsslapd-bakdir).": [
  null,
  "データベースのバックアップが保存される場所を設定します。(nsslapd-bakdir)"
 ],
 "Backup Directory": [
  null,
  "バックアップディレクトリ"
 ],
 "Invalid LDIF directory path!": [
  null,
  "LDIFディレクトリパスが無効です！"
 ],
 "The location where the server's LDIF files are located (nsslapd-ldifdir).": [
  null,
  "サーバのLDIFファイルが保存される場所を設定します。(nsslapd-ldifdir)"
 ],
 "LDIF File Directory": [
  null,
  "LDIFファイルディレクトリ"
 ],
 "Invalid schema directory path!": [
  null,
  "スキーマディレクトリパスが無効です！"
 ],
 "The location for the servers custom schema files. (nsslapd-schemadir).": [
  null,
  "サーバのカスタムスキーマファイルの場所を設定します。(nsslapd-schemadir)"
 ],
 "Schema Directory": [
  null,
  "スキーマディレクトリ"
 ],
 "Invalid certificate directory path!": [
  null,
  "証明書ディレクトリパスが無効です！"
 ],
 "The location of the server's certificates (nsslapd-certdir).": [
  null,
  "サーバの証明書の場所を設定します。(nsslapd-certdir)"
 ],
 "Certificate Directory": [
  null,
  "証明書ディレクトリ"
 ],
 "Directory Manager": [
  null,
  "ディレクトリ管理者"
 ],
 "The DN of the unrestricted directory manager (nsslapd-rootdn).": [
  null,
  "特権を持つディレクトリ管理者のDNが表示されます。(nsslapd-rootdn)"
 ],
 "Directory Manager DN": [
  null,
  "ディレクトリ管理者DN"
 ],
 "The password for the Root DN/Directory Manager (nsslapd-rootpw).": [
  null,
  "ルートDN/ディレクトリ管理者のパスワードを設定します。(nsslapd-rootpw)"
 ],
 "Directory Manager Password": [
  null,
  "ディレクトリ管理者パスワード"
 ],
 "Confirm the Directory Manager password": [
  null,
  "ディレクトリ管理者パスワードを再入力してください"
 ],
 "Set the Directory Manager password storage scheme (nsslapd-rootpwstoragescheme).": [
  null,
  "ディレクトリ管理者のパスワードを暗号化する方法を設定します。(nsslapd-rootpwstoragescheme)"
 ],
 "Disk Monitoring": [
  null,
  "ディスク監視"
 ],
 "Enable Disk Space Monitoring": [
  null,
  "ディスク容量監視を有効にする"
 ],
 "Enable schema checking (nsslapd-schemacheck).": [
  null,
  "エントリが追加または変更されたときにデータベーススキーマを適用するかどうかを設定します。(nsslapd-schemacheck)"
 ],
 "Enable Schema Checking": [
  null,
  "スキーマチェックを有効にする"
 ],
 "Enable attribute syntax checking (nsslapd-syntaxcheck).": [
  null,
  "属性の構文チェックを行うかどうかを設定します。(nsslapd-syntaxcheck)"
 ],
 "Enable Attribute Syntax Checking": [
  null,
  "属性の構文チェックを有効にする"
 ],
 "Enable plugins to log access and audit events.  (nsslapd-plugin-logging).": [
  null,
  "アクセスと監査イベントのログ記録をプラグインで有効にするかどうかを設定します。(nsslapd-plugin-logging)"
 ],
 "Enable Plugin Logging": [
  null,
  "プラグインのログ記録を有効にする"
 ],
 "Enable syntax logging (nsslapd-syntaxlogging).": [
  null,
  "構文検証の失敗をエラーログに記録するかどうかを設定します。(nsslapd-syntaxlogging)"
 ],
 "Enable Attribute Syntax Logging": [
  null,
  "属性構文ログを有効にする"
 ],
 "Enable Plugin Bind DN Tracking": [
  null,
  "プラグインのバインドDN追跡を有効にする"
 ],
 "Enabling this feature will write new operational attributes to the modified entry: internalModifiersname & internalCreatorsname. These new attributes contain the plugin DN, while modifiersname will be the original binding entry that triggered the update. (nsslapd-plugin-binddn-tracking).": [
  null,
  "この機能を有効にすると、変更されたエントリに新しい操作属性が書き込まれます。:(internalModifiersname および internalCreatorsname) これらの新しい属性には、プラグインのDNが含まれます。一方、modifiersname は更新をトリガーした元のバインドエントリになります。(nsslapd-plugin-binddn-tracking)"
 ],
 "Allows non-standard characters in attribute names to be used for backwards compatibility with older servers (nsslapd-attribute-name-exceptions).": [
  null,
  "非標準の文字を属性名に使用できるようにします。これにより、古いサーバとの後方互換性を保つことができるようになります。(nsslapd-attribute-name-exceptions)"
 ],
 "Allow Attribute Naming Exceptions": [
  null,
  "属性名の例外を許可する"
 ],
 "Strict DN Syntax Validation": [
  null,
  "厳密にDN構文を検証する"
 ],
 "Enables strict syntax validation for DNs, according to section 3 in RFC 4514 (nsslapd-dn-validate-strict).": [
  null,
  "RFC 4514のセクション3に基づいてDNの厳密な構文検証を有効にします。(nsslapd-dn-validate-strict)"
 ],
 "For USN plugin - maintain unique USNs across all back end databases (nsslapd-entryusn-global).": [
  null,
  "USNプラグイン用に、バックエンドデータベース全体で一意のUSNを維持するかどうかを設定します。(nsslapd-entryusn-global)"
 ],
 "Maintain Unique USNs Across All Backends": [
  null,
  "すべてのバックエンドで一意のUSNを維持する"
 ],
 "Ignore replication time skew when acquiring a replica to start a replciation session (nsslapd-ignore-time-skew).": [
  null,
  "レプリカを取得してレプリケーションセッションを開始する際に、レプリケーションの時刻のずれを無視するかどうかを設定します。(nsslapd-ignore-time-skew)"
 ],
 "Ignore CSN Time Skew": [
  null,
  "CSNの時刻のずれを無視する"
 ],
 "Make entire server read-only (nsslapd-readonly)": [
  null,
  "サーバ全体を読み取り専用にするかどうかを設定します。(nsslapd-readonly)"
 ],
 "Server Read-Only": [
  null,
  "サーバを読み取り専用にする"
 ],
 "Allow anonymous binds to the server (nsslapd-allow-anonymous-access).": [
  null,
  "サーバへの匿名バインドを許可するかどうかを設定します。(nsslapd-allow-anonymous-access)"
 ],
 "Allow Anonymous Access": [
  null,
  "匿名アクセスを許可"
 ],
 "Allows anonymous search and read access to search the root DSE itself, but restricts access to all other directory entries. ": [
  null,
  "ルートDSE自体を検索するための匿名検索および読み取りアクセスを許可しますが、他のすべてのディレクトリエントリへのアクセスを制限します。"
 ],
 "The DN of a template entry containing the resource limits to apply to anonymous connections (nsslapd-anonlimitsdn).": [
  null,
  "匿名接続に適用するリソース制限を含むテンプレートエントリのDNを設定します。(nsslapd-anonlimitsdn)"
 ],
 "Anonymous Resource Limits DN": [
  null,
  "匿名接続のリソース制限DN"
 ],
 "Successfully updated tuning configuration": [
  null,
  "チューニング設定を更新しました"
 ],
 "Error updating tuning configuration - $0": [
  null,
  "チューニング設定の更新に失敗しました - $0"
 ],
 "Loading Tuning Configuration ...": [
  null,
  "チューニング設定を読み込んでいます..."
 ],
 "Tuning & Limits": [
  null,
  "チューニングとリミット"
 ],
 "Refresh settings": [
  null,
  "設定を更新します"
 ],
 "The number of worker threads that handle database operations.  Set to '-1' for enable auto tuning. (nsslapd-threadnumber).": [
  null,
  "データベース操作を処理するワーカースレッドの数を設定します。自動チューニングを有効にするには、'-1'を設定します。(nsslapd-threadnumber)"
 ],
 "Number Of Worker Threads": [
  null,
  "ワーカースレッドの数"
 ],
 "The maximum number of seconds allocated for a search request.  Set to '-1' to disable the time limit (nsslapd-timelimit).": [
  null,
  "検索要求に割り当てられる最大秒数を設定します。タイムリミットを無効にするには、'-1'を設定します。(nsslapd-timelimit)"
 ],
 "Search Time Limit": [
  null,
  "検索時間制限"
 ],
 "The maximum number of entries to return from a search operation.  Set to '-1' to disable the size limit (nsslapd-sizelimit).": [
  null,
  "検索操作から返されるエントリ数の最大値を設定します。サイズ制限を無効にするには、'-1'を設定します。(nsslapd-sizelimit)"
 ],
 "Search Size Limit": [
  null,
  "検索サイズ制限"
 ],
 "The maximum number of entries to return from a paged search operation. Set to '-1' to disable the size limit (nsslapd-pagedsizelimit).": [
  null,
  "ページングされた検索操作から返されるエントリ数の最大値を設定します。サイズ制限を無効にするには、'-1'を設定します。(nsslapd-pagedsizelimit)"
 ],
 "Paged Search Size Limit": [
  null,
  "ページングされた検索サイズ制限"
 ],
 "Sets the amount of time in seconds after which an idle LDAP client connection is closed by the server (nsslapd-idletimeout).": [
  null,
  "LDAPクライアント接続がアイドル状態になってから、サーバによって切断されるまでの時間を秒単位で設定します。(nsslapd-idletimeout)"
 ],
 "Idle Connection Timeout": [
  null,
  "アイドル接続のタイムアウト"
 ],
 "Sets the amount of time in milliseconds after which the connection to a stalled LDAP client is closed (nsslapd-ioblocktimeout).": [
  null,
  "クライアントへの接続が停止した場合に、接続が閉じられるまでの時間をミリ秒単位で設定します。(nsslapd-ioblocktimeout)"
 ],
 "I/O Block Timeout": [
  null,
  "I/Oブロックタイムアウト"
 ],
 "Sets the I/O wait time for all outbound LDAP connections (nsslapd-outbound-ldap-io-timeout).": [
  null,
  "すべてのアウトバウンドLDAP接続のI/O待機時間を制限します。(nsslapd-outbound-ldap-io-timeout)"
 ],
 "Outbound IO Timeout": [
  null,
  "出力IOタイムアウト"
 ],
 "The maximum size in bytes allowed for an incoming message (nsslapd-maxbersize).": [
  null,
  "受信メッセージの最大バイト数を設定します。(nsslapd-maxbersize)"
 ],
 "Maximum BER Size": [
  null,
  "最大BERサイズ"
 ],
 "The maximum allowed SASL IO packet size that the server will accept (nsslapd-maxsasliosize).": [
  null,
  "サーバが受け入れるSASL IOパケットの最大サイズを設定します。(nsslapd-maxsasliosize)"
 ],
 "Maximum SASL IO Size": [
  null,
  "最大SASL IOサイズ"
 ],
 "The maximum length for how long the connection queue for the socket can grow before refusing connections (nsslapd-listen-backlog-size).": [
  null,
  "接続を拒否する前にソケット(sockfd)のキューを拡張する最大期間を設定します。(nsslapd-listen-backlog-size)"
 ],
 "Listen Backlog Size": [
  null,
  "バックログサイズのリッスン"
 ],
 "Sets how deep a nested search filter is analysed (nsslapd-max-filter-nest-level).": [
  null,
  "ネストされた検索フィルタの解析深度を設定します。(nsslapd-max-filter-nest-level)"
 ],
 "Maximum Nested Filter Level": [
  null,
  "最大ネストフィルタレベル"
 ],
 "Disable DNS reverse entries for outgoing connections (nsslapd-connection-nocanon).": [
  null,
  "出力接続のDNS逆エントリを無効にするかどうかを設定します。(nsslapd-connection-nocanon)"
 ],
 "Disable Reverse DNS Lookups": [
  null,
  "逆DNSルックアップを無効にする"
 ],
 "Sets the worker threads to continuously read a connection without passing it back to the polling mechanism. (nsslapd-enable-turbo-mode).": [
  null,
  "接続をポーリングメカニズムに戻さずに、ワーカースレッドを継続的に読み取るようにするかどうかを設定します。(nsslapd-enable-turbo-mode)"
 ],
 "Enable Connection Turbo Mode": [
  null,
  "接続ターボモードを有効にする"
 ],
 "Disable the virtual attribute lookup in a search entry (nsslapd-ignore-virtual-attrs).": [
  null,
  "検索エントリの仮想属性の検索を無効にするかどうかを設定します。(nsslapd-ignore-virtual-attrs)"
 ],
 "Disable Virtual Attribute Lookups": [
  null,
  "仮想属性の検索を無効にする"
 ],
 "Enable the normalized DN cache.  Each thread has its own cache (nsslapd-ndn-cache-enabled).": [
  null,
  "正規化されたDNキャッシュを有効にするかどうかを設定します。各スレッドに独自のキャッシュがあります。(nsslapd-ndn-cache-enabled)"
 ],
 "Enable Normalized DN Cache": [
  null,
  "正規化されたDNキャッシュを有効にする"
 ],
 "The max NDN cache size in bytes (nsslapd-ndn-cache-max-size).": [
  null,
  "最大NDNキャッシュサイズを設定します。(nsslapd-ndn-cache-max-size)"
 ],
 "NDN Max Cache Size": [
  null,
  "NDN最大キャッシュサイズ"
 ],
 "Default Logging": [
  null,
  "デフォルトログ"
 ],
 "Internal Operations": [
  null,
  "内部操作"
 ],
 "Entry Access and Referrals": [
  null,
  "エントリアクセスと参照"
 ],
 "Successfully updated Access Log settings": [
  null,
  "アクセスログの設定を更新しました"
 ],
 "Error saving Access Log settings - $0": [
  null,
  "アクセスログ設定の保存に失敗しました - $0"
 ],
 "Error loading Access Log configuration - $0": [
  null,
  "アクセスログ設定の読み込みに失敗しました - $0"
 ],
 "Enable access logging (nsslapd-accesslog-logging-enabled).": [
  null,
  "アクセスログを有効にするかどうかを設定します。(nsslapd-accesslog-logging-enabled)"
 ],
 "Enable Access Logging": [
  null,
  "アクセスログを有効にする"
 ],
 "Access Log Location": [
  null,
  "アクセスログのパス"
 ],
 "Enable access logging (nsslapd-accesslog).": [
  null,
  "アクセスログのパスおよびファイル名を設定します。(nsslapd-accesslog)"
 ],
 "Disable access log buffering for faster troubleshooting, but this will impact server performance (nsslapd-accesslog-logbuffering).": [
  null,
  "トラブルシューティングを迅速に行うために、アクセスログのバッファリングを無効にするかどうかを設定します。無効にした場合、サーバのパフォーマンスに影響を与える可能性があります。(nsslapd-accesslog-logbuffering)"
 ],
 "Access Log Buffering Enabled": [
  null,
  "アクセスログバッファリングを有効にする"
 ],
 "The maximum number of logs that are archived (nsslapd-accesslog-maxlogsperdir).": [
  null,
  "アーカイブされるログの最大数を設定します。(nsslapd-accesslog-maxlogsperdir)"
 ],
 "The maximum size of each log file in megabytes (nsslapd-accesslog-maxlogsize).": [
  null,
  "最大ログサイズをMB単位で設定します。(nsslapd-accesslog-maxlogsize)"
 ],
 "Rotate the log based this number of time units (nsslapd-accesslog-logrotationtime).": [
  null,
  "アクセスログファイルのローテーションの間隔を設定します。(nsslapd-accesslog-logrotationtime)"
 ],
 "The time when the log should be rotated (nsslapd-accesslog-logrotationsynchour, nsslapd-accesslog-logrotationsyncmin).": [
  null,
  "アクセスログをローテーションする時刻を設定します。(nsslapd-accesslog-logrotationsynchour, nsslapd-accesslog-logrotationsyncmin)"
 ],
 "The server deletes the oldest archived log when the total of all the logs reaches this amount (nsslapd-accesslog-logmaxdiskspace).": [
  null,
  "アクセスログが消費できる最大ディスク容量をMB単位で設定します。ログの容量が設定した容量を超えると、サーバはアーカイブされた最も古いログを削除します。(nsslapd-accesslog-logmaxdiskspace)"
 ],
 "The server deletes the oldest archived log file when available disk space is less than this amount. (nsslapd-accesslog-logminfreediskspace).": [
  null,
  "許可される最小空きディスク容量をMB単位で設定します。空きディスク容量が設定した容量未満になると、十分なディスク領域が解放されるまでアーカイブされた最も古いログファイルを削除します。(nsslapd-accesslog-logminfreediskspace)"
 ],
 "Server deletes an old archived log file when it is older than the specified age. (nsslapd-accesslog-logexpirationtime).": [
  null,
  "指定された期間よりも古い場合、サーバはアーカイブされた古いログファイルを削除します。(nsslapd-accesslog-logexpirationtime)"
 ],
 "Loading Access Log Settings ...": [
  null,
  "アクセスログ設定を読み込んでいます..."
 ],
 "Access Log Settings": [
  null,
  "アクセスログ設定"
 ],
 "Create Mapping": [
  null,
  "マッピングを作成"
 ],
 "Save Mapping": [
  null,
  "マッピングを保存"
 ],
 "$0 SASL Mapping": [
  null,
  "SASLマッピングを$0"
 ],
 "SASL Mapping entry name": [
  null,
  "SASLマッピングのエントリ名を設定します"
 ],
 "SASL Mapping Name": [
  null,
  "SASLマッピング名"
 ],
 "You must provide a name for this mapping": [
  null,
  "マッピングの名前は必ず設定してください"
 ],
 "SASL Mapping Regular Expression": [
  null,
  "SASLマッピングに使用する正規表現を設定します"
 ],
 "Regular Expression": [
  null,
  "正規表現"
 ],
 "You must provide a valid regular expression": [
  null,
  "有効な正規表現を設定してください"
 ],
 "Test Regular Expression": [
  null,
  "正規表現をテストし確認できます"
 ],
 "* Test Regex": [
  null,
  "* 正規表現の確認"
 ],
 "Enter text to test regex": [
  null,
  "テストする正規表現を入力してください"
 ],
 "Test It": [
  null,
  "確認"
 ],
 "The search base or a specific entry DN to match against the constructed DN": [
  null,
  "構築されたDNと照合する検索ベースまたは特定のエントリを設定します"
 ],
 "SASL Mapping Base": [
  null,
  "SASLマッピングのベース"
 ],
 "You must provide a search base": [
  null,
  "有効な検索ベースを設定してください"
 ],
 "SASL mapping search filter": [
  null,
  "SASLマッピングフィルタを設定します"
 ],
 "SASL Mapping Filter": [
  null,
  "SASLマッピングフィルタ"
 ],
 "You must provide an LDAP search filter": [
  null,
  "有効なLDAP検索フィルタを設定してください"
 ],
 "Set the mapping priority for which mappings should be tried first": [
  null,
  "最初に試行されるマッピングの優先順位を設定します"
 ],
 "SASL Mapping Priority": [
  null,
  "SASLマッピングの優先順位"
 ],
 "Priority must be between 1 and 100": [
  null,
  "優先順位は1から100の範囲内で設定してください"
 ],
 "The test string matches the Regular Expression": [
  null,
  "テスト文字列が正規表現と一致しました"
 ],
 "The test string does not match the Regular Expression": [
  null,
  "テスト文字列が正規表現と一致しません"
 ],
 "Invalid regular expression": [
  null,
  "無効な正規表現です"
 ],
 "Successfully create new SASL Mapping": [
  null,
  "SASLマッピングを新しく作成しました"
 ],
 "Error creating new SASL Mapping - $0": [
  null,
  "SASLマッピングの新規作成に失敗しました - $0"
 ],
 "Successfully updated SASL Mapping": [
  null,
  "SASLマッピングを更新しました"
 ],
 "Error updating SASL Mapping - $0": [
  null,
  "SASLマッピングの更新に失敗しました - $0"
 ],
 "Error replacing SASL Mapping - $0": [
  null,
  "SASLマッピングの置換に失敗しました - $0"
 ],
 "Successfully deleted SASL Mapping": [
  null,
  "SASLマッピングを削除しました"
 ],
 "Error deleting SASL Mapping - $0": [
  null,
  "SASLマッピングの削除に失敗しました - $0"
 ],
 "Successfully updated SASL configuration.  These changes require the server to be restarted to take effect.": [
  null,
  "SASL設定を更新しました。変更を適用するにはサーバを再起動してください。"
 ],
 "Error updating SASL configuration - $0": [
  null,
  "SASL設定の更新に失敗しました - $0"
 ],
 "Loading SASL Configuration ...": [
  null,
  "SASL設定を読み込んでいます..."
 ],
 "SASL Settings": [
  null,
  "SASL設定"
 ],
 "Refresh SASL settings": [
  null,
  "SASL設定を更新します"
 ],
 "The maximum SASL buffer size in bytes (nsslapd-sasl-max-buffer-size).": [
  null,
  "最大SASLバッファサイズを設定します。(nsslapd-sasl-max-buffer-size)"
 ],
 "Max SASL Buffer Size": [
  null,
  "最大SASLバッファサイズ"
 ],
 "A list of SASL mechanisms the server will only accept (nsslapd-allowed-sasl-mechanisms).  The default is all mechanisms are allowed.": [
  null,
  "SASLメカニズムのリストを設定します。設定したSASLメカニズムのみが有効になります。なお、デフォルトでは、すべてのメカニズムが許可されます。(nsslapd-allowed-sasl-mechanisms)。"
 ],
 "Allowed SASL Mechanisms": [
  null,
  "許可するSASLメカニズム"
 ],
 "Type SASL mechanism to allow...": [
  null,
  "許可するSASLメカニズムを入力してください..."
 ],
 "Check all sasl mappings until one succeeds or they all fail (nsslapd-sasl-mapping-fallback).": [
  null,
  "1つ成功するか、すべてが失敗するまで、すべてのSASLマッピングを確認します"
 ],
 "Allow SASL Mapping Fallback": [
  null,
  "SASLマッピングのフォールバックを許可する"
 ],
 "Create New Mapping": [
  null,
  "マッピングの新規作成"
 ],
 "Delete SASL Mapping": [
  null,
  "SASLマッピングを削除"
 ],
 "Are you sure you want to delete this SASL mapping?": [
  null,
  "このSASLマッピングを削除しますか？"
 ],
 "Deleting SASL Mapping ...": [
  null,
  "SASLマッピングを削除しています..."
 ],
 "Error loading server configuration for database- $0": [
  null,
  "データべースからサーバ構成を読み込めませんでした - $0"
 ],
 "Error loading database configuration - $0": [
  null,
  "データベース構成を読み込めませんでした - $0"
 ],
 "Error loading default chaining configuration - $0": [
  null,
  "デフォルトのチェイン構成を読み込めませんでした - $0"
 ],
 "Chaining Configuration": [
  null,
  "チェイン構成"
 ],
 "Backups & LDIFs": [
  null,
  "バックアップ & LDIF"
 ],
 "Password Policies": [
  null,
  "パスワードポリシー"
 ],
 "Global Policy": [
  null,
  "グローバルポリシー"
 ],
 "Local Policies": [
  null,
  "ローカルポリシー"
 ],
 "Suffixes": [
  null,
  "サフィックス"
 ],
 "Create new suffix": [
  null,
  "サフィックスの新規作成"
 ],
 "Error getting chaining link configuration - $0": [
  null,
  "チェインリンク設定取得に失敗しました - $0"
 ],
 "Missing the suffix DN": [
  null,
  "サフィックスDNが見つかりません"
 ],
 "Missing the suffix backend name": [
  null,
  "サフィックスバックエンド名が見つかりません"
 ],
 "Successfully create new suffix": [
  null,
  "新しいサフィックスを作成しました"
 ],
 "Error creating suffix - $0": [
  null,
  "サフィックスの作成に失敗しました - $0"
 ],
 "Error loading indexes for $0 - $1": [
  null,
  "$0のインデックスの読み込みに失敗しました - $1"
 ],
 "Error attribute encryption for $0 - $1": [
  null,
  "$0の属性の暗号化に失敗しました - $1"
 ],
 "Error loading VLV indexes for $0 - $1": [
  null,
  "$0のVLVインデックスの読み込みに失敗しました - $1"
 ],
 "Error loading config for $0 - $1": [
  null,
  "$0の設定の読み込みに失敗しました - $1"
 ],
 "Loading suffix configuration for ": [
  null,
  "サフィックス設定を読み込んでいます "
 ],
 "Loading Chaining configuration for ": [
  null,
  "チェイン設定を読み込んでいます"
 ],
 "Loading Database Configuration ...": [
  null,
  "データベース設定を読み込んでいます..."
 ],
 "Create Suffix": [
  null,
  "サフィックスの作成"
 ],
 "Create New Suffix": [
  null,
  "サフィックスの新規作成"
 ],
 "Suffix DN": [
  null,
  "サフィックスDN"
 ],
 "Database suffix, like 'dc=example,dc=com'.  The suffix must be a valid LDAP Distinguished Name (DN).": [
  null,
  "'dc=example,dc=com'のようなデータベースサフィックスを設定します。サフィックスには有効なLDAP識別名(DN)を設定してください。"
 ],
 "The DN of the suffix is invalid": [
  null,
  "サフィックスのDNが無効な形式です"
 ],
 "You must enter a name for the database": [
  null,
  "データベース名は必ず入力してください"
 ],
 "Create The Top Suffix Entry": [
  null,
  "トップサフィックスエントリを作成する"
 ],
 "There is no <b>389-ds-base</b> package installed on this system. Sorry there is nothing to manage...": [
  null,
  "このシステムには389-ds-baseパッケージがインストールされていません。管理するものがありません。"
 ],
 "There are no Directory Server instances to manage": [
  null,
  "管理するディレクトリサーバのインスタンスはありません"
 ],
 "This server instance is not running, either start it from the <b>Actions</b> dropdown menu, or choose a different instance": [
  null,
  "このインスタンスは起動していません。アクションメニューから開始するか、別のインスタンスを選択してください。"
 ],
 "This server instance is running, but we can not connect to it. Check LDAPI is properly configured on this instance.": [
  null,
  "このサーバインスタンスは実行されていますが、接続できません。LDAPIがこのインスタンスで適切に構成されているか確認してください。"
 ],
 "Problem accessing required server configuration. Check LDAPI is properly configured on this instance.": [
  null,
  "必要なサーバ設定にアクセスできません。このインスタンスでLDAPIが適切に構成されているか確認してください。"
 ],
 "Instance was successfully removed": [
  null,
  "インスタンスを削除しました"
 ],
 "Error during instance remove operation - $0": [
  null,
  "インスタンスの削除に失敗しました - $0"
 ],
 "Instance is already running": [
  null,
  "インスタンスはすでに起動しています"
 ],
 "Instance is already stopped": [
  null,
  "インスタンスはすでに停止しています"
 ],
 "Instance was successfully $0": [
  null,
  "インスタンスの処理が成功しました $0"
 ],
 "Error during instance $0 operation - $1": [
  null,
  "インスタンスの$0に失敗しました - $1"
 ],
 "Error during instance check status operation - $0": [
  null,
  "インスタンスのステータスチェックに失敗しました - $0"
 ],
 "Start Instance": [
  null,
  "インスタンスの開始"
 ],
 "Stop Instance": [
  null,
  "インスタンスの停止"
 ],
 "Restart Instance": [
  null,
  "インスタンスの再起動"
 ],
 "Manage Backups": [
  null,
  "バックアップの管理"
 ],
 "Reload Schema Files": [
  null,
  "スキーマファイルを更新する"
 ],
 "Remove This Instance": [
  null,
  "このインスタンスを削除"
 ],
 "Create New Instance": [
  null,
  "新しいインスタンスを作成"
 ],
 "Loading Directory Server Configuration ...": [
  null,
  "ディレクトリサーバの設定を読み込んでいます..."
 ],
 "Directory Server instance list": [
  null,
  "ディレクトリサーバのインスタンス一覧"
 ],
 "Actions": [
  null,
  "アクション"
 ],
 "Server": [
  null,
  "サーバ"
 ],
 "Database": [
  null,
  "データベース"
 ],
 "Schema": [
  null,
  "スキーマ"
 ],
 "Plugins": [
  null,
  "プラグイン"
 ],
 "Monitoring": [
  null,
  "監視"
 ],
 "LDAP Browser": [
  null,
  "LDAPブラウザ"
 ],
 "Remove Instance": [
  null,
  "インスタンスの削除"
 ],
 "Are you really sure you want to delete this instance?": [
  null,
  "本当にこのインスタンスを削除しますか？"
 ],
 "Removing Instance...": [
  null,
  "インスタンスを削除しています..."
 ],
 "Instance name must be less than 80 characters": [
  null,
  "インスタンス名は80文字以下で設定してください。"
 ],
 "Instance name can only contain letters, numbers, and these 4 characters:  - @ : _": [
  null,
  "インスタンス名には、英数字および次の4つの文字(- @ : _)のみを含めることができます。"
 ],
 "Failed to get hostname! $0": [
  null,
  "ホスト名の取得に失敗しました！ $0"
 ],
 "Failed to create installation file! $0": [
  null,
  "インストールファイルの作成に失敗しました！ $0"
 ],
 "Failed to set permissions on setup file $0: $1": [
  null,
  "設定アップファイル $0 の権限設定に失敗しました: $1"
 ],
 "Failed to populate installation file! $0": [
  null,
  "インストールファイルへのデータ入力に失敗しました！ $0"
 ],
 "Successfully created instance: slapd-$0": [
  null,
  "インスタンスの作成に成功しました: slapd-$0"
 ],
 "Failed to set Directory Manager password": [
  null,
  "ディレクトリ管理者パスワードの設定に失敗しました"
 ],
 "Set Directory Manager password...": [
  null,
  "ディレクトリ管理者パスワードを設定しています..."
 ],
 "Create Instance": [
  null,
  "インスタンスの作成"
 ],
 "Creating Instance ...": [
  null,
  "インスタンスを作成中..."
 ],
 "Create New Server Instance": [
  null,
  "新しくサーバインスタンスを作成"
 ],
 "The instance name, this is what gets appended to 'slapd-'. The instance name can only contain letters, numbers, and: # @ : - _": [
  null,
  "'slapd-'に追加されるインスタンス名を設定します。インスタンス名は、英数字および次の5つの文字(# @ : - _)のみ利用可能です。"
 ],
 "Instance Name": [
  null,
  "インスタンス名"
 ],
 "The server port number should be in the range of 0 to 65534.": [
  null,
  "サーバポート番号は0から65534の範囲で指定してください。"
 ],
 "Port 0 will disable non-TLS connections": [
  null,
  "ポート0は非TLS接続を無効にします"
 ],
 "The secure port number for TLS connections. It should be in the range of 1 to 65534.": [
  null,
  "TLS接続用のセキュアポート番号を設定します。1から65534の範囲で指定してください。"
 ],
 "Secure Port": [
  null,
  "セキュアポート"
 ],
 "Create a self-signed certificate database in /etc/dirsrc/ssca directory.": [
  null,
  "/etc/dirsrc/sscaディレクトリに自己署名証明書データベースを作成します。"
 ],
 "Create Self-Signed TLS Certificate": [
  null,
  "自己署名TLS証明書を作成する"
 ],
 "The DN for the unrestricted user": [
  null,
  "特権を持つディレクトリ管理者のDNを設定します"
 ],
 "Enter a valid DN": [
  null,
  "有効なDNを入力してください"
 ],
 "Directory Manager password must be at least 8 characters in length.": [
  null,
  "ディレクトリ管理者のパスワードは8文字以上で設定してください。"
 ],
 "Password must be set and it must match the confirmation password.": [
  null,
  "パスワードを設定してください。パスワードの確認にも同じパスワードを入力してください。"
 ],
 "Confirm the previously entered password.": [
  null,
  "パスワードを確認するため再入力してください。"
 ],
 "Confirmation password must be set and it must match the first password.": [
  null,
  "入力したパスワードが上記で設定したパスワードと異なります。"
 ],
 "Create a database during the installation.": [
  null,
  "インストール中にデータベースを作成します。"
 ],
 "Create Database": [
  null,
  "データベースを作成する"
 ],
 "Database suffix, like 'dc=example,dc=com'. The suffix must be a valid LDAP Distinguished Name (DN)": [
  null,
  "'dc=example,dc=com'のようなデータベースサフィックスを設定します。サフィックスには有効なLDAP識別名(DN)を設定してください。"
 ],
 "The name for the backend database, like 'userroot'. The name can be a combination of alphanumeric characters, dashes (-), and underscores (_). No other characters are allowed, and the name must be unique across all backends.": [
  null,
  "'userroot'のようなバックエンドデータベースの名前を設定します。名前は、英数字、ダッシュ(-)、アンダースコア(_)の組み合わせを使用することができます。他の文字は許可されません。名前は、すべてのバックエンドで一意でなければなりません。"
 ],
 "Name is required": [
  null,
  "データベース名は必須です"
 ],
 "Database Initialization": [
  null,
  "データベースの初期化"
 ],
 "Create Suffix Entry": [
  null,
  "サフィックスエントリの作成"
 ],
 "Create Sample Entries": [
  null,
  "サンプルエントリの作成"
 ],
 "Successfully reloaded schema": [
  null,
  "スキーマを更新しました"
 ],
 "Failed to reload schema files - $0": [
  null,
  "スキーマの更新に失敗しました - $0"
 ],
 "Reload Schema": [
  null,
  "スキーマを更新する"
 ],
 "The name of the database link.": [
  null,
  "データベースリンクの名前"
 ],
 "Server has been backed up": [
  null,
  "サーバがバックアップされました。"
 ],
 "Are you sure you want to restore this backup?": [
  null,
  "このバックアップを復元しますか？"
 ],
 " backup already eixsts with the same name, do you want to replace it?": [
  null,
  "同じ名前のバックアップがすでに存在しています。置き換えますか？"
 ],
 "Replacing ...": [
  null,
  "置き換えています..."
 ],
 "Failed to get objectClasses - $0": [
  null,
  "オブジェクトクラスの取得に失敗しました - $0"
 ],
 "$0 error during plugin loading": [
  null,
  "$0 プラグインの読み込みに失敗しました"
 ],
 "Plugin $0 was successfully modified": [
  null,
  "プラグイン $0 を変更しました"
 ],
 "$0 error during $1 modification": [
  null,
  "$0 $1の変更に失敗しました"
 ],
 "$0 plugin was successfully set to $1. Please, restart the instance.": [
  null,
  "$0 プラグインを$1に設定しました。インスタンスを再起動してください。"
 ],
 "Plugin is not configured": [
  null,
  "プラグインが設定されていません"
 ],
 "All Plugins": [
  null,
  "すべてのプラグイン"
 ],
 "Loading Plugins ...": [
  null,
  "プラグインを読み込んでいます..."
 ],
 "Disable Plugin": [
  null,
  "プラグインを無効にする"
 ],
 "Enable Plugin": [
  null,
  "プラグインを有効にする"
 ],
 "Are you really sure you want to disable this plugin?  Disabling some plugins can cause the server to not start, please use caution.": [
  null,
  "本当にこのプラグインを無効にしますか？一部のプラグインを無効にすると、サーバーが起動しなくなる可能性があります。注意してください。"
 ],
 "Are you sure you want to enable this plugin?": [
  null,
  "このプラグインを有効にしますか？"
 ],
 "Disabling plugin ...": [
  null,
  "プラグインを無効にしています..."
 ],
 "Enabling plugin ...": [
  null,
  "プラグインを有効にしています..."
 ],
 "Error loading suffix RUV - $0": [
  null,
  "サフィックスRUVの読み込みに失敗しました - $0"
 ],
 "Error loading winsync agreements - $0": [
  null,
  "winsync合意の読み込みに失敗しました - $0"
 ],
 "Error loading replication agreements configuration - $0": [
  null,
  "レプリカ合意設定の読み込みに失敗しました - $0"
 ],
 "Error loading replication changelog configuration - $0": [
  null,
  "レプリケーション変更ログ設定の読み込みに失敗しました - $0"
 ],
 "Loading Replication Information ...": [
  null,
  "レプリケーション情報を読み込んでいます..."
 ],
 "Loading Replication Configuration For ": [
  null,
  "レプリケーション情報を読み込んでいます "
 ],
 "Agreements": [
  null,
  "合意"
 ],
 "Tasks": [
  null,
  "タスク"
 ],
 "Conflict Entries": [
  null,
  "競合エントリ"
 ],
 "Logging": [
  null,
  "ログ"
 ],
 "Loading Database Monitor Information ...": [
  null,
  "データベース監視情報を読み込んでいます..."
 ],
 "Loading Server Monitor Information ...": [
  null,
  "サーバ監視情報を読み込んでいます..."
 ],
 "Loading Replication Monitor Information ...": [
  null,
  "レプリケーション監視情報を読み込んでいます..."
 ],
 "Loading Chaining Monitor Information for $0 ...": [
  null,
  "$0のチェイン監視情報を読み込んでいます..."
 ],
 "Loading Monitor Information ...": [
  null,
  "監視情報を読み込んでいます..."
 ],
 "Error loading security configuration - $0": [
  null,
  "セキュリティ設定の読み込みに失敗しました - $0"
 ],
 "Error loading security RSA configuration - $0": [
  null,
  "セキュリティRSA設定の読み込みに失敗しました - $0"
 ],
 "There must be at least one server certificate present in the security database to enable security": [
  null,
  "セキュリティを有効にするには、セキュリティデータベースに少なくとも1つのサーバー証明書が必要です。"
 ],
 "Successfully enabled security.": [
  null,
  "セキュリティを有効にしました。"
 ],
 "Error enabling security - $0": [
  null,
  "セキュリティの有効化に失敗しました - $0"
 ],
 "Error enabling security (RSA cert name)- $0": [
  null,
  "セキュリティの有効化に失敗しました(RSA証明書名) - $0"
 ],
 "Successfully disabled security.": [
  null,
  "セキュリティを無効にしました。"
 ],
 "Error disabling security - $0": [
  null,
  "セキュリティの無効化に失敗しました - $0"
 ],
 "The TLS minimum version must be less than or equal to the TLS maximum version": [
  null,
  "TLSの最小バージョンには、TLSの最大バージョン以下のバージョンを指定してください"
 ],
 "Successfully updated security RSA configuration.": [
  null,
  "セキュリティRSA設定を更新しました。"
 ],
 "Error updating security RSA configuration - $0": [
  null,
  "セキュリティRSA設定の更新に失敗しました - $0"
 ],
 "Successfully updated security configuration.": [
  null,
  "セキュリティ設定を更新しました。"
 ],
 "Error updating security configuration - $0": [
  null,
  "セキュリティ設定の更新に失敗しました - $0"
 ],
 "The name, or nickname, of the server certificate inthe NSS database the server should use (nsSSLPersonalitySSL).": [
  null,
  "サーバが使用するNSSデータベース内のサーバ証明書の名前またはニックネームを設定します。(nsSSLPersonalitySSL)"
 ],
 "Server Certificate Name": [
  null,
  "サーバ証明書名"
 ],
 "Type a sever certificate nickname...": [
  null,
  "サーバ証明書のニックネームを入力してください..."
 ],
 "The minimum SSL/TLS version the server will accept (sslversionmin).": [
  null,
  "サーバが受け入れる最小のSSL/TLSバージョンを設定します。(sslversionmin)"
 ],
 "Minimum TLS Version": [
  null,
  "最小TLSバージョン"
 ],
 "The maximum SSL/TLS version the server will accept (sslversionmax).": [
  null,
  "サーバが受け入れる最大のSSL/TLSバージョンを設定します。(sslversionmax)"
 ],
 "Maximum TLS Version": [
  null,
  "最大TLSバージョン"
 ],
 "Sets how the Directory Server enforces TLS client authentication (nsSSLClientAuth).": [
  null,
  "ディレクトリサーバがTLSクライアント認証を強制する方法を設定します。(nsSSLClientAuth)"
 ],
 "Client Authentication": [
  null,
  "クライアント認証"
 ],
 "Validate server's certificate expiration date (nsslapd-validate-cert).": [
  null,
  "サーバの証明書の有効期限を検証するかどうかを設定します。(nsslapd-validate-cert)"
 ],
 "Validate Certificate": [
  null,
  "証明書の検証"
 ],
 "This parameter can be used to restrict the Directory Server instance to a single IP interface (hostname, or IP address).  This parameter specifically sets what interface to use for TLS traffic.  Requires restart. (nsslapd-securelistenhost).": [
  null,
  "ディレクトリサーバインスタンスを単一のIPインターフェース(ホスト名またはIPアドレス)に制限するために使用できます。特にTLSトラフィックに使用するインターフェースを設定します。設定を反映するには再起動が必要です。(nsslapd-securelistenhost)"
 ],
 "Secure Listen Host": [
  null,
  "セキュアリッスンホスト"
 ],
 "Require all connections use TLS (nsslapd-require-secure-binds).": [
  null,
  "すべての接続にTLSを要求するかどうかを設定します。(nsslapd-require-secure-binds)"
 ],
 "Require Secure Connections": [
  null,
  "セキュアな接続を要求する"
 ],
 "Verify authenticity of a request by matching the host name against the value assigned to the common name (cn) attribute of the subject name (subjectDN field) in the certificate being presented. (nsslapd-ssl-check-hostname).": [
  null,
  "リクエストの真正性を確認するために、証明書のサブジェクト名(subjectDNフィールド)の共通名(cn)属性に割り当てられた値に対してホスト名を照合するかどうかを設定します。(nsslapd-ssl-check-hostname)"
 ],
 "Verify Certificate Subject Hostname": [
  null,
  "証明書のサブジェクトホスト名の検証を行う"
 ],
 "Allow weak ciphers (allowWeakCipher).": [
  null,
  "弱い暗号を許可するかどうかを指定します。(allowWeakCipher)"
 ],
 "Allow Weak Ciphers": [
  null,
  "弱い暗号を許可する"
 ],
 "Allow client-initiated renegotiation (nsTLSAllowClientRenegotiation).": [
  null,
  "クライアント起点の再ネゴシエーションを許可するかどうかを設定します。(nsTLSAllowClientRenegotiation)"
 ],
 "Allow Client Renegotiation": [
  null,
  "クライアント再ネゴシエーションを許可する"
 ],
 "Security Settings": [
  null,
  "セキュリティ設定"
 ],
 "Security Configuration": [
  null,
  "セキュリティ設定"
 ],
 "Security Enabled": [
  null,
  "セキュリティを有効にする"
 ],
 "Certificate Management": [
  null,
  "証明書管理"
 ],
 "Cipher Preferences": [
  null,
  "暗号設定"
 ],
 "Loading Security Information ...": [
  null,
  "セキュリティ情報を読み込んでいます..."
 ],
 "Disable Security": [
  null,
  "セキュリティを無効にする"
 ],
 "Are you sure you want to disable security?": [
  null,
  "セキュリティを無効にしますか？"
 ],
 "ObjectClass $0 was successfully deleted": [
  null,
  "オブジェクトクラス $0 を削除しました"
 ],
 "Error during ObjectClass removal operation - $0": [
  null,
  "オブジェクトクラスの削除に失敗しました - $0"
 ],
 "ObjectClass Name is required.": [
  null,
  "オブジェクトクラス名は必須です。"
 ],
 "ObjectClass $0 - $1 operation was successful": [
  null,
  "オブジェクトクラス $0 の $1 を行いました"
 ],
 "Error during the ObjectClass $0 operation - $1": [
  null,
  "オブジェクトクラスの$0に失敗しました - $1"
 ],
 "Attribute $0 was successfully deleted": [
  null,
  "属性 $0 を削除しました"
 ],
 "Error during Attribute removal operation - $0": [
  null,
  "属性の削除に失敗しました - $0"
 ],
 "Attribute $0 - add operation was successful": [
  null,
  "属性 $0 を追加しました"
 ],
 "Error during the Attribute add operation - $0": [
  null,
  "属性の追加に失敗しました - $0"
 ],
 "Attribute $0 - replace operation was successful": [
  null,
  "属性 $0 を置き換えました"
 ],
 "Error during the Attribute replace operation - $0": [
  null,
  "属性の置き換えに失敗しました - $0"
 ],
 "Loading Schema Information ...": [
  null,
  "スキーマ情報を読み込んでいます..."
 ],
 "Show only the objectclasses that are defined by a user and have the X-ORIGIN set to 'user defined'": [
  null,
  "ユーザによって定義され、X-ORIGIN が 'user defined' に設定されているオブジェクトクラスのみ表示します"
 ],
 "Only Show Non-standard/Custom Schema": [
  null,
  "非標準/カスタムスキーマのみ表示する"
 ],
 "Add ObjectClass": [
  null,
  "オブジェクトクラスを追加"
 ],
 "Show only the attributes that are defined by a user, and have the X-ORIGIN set to 'user defined'": [
  null,
  "ユーザによって定義され、X-ORIGIN が 'user defined' に設定されている属性のみ表示します"
 ],
 "Delete An Objectclass": [
  null,
  "オブジェクトクラスを削除"
 ],
 "Are you sure you want to delete this Objectclass?": [
  null,
  "このオブジェクトクラスを削除しますか？"
 ],
 "Deleting objectclass ...": [
  null,
  "オブジェクトクラスを削除しています..."
 ],
 "Delete An Attribute": [
  null,
  "属性を削除"
 ],
 "Are you sure you want to delete this Attribute?": [
  null,
  "この属性を削除しますか？"
 ],
 "Deleting attribute ...": [
  null,
  "属性を削除しています..."
 ],
 "Database Suffixes": [
  null,
  "データベースサフィックス"
 ],
 "error: please, check browser logs": [
  null,
  "エラー: ブラウザのログを確認してください"
 ],
 "This suffix is empty!": [
  null,
  "このサフィックスは空です！"
 ],
 "Click here to create the top entry": [
  null,
  "トップエントリの作成はこちらから"
 ],
 "Create Top Entry...": [
  null,
  "トップエントリを作成"
 ],
 "Refresh ...": [
  null,
  "更新..."
 ],
 "Tree View": [
  null,
  "ツリー一覧"
 ],
 "remove": [
  null,
  "削除"
 ],
 "to": [
  null,
  "に"
 ],
 "from": [
  null,
  "から"
 ],
 "It will $0 nsAccountLock attribute $1 the entry - $2.": [
  null,
  "$2エントリ$1 nsAccountLock属性を$0します。"
 ],
 "This operation will make sure that these five entries are created at the entry's root suffix (if not, they will be created):": [
  null,
  "この操作により、以下の5つのエントリがエントリのルートサフィックスに作成されます(作成されていない場合):"
 ],
 "added to": [
  null,
  "に追加"
 ],
 "removed from": [
  null,
  "から削除"
 ],
 "The entry - $0 - will be $1 nsRoleDN attribute in cn=nsDisabledRole entry in the root suffix.": [
  null,
  "エントリ $0 はルートサフィックスのcn=nsDisabledRoleエントリ内にあるnsRoleDN属性$1されます。"
 ],
 "Security": [
  null,
  "セキュリティ"
 ],
 "SASL Settings & Mappings": [
  null,
  "SASL設定とマッピング"
 ],
 "LDAPI & Autobind": [
  null,
  "LDAPI&自動バインド"
 ],
 "Loading Server Configuration ...": [
  null,
  "サーバ設定を読み込んでいます..."
 ]
});
